//
// THIS SOFTWARE IS FURNISHED UNDER A LICENSE AND  MAY
// BE USED OR COPIED ONLY IN ACCORDANCE WITH THE TERMS
// OF SUCH LICENSE.
//
//
// COPYRIGHT (c) 1997,1998 BY DIGITAL EQUIPMENT CORPORATION.
// ALL RIGHTS RESERVED.
//

// This is not a Trek application, it is the code inserted in a
// a program by running a Trek application (namely profile).
import dec.trek.RunLog;
import java.io.*;

//
// ProfileLog -- run-time methods associated with profile application.
// Calls to startMethod were inserted at the start of each method being
// profiled, and calls to endMethod at each such method's Return stmts.
//
public final class ProfileLog extends RunLog
{
	// If Thread.class has been instrumented to support per-thread
	// instrumentation, one way to do per-thread stuff is as follows:
	// 1) Make all relevant variables in your xxxLog class NOT be static
	// 2) Have a static jacket method for each real instrumentation method:
	/*
	public static jacket_meth1(...) {
		Thread thr = Thread.currentThread();
		if (thr.userDefined==null)
			thr.userDefined = new XxxLog("xxx");
		XxxLog xxxlog = (XxxLog)thr.userDefined;
		xxxlog.real_meth1(...);
	}
	*/

	private static ProfileLog log;					// Handle for log machinery
	private static boolean toldUser;				// For one-time error msg

	// The info saved for each method being profiled
	private static int profiled;
	private static long elapsed[];
	private static long entered[];
	private static int recurse[];
	private static int times[];
	private static String names[];

	// Tabulate a return to a method. Store elapsed time for this usage
	// of the method being profiled
	// ctr = slot assigned to method from which call to endMethod is coming
	public static void endMethod(int ctr)
	{
		if (names==null) return;
		if (active) return;		// In case currentTimeMillis is being profiled
		active = true;

		// When method called recursively, log 1 entry for whole chain.
		// Stop gathering real data for 0-duration methods eventually
		if (--recurse[ctr]==0)
		{
			if (times[ctr]<=1000 || (double)elapsed[ctr]/1000>.04)
				elapsed[ctr] += System.currentTimeMillis() - entered[ctr];
		}
		active = false;
	}

	// Generate the profiling report to profile.log. A call to
	// this method is inserted at the end of your program's main method.
	public static void endTrek()
	{
		// Do the one and only endMethod call for main
		// Open the report file
		endMethod(0);
		if (active) return;		// In case java.* method is being profiled
		active = true;
		log = new ProfileLog("profile");

		// Do the work. Output report hdr and line for each profiled method
		log.writeEntry(" Number   MS/Call   Total MS     Method Name");
		for (int i=0; i<profiled; i++) {
			if (names[i]==null) continue;
			double dlap = (double)elapsed[i];
			int cnt = times[i];

			// Extrapolate elapsed time of high freq quick methods from 1st
			// 1000 calls to such methods. (Logging of such calls ends after 
			// 1000th call on assumption we now know the proper average).
			if (cnt>1000 && dlap/1000<=.04)
				dlap *= cnt/1000;
			
			// Output line for curr method, if "significant" -- meaning
			// at least 1% of time was spent in the method.
			if (dlap>0 && elapsed[0]/dlap <= 100)
			{
				double ms = dlap / cnt;
				log.writeItem(cnt, 7);
				log.writeItem(ms, 8, 1);
				log.writeItem(dlap, 11, 0);
				log.writeItem(names[i], 5, 0);
				log.writeEntry();
			}

			// Chk for patho behavior during run
			if (recurse[i]<0)
				log.writeEntry("endMethod without startMethod for " + names[i]);
		}
		log.closeLog();
		active = false;
	}

	// Construct a profile log
	private ProfileLog(String sug)
	{
		super(sug);
	}

	// Tabulate a call to a method
	// ctr = slot assigned to method from which call to startMethod is coming
	// name = that method's name, for the report. (A more extensive profiler
	//		would include the arglist to distinguish overloaded names)
	// Notes:
	//	Even though time is tabulated in milliseconds, and time per call might
	//	be much less than that, the reported data is statistically accurate
	//	if the method is called many times. For example, if time/call is .1ms
	//	and times called was 100, there would on average be 90 0ms's and 10
	//	1ms's tabulated (and (90*0 + 10*1)/100 = .1)
	public static void startMethod(int ctr, String name)
	{
		if (active) return;		// In case currentTimeMillis is being profiled
		active = true;

		// Insure dont get here without having called startTrek
		if (names==null)
		{
			tellOnce("Reached profiled method and main did not init profiling");
			active = false;
			return;
		}

		// Gather the data unless recursed already, but process only
		// count if have already determined time/call is effectively 0
		if (++recurse[ctr]==1)
		{
			if (times[ctr]<=1000 || (double)elapsed[ctr]/1000>.04)
			{
				names[ctr] = name;
				entered[ctr] = System.currentTimeMillis();
			}
			times[ctr] += 1;		// Tabulate another call
		}
		active = false;
	}

	// Create the arrays to tabulate the profiling data. A call to startTrek
	// is inserted at the end of the program's main method. Note that the
	// insertion occurs at the end of the trek. In other words, dont confuse
	// where the code is inserted with when the insertion is done.
	// num = number of profiled methods
	// name = class.main
	public static void startTrek(int num, String name)
	{
		profiled = num;

		// Init storage before allowing other instrumentation methods to be
		// called. In theory, the storage inits could trigger such calls.
		elapsed = new long[profiled];
		entered = new long[profiled];
		recurse = new int[profiled];
		times = new int[profiled];
		names = new String[profiled];
		sysInited = 1;

		// Do the one and only startMethod call for main. In other words,
		// the entry for main will simply show program's entire runtime.
		startMethod(0, name);
	}

	// Output msg about bypassing init-method
	private static void tellOnce(String msg)
	{
		if (!toldUser)
		{
			System.out.println(msg);
			toldUser = true;
		}
	}

}
