//
// THIS SOFTWARE IS FURNISHED UNDER A LICENSE AND MAY
// BE USED OR COPIED ONLY IN ACCORDANCE WITH THE TERMS
// OF SUCH LICENSE.
//
//
// COPYRIGHT (c) 1997 BY DIGITAL EQUIPMENT CORPORATION.
// ALL RIGHTS RESERVED.
//

import dec.trek.RunLog;
import java.io.*;

//
//
// WatchLog. Run-time environment for watch class
// To support multiple threads efficiently (without having to have special
// Thread.class file), interface is setup so that there is no per-thread 
// state maintained between calls 
//
//
public final class WatchLog extends RunLog
{
	private static WatchLog log;					// Handle for log machinery
	private static int counters[] = new int[10];	// Loop iters. 1 entry per loop
	private static int options;						// Watch context options if any
	private static String trekId;					// Object Id of trek if any
	private static int valStart = 62;

	// Display numeric arg. These methods may all look the same,
	// but newval's data type controls which overloaded method is called.
	public static void argValue(boolean newval, String name)
	{
		if (active) return;
		active = true;
		log.startArg(name);
		log.writeItem(newval);
		log.writeEntry();
		active = false;
	}
	public static void argValue(double newval, String name)
	{
		if (active) return;
		active = true;
		log.startArg(name);
		log.writeItem(newval);
		log.writeEntry();
		active = false;
	}
	public static void argValue(int newval, String name)
	{
		if (active) return;
		active = true;
		log.startArg(name);
		log.writeItem(newval);
		log.writeEntry();
		active = false;
	}
	public static void argValue(long newval, String name)
	{
		if (active) return;
		active = true;
		log.startArg(name);
		log.writeItem(newval);
		log.writeEntry();
		active = false;
	}

	// Display non-numeric arg
	public static void argValue(String newval, String name)
	{
		if (active) return;
		active = true;
		String str = newval!=null ? '"' + newval + '"' : "null";
		log.startArg(name);
		log.writeItem(str);
		log.writeEntry();
		active = false;
	}
	public static void argValue(Object newval, String name)
	{
		if (active) return;
		active = true;

		// Normalize string representation of the object
		String str = null;
		if (newval!=null)
		{
			int objadd;
			str = toObjectString(newval);
			if (str.length()==0)
				str = "empty";
			else if ((objadd=str.indexOf('@'))>=0
				&& newval.getClass().toString().endsWith(str.substring(0,objadd)))
				str = str.substring(objadd);
		}

		// Append this arg to arglist
		log.startArg(name);
		log.writeItem(str);
		log.writeEntry();
		active = false;
	}

	// Display method name, and StackTrace and thread info if requested
	public static void enterMethod(String methname, String trekid, int opts)
	{
		if (active) return;
		active = true;
		options = opts;					// Must be constant for whole app
		trekId = trekid;
		insureInited();

		// Identify method, and do other stuff as requested
		log.writeEnter(methname);
		if ((options&watch.ctx_thread)!=0)
			log.writeThreadInfo();
		if ((options&watch.ctx_stack)!=0)
			log.writeStackTraceEntry();
		active = false;
	}

	// Display method-exited (called if no retval or user not interested)
	// When retval desired, exitValue via printLine is used
	public static void exitMethod(String methname)
	{
		if (active) return;
		active = true;
		log.writeExit(methname);
		active = false;
	}

	// Init resources needed by watch logic
	// Notes:
	//	Call dynamically rather than in fixed loc to allow someone to instrument
	//	class library and use it with multiple programs. However all except
	//	printValue must appear after an enterMethod call.
	private static void insureInited()
	{
		// Chk log outside sync to avoid unnec sync
		// Note: a thread may not immediately see changes made by another
		// thread to non-volatile variables. But that is okay here because it
		// will only lead to a "false negative" -- which will be corrected
		// inside the synchronized block
		if (log!=null)
			return;

		// Chk log inside sync so 2 threads dont try to create log
		// Open log and do other 1-time inits
		// If less room because of thread id, adjust val-start
		synchronized (counters) {
			if (log!=null)
				return;
			if ((options&watch.ctx_thread)!=0)
				valStart -= thridLength/2;

			// Hack: See if data log... Currenly yes means doctorj
			// and no means Watch console app.
			// If former, 2 cases:
			// 1) App running in doctorj, create temp log (djwatch.tmp)
			// 2) App running standalone, create mainclass.log
			try {
				if (trekId!=null)
				{
					String permlog = System.getProperty("DOCTORJ");
					if (permlog==null || !permlog.equals("1"))
					{
						String clname;
						int end = trekId.indexOf('\t');
						clname = end>=0 ? trekId.substring(0, end) : trekId;
						end = clname.indexOf(';');
						clname = end>=0 ? clname.substring(0, end) : clname;
						log = new WatchLog(clname, LOG_DATA);
					}
					else log = new WatchLog("djwatch.tmp", LOG_DATA);

					// Setup to restore trek when log read, if applic
					log.dataStream.writeInt(options);
					log.dataStream.writeUTF(trekId);
				}
				else log = new WatchLog("watch", LOG_TEXT);
			}
			catch (Exception e) {
				throw new RuntimeException("Error initializing log: " + e.toString());
			}
			RunLog.sysInited = 1;
		}
	}

	// Display if result
	private static final String path[] = {"ELSE", "THEN", "Past IF"};
	public static void printIfResult(int result, String context)
	{
		if (active) return;
		active = true;

		// Do then, else, or past-if
		log.startIf(context);
		log.writeItem(path[result]);
		log.writeEntry();
		active = false;
	}

	// Display stmt without an associated value (eg. catch stmt & before SHOW)
	public static void printStatement(String context)
	{
		if (active) return;
		active = true;
		log.writeStatement(context);
		active = false;
	}

	// Display case branched to by switch
	public static void printSwitch(int caseval, String context)
	{
		if (active) return;
		active = true;
		log.startIf(context);
		log.writeItem(caseval);
		log.writeEntry();
		active = false;
	}

	// Display statement
	public static void printWhile(int ctr, String context)
	{
		if (active) return;
		active = true;

		// Insure counter array big enough
		if (ctr >= counters.length)
		{
			int more[] = new int[ctr+50];
			System.arraycopy(counters, 0, more, 0, counters.length);
			counters = more;
		}
		
		// Create log entry
		log.startWhile(context);
		log.writeItem(counters[ctr]);
		counters[ctr] += 1;
		log.writeEntry();
		active = false;
	}

	// Display new value of numeric watched variable
	public static void printValue(boolean newval, String context)
	{
		if (active) return;
		active = true;
		insureInited();
		log.startValue(context);
		log.writeItem(newval);
		log.writeEntry();
		active = false;
	}
	public static void printValue(int newval, String context)
	{
		if (active) return;
		active = true;
		insureInited();
		log.startValue(context);
		log.writeItem(newval);
		log.writeEntry();
		active = false;
	}
	public static void printValue(long newval, String context)
	{
		if (active) return;
		active = true;
		insureInited();
		log.startValue(context);
		log.writeItem(newval);
		log.writeEntry();
		active = false;
	}
	public static void printValue(double newval, String context)
	{
		if (active) return;
		active = true;
		insureInited();
		log.startValue(context);
		log.writeItem(newval);
		log.writeEntry();
		active = false;
	}
	public static void printValue(float newval, String context)
	{
		if (active) return;
		active = true;
		insureInited();
		log.startValue(context);
		log.writeItem(newval);
		log.writeEntry();
		active = false;
	}

	// Display new value of non-numeric watched variable
	// Note that all come to String variation eventualy
	public static void printValue(Object newval, String context)
	{
		if (active) return;
		active = true;
		insureInited();

		// Output string for object
		String str = toObjectString(newval);
		log.startValue(context);
		log.writeItem(str);
		log.writeEntry();
		active = false;
	}
	public static void printValue(String newval, String context)
	{
		if (active) return;
		active = true;
		insureInited();
		log.startValue(context);
		log.writeItem(newval);
		log.writeEntry();
		active = false;
	}

	// Display new value of numeric watched array element
	public static void printValue(boolean newval, int arridx, String context)
	{
		if (active) return;
		active = true;
		insureInited();
		log.startValue(context, arridx);
		log.writeItem(newval);
		log.writeEntry();
		active = false;
	}
	public static void printValue(int newval, int arridx, String context)
	{
		if (active) return;
		active = true;
		insureInited();
		log.startValue(context, arridx);
		log.writeItem(newval);
		log.writeEntry();
		active = false;
	}
	public static void printValue(long newval, int arridx, String context)
	{
		if (active) return;
		active = true;
		insureInited();
		log.startValue(context, arridx);
		log.writeItem(newval);
		log.writeEntry();
		active = false;
	}
	public static void printValue(float newval, int arridx, String context)
	{
		if (active) return;
		active = true;
		insureInited();
		log.startValue(context, arridx);
		log.writeItem(newval);
		log.writeEntry();
		active = false;
	}
	public static void printValue(double newval, int arridx, String context)
	{
		if (active) return;
		active = true;
		insureInited();
		log.startValue(context, arridx);
		log.writeItem(newval);
		log.writeEntry();
		active = false;
	}

	// Display new value of numeric watched array element
	public static void printValue(String newval, int arridx, String context)
	{
		if (active) return;
		active = true;
		insureInited();
		log.startValue(context, arridx);
		log.writeItem(newval);
		log.writeEntry();
		active = false;
	}
	public static void printValue(Object newval, int arridx, String context)
	{
		if (active) return;
		active = true;
		insureInited();
		log.startValue(context, arridx);
		log.writeItem(toObjectString(newval));
		log.writeEntry();
		active = false;
	}

	// Do various entries that start with a string item
	private void startArg(String name)
	{
		if (logType==LOG_TEXT)
			name = "Arg " + name + "=";
		writeString(name, Jtu.wit_arg, false);
	}

	// If-clause
	private void startIf(String context)
	{
		if (logType==LOG_TEXT)
			context = toPrefix(context, Jtu.wit_if);
		writeString(context, Jtu.wit_if, false);
	}

	// Switch case
	private void startSwitch(String context)
	{
		if (logType==LOG_TEXT)
			context = toPrefix(context, Jtu.wit_switch);
		writeString(context, Jtu.wit_switch, false);
	}

	// Start new-value entry, including retvals
	private void startValue(String context)
	{
		// Chk for exit-method hack
		int exithack = context.charAt(0);
		if (context.startsWith("?"))
		{
			context = context.substring(1);
			if (logType==LOG_TEXT)
				context = toPrefix(context + " exited with ", Jtu.wit_exitvalue);
			writeString(context, Jtu.wit_exitvalue, false);
		}

		// Else normal case
		else {
			if (logType==LOG_TEXT)
				context = toPrefix(context, Jtu.wit_value);
			writeString(context, Jtu.wit_value, false);
		}
	}

	// Start value-change entry for array element
	// context = src stmt of use
	// arridx = index of element that has changed
	private void startValue(String context, int arridx)
	{
		if (logType==LOG_TEXT)
			context = toPrefix(context, Jtu.wit_value);
		writeString(context, Jtu.wit_element, false);
		if (logType==LOG_TEXT)
			writeItem("[" + arridx + "]=");
		else writeItem(arridx);
	}

	// While stmt
	private void startWhile(String context)
	{
		if (logType==LOG_TEXT)
			context = toPrefix(context, Jtu.wit_while);
		writeString(context, Jtu.wit_while, false);
	}

	// Return type-specific string repr of newval. Note that this
	// has the risk that the object's envir has not been adequately setup
	// at time watchpt is hit for valueOf/toString to perform correctly.
	// So fallback to Object.toString if need be (ie. dont use toString
	// of newval's actual class, instead specifically use Object.toString).
	// Notes:
	//	valueOf chks for null before calling toString
	private static String toObjectString(Object newval)
	{
		try {
			return String.valueOf(newval);
		}
		catch (Throwable t) {
			return ((Object)newval).toString();
		}
	}

	// Return formatted context string to output
	// context = source statement usually
	// wit = type of watch entry
	// valstart = column to abut context string with
	private static final int thridLength = 10;
	private String toPrefix(String context, int wit)
	{
		return Jtu.toPaddedPrefix(context, wit, valStart);
	}
	// Constructor
	private WatchLog(String sug, int logtype)
	{
		super(sug, logtype);
	}

	// Do various whole entries that start with a string item
	private void writeEnter(String name)
	{
		if (logType==LOG_TEXT)
			writeString(name + " Called", Jtu.wit_enter, true);
		else writeString(name, Jtu.wit_enter, true);
	}
	private void writeExit(String name)
	{
		if (logType==LOG_TEXT)
			writeString(name + " exited", Jtu.wit_exit, true);
		else writeString(name, Jtu.wit_exit, true);
	}
	private void writeStatement(String name)
	{
		writeString(name, Jtu.wit_stmt, true);
	}

	// Write a stack dump as a single entry
	private PrintWriter traceStream;
	private ByteArrayOutputStream traceBuffer;
	private void writeStackTraceEntry()
	{
		String trace;

		// Init trace-building buffer if havent
		try {
			if (traceBuffer==null)
			{
				traceBuffer = new ByteArrayOutputStream();
				traceStream = new PrintWriter(traceBuffer);
			}

			// Gen the trace
			new Exception("Stack").printStackTrace(traceStream);
			traceStream.flush();
			trace = traceBuffer.toString();
			traceBuffer.reset();
		}

		catch (Exception e) {
			e.printStackTrace();
			throw new RuntimeException("Could not log item");
		}

		// Strip the in-WatchLog stuff at start
		int endjunk = trace.indexOf("enterMethod");
		if (endjunk>=0)
		{
			endjunk = trace.indexOf('\n', endjunk+10);
			if (endjunk>=0)
			{
				if (logType!=LOG_TEXT)
					trace = trace.substring(endjunk+1);
				else trace = "Stack trace:" + trace.substring(endjunk);
			}
		}

		// Strip new-lines at end and write entry
		int len;
		for (len=trace.length(); trace.charAt(len-1)<' '; len--) ;
		trace = trace.substring(0, len);
		writeString(trace, Jtu.wit_stack, true);
	}

	// Write info describing current thread
	private void writeThreadInfo()
	{
		// Gather thread info
		Thread thread = Thread.currentThread();
		String name = thread.getName();
		String group = thread.getThreadGroup().getName();
		int priority = thread.getPriority();

		// Write data or text
		if (logType==LOG_TEXT)
		{
			String info = "Thread " + group + "." + name
				+ " (Priority=" + priority + ")";
			writeString(info, Jtu.wit_thread, true);
		}
		else {
			writeString(name, Jtu.wit_thread, false);
			writeItem(group);
			writeItem(priority);
			writeEntry();
		}
	}

	// Write first item of an entry that starts with a string
	// str = string to start entry with
	// wit = type code for entry
	private void writeString(String str, int wit, boolean done)
	{
		// Use thread object's hash code as id if applic
		int thrid = (options&watch.ctx_thread)!=0
			? Thread.currentThread().hashCode()
			: 0;

		// Write the item (and thread id if requested)
		//
		try {
			if (logType==LOG_TEXT)
			{
				if (thrid!=0)
				{
					writeItem(thrid, thridLength-2);
					writeItem(": ");
				}
				writeItem(str);
			}

			// For data log. Dont use writeItem so that datatype tags
			// (like ITEM_STRING) are not written out. In this context,
			// the functional tag "wit" subsumes need.
			else {
				dataStream.writeByte(wit);
				if (thrid!=0)
					dataStream.writeInt(thrid);
				dataStream.writeUTF(str);
			}
		}

		catch (Exception e) {
			e.printStackTrace();
			throw new RuntimeException("Could not log item");
		}

		// If 1-item entry, term it now
		if (done)
			writeEntry();
	}

}

