//
// THIS SOFTWARE IS FURNISHED UNDER A LICENSE AND  MAY
// BE USED OR COPIED ONLY IN ACCORDANCE WITH THE TERMS
// OF SUCH LICENSE.
//
//
// COPYRIGHT (c) 1997 BY DIGITAL EQUIPMENT CORPORATION.
// ALL RIGHTS RESERVED.
//

package dec.trek;
import java.io.*;

//
//
// Attr. Handle attributes generically. Some "attributes" are pretty
// important -- in partic Code, and they are handled as their own classes.
//
//
class Attr 
{
	private int nameIdx;		// Cp entry of attr name
	byte value[];				// Raw bytes in attr value if any,
								// or null for internalized attr.
								// Note: 0-len possible in 1st case

	Attr(int idx)
	{
		nameIdx = idx;
	}

	// Outputs presentation string for an un-internalized attribute
	void dump(PrintWriter chan, ClassFile cf, String pad)
	{
		String tail = value.length!=1
			? " (" + value.length + " bytes)" : " (1 byte)";
		Trek.println(chan, pad + toName(cf) + " attribute" + tail);

		// Output hex bytes for the value
		String str = null;
		pad += "  ";					// Data lines indented by 2
		for (int i=0; i<value.length; i++ ) {
			if (i%16 == 0)
			{
				if (i!=0) Trek.println(chan, str);

				// Insure rows are column-aligned
				str = pad;
				if (value.length > 16 && i<16) str += " ";
				if (value.length > 112 && i <112) str += " ";
				str += i + ": ";
			}
			else if (i%4 == 0)
				str += " ";

			// A pair of hex digits for each byte of data
			int val = value[i] & 0xff;
			String hex = Integer.toHexString(val);
			if (hex.length()==1)
				str += "0" + hex;
			else str += hex;
		}

		// Dump final line if not empty
		if (value.length%16!=0)
			Trek.println(chan, str);
		return;
	}

	// Create context string for error during processing of attributes
	// classfile = class being processed
	// doing = action being performed at time of exception
	// mem = mem being processed if any (or null)
	// name = name of the attribute
	// i = index of attr when problem occurred
	// e = exception that is occurring
	private static void logError(String doing, ClassFile cf,
		Member mem, String name, int i, Exception e)
	{
		String mname = "";
		String elem = "";

		if (mem != null) mname = "M=" + mem.toName() + ";";
		if (name != null) elem = "(" + name + ")";
		elem = "E=attribute[" + i + "]" + elem + ";";
		cf.logError("D="+doing+";" + mname + elem, e);
	}

	// Read the current object's attributes, exploding important attr
	// out as separate types of objects. Note that this leaves bodyless
	// entries in the attrs array (ie. code must use attr name to get to
	// object(s) where the data of the attr is stored.
	// chan = where to read from
	// cf = classfile being read from
	// mem = null says reading class attr. non-null is either method or field
	static Attr[] readAttrs(DataInput chan, ClassFile cf, Member mem)
		throws IOException
	{
		Attr attrs[] = null;
		String name = "";
		int i = -1;

		// Materialize obj
		Field field = mem instanceof Field ? (Field)mem : null;
		Method method = mem instanceof Method ? (Method)mem : null;

		// Get # of attributes and alloc room for them
		try {
			int num = chan.readUnsignedShort();
			attrs = new Attr[num];				// Get cnt & alloc attr array

			// Read each attr. Complicated because there are diff kinds of attr
			for (i=0; i<attrs.length; i++) {
				int idx = chan.readUnsignedShort();	// Identifies name of attr
				int len = chan.readInt();			// Bytes of data in attr
				Attr attr = new Attr(idx);		// Create attr & fill-in hdr
				attrs[i] = attr;				// Insert it in array

				// Reading the rest is name-of-attr specific.
				name = cf.constant(idx).toInternalString();
				if (mem instanceof Method && name.equals("Code"))
					method.readCode(chan, len);
				else if (mem instanceof Method && name.equals("LocalVariableTable"))
					Local.readLocals(chan, method);
				else if (mem instanceof Method && name.equals("LineNumberTable"))
					Linum.readLinums(chan, method);
				else if (mem instanceof Method && name.equals("Exceptions"))
					method.readThrown(chan);
				else if (mem instanceof Field && name.equals("ConstantValue"))
					field.initIdx = chan.readUnsignedShort();
				else if (mem==null && name.equals("SourceFile"))
					cf.fileIdx = chan.readUnsignedShort();
				else if (mem==null && name.equals("com.compaq.JTrekCreated"))
					cf.jtrekCreated = true;

				// If attr not needed otherwise, create/read attr object
				// (which is just an opaque seq of bytes) into attrs array.
				else {
					attr.value = new byte[len];
					chan.readFully(attr.value);
				}
			}									// End loop
		}										// End try

		// Gen msg for error if any
		catch (Exception e) {
			logError("reading", cf, mem, name, i, e);
		}
		return attrs;
	}											// End readAttrs

	// Returns name of this attribute
	String toName(ClassFile cf)
	{
		return cf.constant(nameIdx).toInternalString();
	}

	// Write the current object's attributes, un-exploding internalized attr
	// chan = where to write to
	// attrs = array of attrs to write
	// cf = classfile being written to
	// obj = null says reading class attr. non-null is either method or field
	static void writeAttrs(DataOutput chan, Attr attrs[],
		ClassFile cf, Member mem)
		throws IOException
	{
		String name = "";
		int i = -1;

		// Materialize obj
		Field field = mem instanceof Field ? (Field)mem : null;
		Method method = mem instanceof Method ? (Method)mem : null;

		// Get # of attributes and alloc room for them
		try {
			chan.writeShort(attrs.length);			// # of attr

			// Write each attr. Complicated because there are diff kinds of attr
			for (i=0; i<attrs.length; i++) {
				Attr attr = attrs[i];
				chan.writeShort(attr.nameIdx);		// Identifies name of attr

				// Writing the rest is name-of-attr specific.
				name = cf.constant(attr.nameIdx).toInternalString();
				if (mem instanceof Method && name.equals("Code"))
					method.writeCode(chan);
				else if (mem instanceof Method && name.equals("LocalVariableTable"))
					Local.writeLocals(chan, method);
				else if (mem instanceof Method && name.equals("LineNumberTable"))
					Linum.writeLinums(chan, method);
				else if (mem instanceof Method && name.equals("Exceptions"))
					method.writeThrown(chan);
				else if (mem instanceof Field && name.equals("ConstantValue"))
				{
					chan.writeInt(2);
					chan.writeShort(field.initIdx);
				}
				else if (mem==null && name.equals("SourceFile"))
				{
					chan.writeInt(2);
					chan.writeShort(cf.fileIdx);
				}
				else if (mem==null && name.equals("com.compaq.JTrekCreated"))
					chan.writeInt(0);

				// Not-internalized attr. Write the opaque bytes
				else {
					chan.writeInt(attr.value.length);
					chan.write(attr.value);
				}
			}									// End loop
		}										// End try

		// Gen msg for error if any
		catch (Exception e) {
			logError("writing", cf, mem, name, i, e);
		}
		return;
	}

}
