//
// THIS SOFTWARE IS FURNISHED UNDER A LICENSE AND  MAY
// BE USED OR COPIED ONLY IN ACCORDANCE WITH THE TERMS
// OF SUCH LICENSE.
//
//
// COPYRIGHT (c) 1997 BY DIGITAL EQUIPMENT CORPORATION.
// ALL RIGHTS RESERVED.
//

package dec.trek;
import java.util.*;
import java.util.zip.*;
import java.io.*;

//
// ClassFile. The data/methods for one .class file
//
public class ClassFile implements Sortable
{
	private boolean caseMiss;		// True if another file in classpath has 
									// same case-ignored name
	Method clinit;					// Pts to class's <clinit> method if any
	int clinitStatus;				// See clinit_xxx
	private int cpVirginSize;		// # of entries originally in cp
	private String fullname = "";	// External-fmt qualified name of classfile
	boolean importMode;				// True if doing import-driven name display
	private boolean hadError;		// Dont write out classfiles that had errors
	private boolean importSetup;	// True if import-info has been setup
	boolean jtrekCreated;			// This classfile was instrumented by JTrek
	private long lastWritten;		// Sys-dependent time stamp when last saved
	boolean modified;				// True if classfile updated during scan
	Vector rawFields;				// Unsorted fields (see dump)
	private long size;				// Number of bytes in file
	public Object userDefined;		// User-owned field

	// Context for relative navigation of trek's classfiles
	private int idx;			// Index of classfile in its trek
	Trek parent;

	// Contents of classfile (*=user-visible component of program hierarchy)
	private int magic;			// Number that marks file as .class file
	private short minorVersion;
	private short majorVersion;	// JVM version it was written for
	Vector cp;					// Contains constants/symbols of this class
	Hashtable cphash;			// Ditto, but hashed for value-based lookup
	private short accessFlags;
	int thisClass;				// Index of this class in cp
	int superClass;				// Index of its super class
	private int interfaces[];	// Indexes of its interfaces it implements
	int fileIdx;				// Index source file name if specified
    Vector fields;				// *Each data member of this class
	Vector methods;				// *Each method member of this class
	Attr attrs[];				// Each attribute of this class

	// Clinit states
	static final int clinit_unread = -1;	// Whole classfile not read (nyi)
	private static final int clinit_unknown = 0;	// Read but not yet analyzed
	private static final int clinit_simple = 1;		// Just init clauses
	private static final int clinit_fancy = 2;		// Fancy init code


	// Add a constant to pool
	// cpi = constant to add, or null to add empty slot (needed for 64-bit consts)
	// Notes:
	//	cphash created at end of read phase because key of entry depends on
	//	its lower-level entries, and they can be later in pool than curr entry
	void addConstant(CpInfo cpi)
	{
		cp.addElement(cpi);
		if (cpi!=null)
		{
			cpi.idx = cp.size() - 1;
			if (cphash!=null)
				cphash.put(cpi.keyOf(), cpi);
		}
	}

	// Adds the specified class to this classfile's interfaces
	// qcname = fully qualified name of class to add
	// Notes:
	//	Mainly useful for code-less interfaces, like Serializable
	public void addInterface(String qcname)
	{
		int newif[] = new int[interfaces.length+1];
		System.arraycopy(interfaces, 0, newif, 0, interfaces.length);
		TypeRef tref = TypeRef.add(qcname, this);
		newif[interfaces.length] = tref.idx;
		interfaces = newif;
		modified = true;
	}

	// Create pool and set its capacity. Extra room in hash table helps spread
	// entries over available hash slots.
	void addPool(int len)
	{
		cp = new Vector(len);
		addConstant(null);			// 0th entry always empty
	}

	// Insure item index matches where stored in vector. Binding is 
	// delayed until all read because idx's in flux while sorting going on
	static void bindIndexes(Trek trek)
	{
		int vecsize = trek.cfiles.size();
		for (int i=0; i<vecsize; i++) {
			ClassFile cf = (ClassFile)trek.cfiles.elementAt(i);
			cf.idx = i;
		}
	}

	// Create obj so have something to read classfile into
	ClassFile(Trek trek)
	{
		parent = trek;
		idx = -1;
	}

	// Restore classfile in-memory to what it was originally
	void clean()
	{
		modified = false;

		// Restore pool to what it had when 1st read in
		int len = cp.size();
		for (int i=cpVirginSize; i<len; i++) {
			CpInfo cpi = constant(i);
			if (cpi!=null && cphash!=null)
				cphash.remove(cpi.keyOf());
		}
		cp.setSize(cpVirginSize);

		// Remove inserted code and its side-effects on method properties
		for (Method meth=firstMethod(); meth!=null; meth=meth.next())
		{
			meth.addLocals = meth.addStack = 0;
			meth.callLocalsBase = -1;
			Statement stn = null;
			for (Statement st=meth.first; st!=null; st=stn) {
				stn = st.next;
				if (st.type<=Trek.ST_PIECE)
					st.remove();
				else st.mate = null;
			}
		}
	}

	// Returns std 1,0,-1 from comparing This to arg
	// this = the object that is in the vector
	// obj = the object/value outside the vector
	public int compareTo(Object obj)
	{
		String key, objkey;

		// Comparing to a string?
		if (obj instanceof String)
		{
			objkey = (String)obj;

			// Turn around key to by-class
			if (parent.sortBy==Trek.CLORD_BYCLASS)
			{
				int idx = objkey.lastIndexOf('/');
				if (idx<0)
					objkey += "@";			// Not in package
				else objkey = objkey.substring(idx+1) + "@"
					+ objkey.substring(0, idx);
			}
		}

		// Key in another object
		else {
			ClassFile cf = (ClassFile)obj;
			objkey = cf.constructKey();
		}

		// Compare it to key from This and return result
		key = constructKey();
		return key.compareTo(objkey);
	}

	// Convert external key to actual key for sort order
	private String constructKey()
	{
		if (parent.sortBy==Trek.CLORD_BYPACKAGE)
			return constant(thisClass).toInternalString();
		else return toName() + "@" + toPackage('/');
	}

	// Access a constant of this classfile
	final CpInfo constant(int idx)
	{
		return (CpInfo)cp.elementAt(idx);
	}

	int constants()
	{
		return cp.size();
	}

	// Returns size of constant pool vector
	int cpSize()
	{
		return cp.size();
	}

	// Dump the specified aspects of classfile into outdir\its-name.dmp
	public void dump(int flags)
	{
		// Log where we are
		if ((Trek.debug&Trek.Db_at) != 0)
			Trek.println(System.out, "Dumping " + toQualifiedName());

		// Set output file and setup pkg-import info
		// Open the dump file and do its hdr and body
		String fname = toInternalName();
		int namepos = fname.lastIndexOf('/') + 1;
		fname = parent.outDir + fname.substring(namepos) + ".dmp";
		setImportedPackages();
		importMode = true;
		try {
			PrintWriter chan = new PrintWriter(
				new BufferedWriter(new FileWriter(fname)));
			dumpHeader(chan, flags);
			dumpClassBody(chan, flags);
			chan.close();
		}
		catch (Exception e) {
			logError("C="+fname+";D=dumping;", e);
		}
		importMode = false;
	}									// End of dump

	// Dump class body: {...}
	private void dumpClassBody(PrintWriter chan, int flags)
	{
		Trek.println(chan, "{");

		// Do Fields. Use sorted order except for source-only dump.
		// (Can get uncompilable class with sorted order because of
		// forward refs in static initializers)
		Field fld;
		for (int i=0; i<fields.size(); i++) {
			if (flags==Trek.DUMP_STMT)
				fld = (Field)rawFields.elementAt(i);
			else fld = (Field)fields.elementAt(i);
			fld.dump(chan, flags);
		}

		// Do non-simple static inits as code fragment inside static {}.
		// The simple assigns in clinit are marked by "in-line init" cmt
		if (clinit!=null && !isClinitSimple())
		{
			Statement st;
			Trek.println(chan, "    static {");
			for (st=clinit.firstStatement(); st!=null; st=st.next()) {
				if (st.type==Trek.ST_RETURN) continue;
				String src = st.toString();
				if (src.endsWith(" // in-line init")) continue;
				Trek.printPretty(chan, st.toIndent(8, 4, 80), st.toString());
			}
			Trek.println(chan, "    }");
		}
		if (fields.size()>0) Trek.println(chan);

		// Do Methods
		for (Method meth=firstMethod(); meth!=null; meth=meth.next()) {
			meth.dump(chan, flags);
		}

		// Do inner classes. Force all to static so as to handle all outer
		// var refs (incl this) by generating source code of blank final flds
		// set to constructor args (which matches internal code)
		for (TypeRef ref=firstTypeRef(); ref!=null; ref=ref.next()) {
			if (ref.inner!=null)
			{
				// Only do immediately inner classes
				String suffix = ref.toInternalString();
				suffix = suffix.substring(toInternalName().length());
				if (suffix.indexOf('$')!=suffix.lastIndexOf('$'))
					continue;		// At a.b.c while scanning a
				Trek.println(chan, ref.inner.toString());
				ref.inner.dumpClassBody(chan, flags);
			}
		}
		Trek.println(chan, "}");
	}

	// Dump package, imports, class dcl, attr, and pool
	private void dumpHeader(PrintWriter chan, int flags) throws IOException
	{
		// If class in a package (ie. has qualified name), output package stmt
		boolean packimp = false;
		String pkg = toPackage();
		if (pkg.length() > 0)
		{
			packimp = true;
			Trek.println(chan, "package " + pkg + ";");
		}

		// Scan class refs, to identify classes to import
		for (TypeRef ref=firstTypeRef(); ref!=null; ref=ref.next()) {
			String name = ref.toQualifiedName();
			if (name==null) continue;		// Array ref

			// Import each used class -- except those in java.lang (which java
			// imports implicitly), the curr class, & the curr package if any
			int dotidx = name.lastIndexOf('.');
			if (ref.used && ref.inner==null
				&& !isBuiltinClass(ref.toInternalString())
				&& ref.idx!=thisClass
				&& (dotidx<0 || !pkg.equals(name.substring(0, dotidx))))
			{
				packimp = true;
				Trek.println(chan, "import " + name + ";");
			}
		}

		// Class def line. If did package or import's, precede with blank line
		if (packimp) Trek.println(chan);
		Trek.println(chan, toString());

		// Look thru attributes if so requested
		if ((flags & Trek.DUMP_ATTR) != 0)
		{
			Trek.println(chan, "Version " + majorVersion + "." + minorVersion);
			if (fileIdx > 0)
			{
				String sf = constant(fileIdx).toInternalString();
				Trek.println(chan, "Source File: " + sf);
			}
			if (jtrekCreated)
				Trek.println(chan, "JTrek-created Class File");

			// Do any custom attr
			for (int i=0; i<attrs.length; i++) {
				Attr attr = attrs[i];
				if (attr.value != null)
					attr.dump(chan, this, "");
			}
		}

		// Constant pool
		if ((flags & Trek.DUMP_POOL) != 0)
		{
			Trek.println(chan);
			for (int i=1; i<cp.size(); i++) {
				CpInfo cpi = constant(i);
				if (cpi==null) continue;
				Trek.println(chan, i + ": " + cpi.toTag() + cpi.toString());
			}
		}
	}

	// See if file has exactly the same name as exact
	private boolean exactlySameName(File file, String exact)
	{
		try {
			String realname = file.getCanonicalPath();
			return realname.endsWith(exact);
		}
		catch (Throwable th) {
			return true;
		}
	}

	// Return field at specified index
	public Field fieldAt(int fldidx)
	{
		return (Field)fields.elementAt(fldidx);
	}

	// First of things in constant pool
	public ValueRef firstValueRef()
	{
		return new IntegerConstant(this).next();
	}
	public FieldRef firstFieldRef()
	{
		return new FieldRef(this).next();
	}
	public MethodRef firstMethodRef()
	{
		return new MethodRef(this).next();
	}
	public TypeRef firstTypeRef()
	{
		return new TypeRef(this).next();
	}

	// Return 1st field of this class
	public Field firstField()
	{
		Field fld;
		if (fields.size()==0)
			return null;
		else fld = (Field)fields.firstElement();
		return fld;
	}

	// Return 1st method of this class
	public Method firstMethod()
	{
		Method meth;
		if (methods.size()==0)
			return null;
		else meth = (Method)methods.firstElement();
		return meth;
	}

	// Returns # of bytes in class file
	public long getSize() {
		return size;
	}

	// Return classfile of this class's super class, reading it if neces.
	// If cant read superclass, throws exception
	ClassFile getSuperClass()
	{
		// Go from name to object if it exists
		if (superClass==0) return null;
		String str = constant(superClass).toInternalString();
		return parent.getClassFile(str);
	}

	public Trek getTrek()
	{
		return parent;
	}

	// Build hash entries for pool now that it has been fully read
	void hashPool()
	{
		int len = cp.size();
		cpVirginSize = len;
		cphash = new Hashtable(len + len/4);
		for (int i=1; i<len; i++) {
			CpInfo cpi = constant(i);
			if (cpi!=null)
				cphash.put(cpi.keyOf(), cpi);
		}
	}

	// Returns true if object's accessFlags has any of the specified flags set
	public boolean hasModifier(int flags)
	{
		return flags==-1 || (accessFlags&flags)!=0;
	}

	// Returns index of this classfile.
	// Notes:
	//	Doing trek.insertClassFile() changes indexes of classfiles
	public int indexOf()
	{
		return idx;
	}

	// Returns true if This is an ancestor class (possib indirect super) of cf 
	public boolean isAncestor(ClassFile cf)
	{
		for (ClassFile ccf=cf.getSuperClass(); ccf!=null; ccf=ccf.getSuperClass())
		{
			if (ccf == this) return true;
		}
		return false;
	}

	// Returns true if there is arbitrary code to init static vars.
	// meth = a <clinit> method
	boolean isClinitSimple()
	{
		// Analyze clinit if havent yet
		if (clinitStatus==clinit_unknown)
		{
			clinitStatus = clinit_fancy;		// Presume this
outer:
			for (Statement st=clinit.firstStatement(); st!=null; st=st.next()) {
				if (st.type==Trek.ST_RETURN)
				{
					clinitStatus = clinit_simple;
					break;
				}

				// Fancy if non-assignments or temp locals crated
				if (st.type!=Trek.ST_ASSIGN) break;
				Instruction inst;
				for (inst=st.firstInstruction(); inst!=null; inst=inst.next()) {
					if (inst.getLocalIdx()>=0) break outer;
				}
			}
		}
		return clinitStatus==clinit_simple ? true : false;
	}

	// Returns true if data type is builtin class (ie. directly in java.lang)
	private boolean isBuiltinClass(String dt)
	{
		return dt.startsWith("java/lang") && dt.substring(10).indexOf('/')<0;
	}
	
	// Returns true if class is in jdk
	boolean isSystemClass()
	{
		String pkg = toPackage();
		return (pkg.startsWith("java.") || pkg.startsWith("sun."));
	}

	// Return 1st field of this class
	public Field lastField()
	{
		Field fld;
		if (fields.size()==0)
			return null;
		else fld = (Field)fields.lastElement();
		return fld;
	}

	// Return last method of this class
	public Method lastMethod()
	{
		Method meth;
		if (methods.size()==0)
			return null;
		else meth = (Method)methods.lastElement();
		return meth;
	}

	// Output msg line(s) to read/write log file
	// context = info about what error occurred for
	// e = exception that just occurred
	void logError(String context, Exception e)
	{
		hadError = true;
		parent.logError(context, e);
	}

	// Returns matching method if any
	// name = name of method (note: constructor is class name, not <init>)
	// desc = null to match any, a descriptor to match specific ocurrence
	//	of this (overloaded) method name -- either just (arg-types) or
	//	full descriptor, namely (arg-types)ret-type
	Method lookupMethod(String name, String desc)
	{
		// If no name match, ret null. If match and no desc, ret meth.
		Method meth = (Method)Trek.lookup(methods, name);
		if (meth==null || desc==null)
			return meth;

		// Look for arglist match "(xxx)" or arg+ret "(xxx)t"
		boolean full = !desc.endsWith(")");
		while (true) {
			String mdesc = meth.toSignature();
			if (!full) mdesc = mdesc.substring(0, mdesc.indexOf(')')+1);
			if (desc.equals(mdesc))
				return meth;
			meth = meth.next();
			if (meth==null || !name.equals(meth.toName()))
				return null;		// Past instances of "name"
		}
	}

	// Return method at specified index
	public Method methodAt(int methidx)
	{
		return (Method)methods.elementAt(methidx);
	}

	// Navigate to next classfile of program. If at end, ret null instead
	// ispub = true if user is calling next (as opposed to pkg-internal usage)
	public ClassFile next() {return next(true);}
	ClassFile next(boolean ispub)
	{
		if (idx >= parent.cfiles.size()-1)
			return null;
		ClassFile cf = (ClassFile)parent.cfiles.elementAt(idx + 1);
		if (ispub)
			parent.writeVerbose("Scan at " + cf.toQualifiedName());
		return cf;
	}

	// Navigate to previous classfile of program. If at start, ret null instead
	public ClassFile prev()
	{
		if (idx <= 0)
			return null;
		ClassFile cf = (ClassFile)parent.cfiles.elementAt(idx - 1);
		parent.writeVerbose("Scan at " + cf.toQualifiedName());
		return cf;
	}

	// Workaround for jdk1.3 bug. ZipFile object not cached by
	// ZipFileInputStream. So caller needs to make sure it is not gc-ed.
	ZipFile zf;

	// Open the classfile.
	// indirs = ordered list of dir paths this class file may be in
	// classname = internal-fmt (/'s not .'s) fully qualified class name
	// filepath = ditto with File.separatorChar's
	// Returns handle to data stream in opened file, null if this file
	// instance has already been read, or fires exception
	private DataInputStream open(String indirs, String classname, String filepath)
	{
		boolean casemiss = false;
		DataInputStream chan = null;
		filepath += ".class";

		// "file-ize" the class name and scan thru indirs for its dir
		// Find next semi-delimited dir path in indirs
		int cp[] = new int[2];
		while (true) {
			String indir = Trek.nextItem(indirs, cp);
			if (indir==null) break;

			// Do special stuff if curr target dir is core-classes zip file.
			// For perf and to min cases where java 1.0.2 gets NoClassDef for
			// zip classes, continue if dir is jdk & class is not in jdk.
			if (indir.endsWith(".zip") || indir.endsWith(".jar"))
			{
				// Access the class file from zip file
				try {
					//ZipFile zf = new ZipFile(indir);
					zf = new ZipFile(indir);
					ZipEntry ze = zf.getEntry(classname + ".class");
					if (ze==null) continue;

					// Chk if rereading same file
					long last = ze.getTime();
					if (lastWritten>0 && lastWritten==last)
						return null;
					lastWritten = last;

					// Create a stream to read from
					chan = new DataInputStream(zf.getInputStream(ze));
					size = ze.getSize();
				}
				catch (Exception e) {
					continue;					// This one failed, try next
				}
				return chan;
			}									// End if jar or zip

			// Canonicalize dir path (eg. insure ends in separator)
			if (indir.equals(".") || indir.equals(""))
				indir = "";
			else if (!indir.endsWith(File.separator))
				indir += File.separatorChar;

			// Try to open the classfile in candidate indir.
			File file = null;
			try {
				file = new File(indir + filepath);
				chan = new DataInputStream(
					new BufferedInputStream(new FileInputStream(file), 2048));
			}
			catch (FileNotFoundException e) {
				continue;						// This one failed, try next
			}

			// xxx is not the same as Xxx. Unfort this is not enforced at
			// Windows level, even tho it is part of Java definition. So 
			// if JTrek tries to write xxx and Xxx to -op's dir,
			// you only end up with one file. So prep to detect this.
			if (!exactlySameName(file, filepath))
			{
				casemiss = true;
				try {chan.close();}
				catch (IOException ex) {}
				continue;
			}

			// Get size and last-modified date. If same, return null.
			try {
				long last = file.lastModified();
				if (lastWritten>0 && lastWritten==last)
				{
					chan.close();
					return null;
				}
				lastWritten = last;
				size = file.length();
			}
			catch (Exception e) {
			}

			// Return the opened classfile.
			caseMiss = true;
			return chan;
		}									// End path-scan while

		// All failed. Tell world
		throw new TrekException("Could not find "
			+ filepath + " within the specified directories");
	}										// End of open

	// For processing individual classfile in an arbitrary context, such
	// as in context of custom class loader
	// chan = input stream from which to read the classfile
	// trek = trek that should be made parent of read-in classfile
	public static ClassFile read(DataInputStream chan, Trek trek)
	{
		ClassFile cf = new ClassFile(trek.mainTrek);
		cf.readWork(chan, chan.toString());
		return cf;
	}

	// Read the specified class's classfile into memory
	// indirs = ordered list of dir paths this class file may be in
	// classname = internal-fmt (/'s not .'s) fully qualified class name
	void read(String indirs, String classname)
	{
		String filepath = classname.replace('/', File.separatorChar);
		DataInputStream chan = open(indirs, classname, filepath);
		readWork(chan, filepath);
		readClose(chan);
	}

	// Close read stream
	void readClose(DataInputStream chan)
	{
		try {
			chan.close();
		}
		catch (Exception e) {
			logError("C=" + toQualifiedName() + ";D=closing;", e);
		}
	}

	// Read class from specified stream
	// filepath = for error msg's
	void readWork(DataInputStream chan, String filepath)
	{
		// Insure collections valid afterwards even if abort
		methods = fields = new Vector();
		attrs = new Attr[0];
		interfaces = new int[0];

		// Read header fields
		try {
			magic = chan.readInt();
			if (magic != 0xCAFEBABE)
				throw new TrekException(filepath + " does not contain a Java class");
			minorVersion = chan.readShort();
			majorVersion = chan.readShort();
			if (majorVersion > 100)
				throw new TrekException(filepath + "'s version is not supported");

			// Read constant pool and hash its entries
			CpInfo.readPool(chan, this);
			hashPool();

			// Read class properties
			accessFlags = chan.readShort();
			thisClass = chan.readShort();
			TypeRef tref = (TypeRef)constant(thisClass);
			fullname = tref.toQualifiedName();
			superClass = chan.readShort();

			// Force inner class to be static
			// If $ in name (as opposed to package), it is
			String full = toInternalName();
			int namidx = full.lastIndexOf('/') + 1;
			if (full.indexOf('$', namidx) >= 0)
				accessFlags |= Trek.ACC_STATIC;

			// Read names of interfaces that this class directly implements
			int len = chan.readShort();
			interfaces = new int[len];
			for (int i=0; i<len; i++) interfaces[i] = chan.readUnsignedShort();

			// Read content of this class:
			// each field (data member), method, and attribute
			Field.readFields(chan, this);
			Member.bindIndexes(fields);
			Method.readMethods(chan, this);
			Member.bindIndexes(methods);
			attrs = Attr.readAttrs(chan, this, null);
			if (jtrekCreated)
				parent.logMsg("Warning - " + toQualifiedName()
					+ " already instrumented by JTrek");
		}

		// This is not top-level. addClasses or nowhere is top-level
		catch (Exception e) {
			logError("C=" + filepath + ";D=reading;", e);
		}
	}										// End of read

	// Returns whether this class references the specified member object
	public boolean references(Member mem)
	{
		return MemberRef.get(mem, this) != null;
	}

	// Reread classfile if changed & return true, else return false
	public boolean refresh()
	{
		String classname = toInternalName();
		String filepath = classname.replace('/', File.separatorChar);
		DataInputStream chan = open(parent.inDirs, classname, filepath);
		if (chan==null)
			return false;

		// Clear state & reread the data
		cphash = null;
		superClass = thisClass = 0;
		clinit = null;
		clinitStatus = clinit_unknown;
		hadError = false;
		importSetup = false;
		modified = false;
		readWork(chan, filepath);
		readClose(chan);
		return true;
	}

	// Do datatype of each arg of a method
	private void setImportedArgs(String desc, Hashtable namehash)
	{
		int pos = 0;
		while (true) {
			int len = Datum.getDescLength(desc, pos);
			if (len<=0) break;
			String argdesc = desc.substring(pos, pos+len);
			pos += len;
			setImported(argdesc, namehash);
		}
	}

	// Scan names and decide which packages to import, if havent yet
	private void setImportedPackages()
	{
		if (importSetup) return;		// Dont repeat
		importSetup = true;

		// Determine whether classnames can be reffed unqualified
		setImportedPackages(new Hashtable(25), this);
	}

	// Scan curr class and all inner classes
	private void setImportedPackages(Hashtable namehash, ClassFile scan)
	{
		for (TypeRef ref=scan.firstTypeRef(); ref!=null; ref=ref.next()) {
			String name = ref.toInternalString();
			if (name.startsWith("["))
				continue;				// Array ref

			// If name indics it is inner class, scan inner class's imports.
			// Else scan curr class's content for imports
			if (name.startsWith(scan.toInternalName() + "$"))
			{
				ref.used = true;
				ref.inner = (ClassFile)ref.getObject();
				setImportedPackages(namehash, ref.inner);
			}
			else {
				TypeRef tref = scan==this ? ref : TypeRef.addTemp(name, this);
				setImported(tref, namehash);
			}
		}									// End loop

		// If only ref to a class is data type of arg or ret val,
		// no class ref is in the const pool.
		// So scan method defs and method refs
		for (Method meth=scan.firstMethod(); meth!=null; meth=meth.next()) {
			String desc = meth.toDesc();
			setImported(desc, namehash);
			desc = meth.toArgsDesc();
			setImportedArgs(desc, namehash);
		}
		for (MethodRef mref=scan.firstMethodRef(); mref!=null; mref=mref.next()) {
			String desc = mref.toDesc();
			setImported(desc, namehash);
			desc = mref.toArgsDesc();
			setImportedArgs(desc, namehash);
		}

		// Similarly chk fields of a class and fields it refs
		for (Field fld=scan.firstField(); fld!=null; fld=fld.next()) {
			String desc = fld.toDesc();
			setImported(desc, namehash);
		}
		for (FieldRef fref=scan.firstFieldRef(); fref!=null; fref=fref.next()) {
			String desc = fref.toDesc();
			setImported(desc, namehash);
		}
	}

	// Get to cp entry via desc (ie. class) of a variable/method.
	// Add class to pool as temp marker if no such class in pool
	private void setImported(String desc, Hashtable namehash)
	{
		String name = Datum.getDescClass(desc);
		if (name==null || name.startsWith("["))
			return;			// Array not a class ref
		TypeRef tref = TypeRef.addTemp(name, this);
		setImported(tref, namehash);
	}

	// Decide if this class should be reffed as if imported (ie. unqualified)
	// newcpi = pool entry representing the class
	// namehash = hashtable of unqualified class names
	private void setImported(TypeRef newcpi, Hashtable namehash)
	{
		// Get simple name from fully qualified internal name
		String qclass = newcpi.toInternalString();
		int slix = qclass.lastIndexOf('/');
		String name = (slix<0) ? qclass : qclass.substring(slix+1);
		TypeRef oldcpi = (TypeRef)namehash.get(name);

		// If same entry as before, no work to do.
		// If entry already there and beats new class, no work to do.
		// 2nd class wins if it is current class or a jdk (java.*) class
		if (oldcpi==newcpi)
			return;
		if (oldcpi!=null
			&& newcpi.idx!=newcpi.parent.thisClass
			&& !qclass.startsWith("java/"))
			return;

		// New class 1st or beat out old class. If latter, unhash old class.
		// Either way, insert new entry
		if (oldcpi!=null)
		{
			oldcpi.used = false;	// Old one must be reffed with qual name
			namehash.remove(name);
		}
		newcpi.used = true;		// Mark that refs can be unqual class name
		namehash.put(name, newcpi);
	}

    // Return source declaration for specified name and data type info
	// name = name of a variable, or null if processing desc only
	// desc = encoded type info for a name
	// mode = 0 says just name (!importMode) or import-relative name
	//        1 says qualified external name if type is class (ie. L)
	String toDef(String name, String desc, int mode)
	{
		// Determ # of dimensions
		String dt = "";			// Data type
		int dims = 0;			// Array dimensions encountered
		int idx;				// Curr index into desc
		String pkgout[] = null;	// For qualified class ref
		for (idx=0; desc.charAt(idx)=='['; idx++) dims += 1;

		// Convert the type-letter to a source data type
		switch (desc.charAt(idx++)) {
		case 'B':
			dt = "byte";
			break;
		case 'C':
			dt = "char";
			break;
		case 'D':
			dt = "double";
			break;
		case 'F':
			dt = "float";
			break;
		case 'I':
			dt = "int";
			break;
		case 'J':
			dt = "long";
			break;
		case 'V':
			dt = "void";
			break;
		case 'S':
			dt = "short";
			break;
		case 'Z':
			dt = "boolean";
			break;

		// The complicated case, the type is a class. Nominally output
		// all chars between the L and the terminating ;.
		case 'L':
			int semipos = desc.indexOf(';', idx); 
			if (semipos<idx+1)
				throw new TrekException("Bad data-type info (" + desc + ")");
			desc = desc.substring(idx, semipos);

			// If caller explicitly wants qualified name, just do it.
			if (mode!=0)
				pkgout = new String[1];	// Want qualified name

			// Else if not in source-code context or name is in java.lang
			// (whose members dont need to be imported), just use unqual name.
			else if (!importMode || isBuiltinClass(desc))
				;						// Want unqualified name

			// If generating source code, model is that names are in
			// context of the imports that were done. If class known and
			// import was done, use unqualified name.
			else {
				TypeRef tref = TypeRef.get(desc, this);
				if (tref==null || !tref.used)
					pkgout = new String[1];	// Want qualified name
			}
			dt = toNamePieces(desc, true, pkgout);
			break;
		default:
			throw new TrekException("Bad data-type info (" + desc + ")");
		}

		// Concat external data type (qualified if desired)
		StringBuffer src = new StringBuffer(32);
		if (pkgout!=null)
			src.append(pkgout[0]);
		src.append(dt);

		// Concat var being declared if any (none if method's ret type)
		if (name!=null)
		{
			src.append(" ");
			src.append(name);
		}
		for (int i=0; i<dims; i++) src.append("[]");
		return src.toString();
    }								// End of toDef

	// Returns source form of possibly inner-class name. 
	// intname = internal qualified name of possibly inner class
	// sep = separator between outer & inner pieces of name
	// both = return outer.tail if true else just tail
	// apkg = non-null says return pkg of name here
	// Notes:
	//	If inner class, map raw tail of outer$inner to inner or Anony_inner
	static String toNamePieces(String intname, boolean both, String apkg[])
	{
		// Isolate simple name, and pkg if desired
		// Note: No / leads to -1 + 1 = 0, which sets name to whole string
		int namepos = intname.lastIndexOf('/') + 1;
		String name = intname.substring(namepos);
		if (apkg!=null)
		{
			if (namepos==0)			// Class in a pkg?
				apkg[0] = "";		// No
			else apkg[0] = intname.substring(0, namepos).replace('/', '.');
		}

		// If not inner class, name is already right
		int sep = '$';
		int sepidx = name.indexOf(sep);
		if (sepidx<0)
			return name;

		// Handle anonymous inner class special case
		String tail = name.substring(sepidx+1);
		if (!Character.isJavaIdentifierStart(tail.charAt(0)))
			tail = "Anony_" + tail;

		// Do outer.tail or tail, and handle multiple nesting levels
		if (both)
			tail = name.substring(0, sepidx) + "." + tail;
		return toNamePieces(tail, both, null);
	}

	// Returns interfaces of class as array of fully qualified names
	public String[] toInterfaces()
	{
		String inh[] = new String[interfaces.length];
		for (int i=0; i<interfaces.length; i++) {
			TypeRef tref = (TypeRef)constant(interfaces[i]);
			inh[i] = tref.toQualifiedName();
		}
		return inh;
	}

	// Return the unqualified external name of the class
	public String toName()
	{
		return toNamePieces(toInternalName(), false, null);
	}

	// Return string that uniquely identifies this classfile
	public String toObjectId()
	{
		return toQualifiedName();
	}

	// Return package of this class (so may be 0-length string)
	// int ch = char to map .'s to before returning pkg name
	public String toPackage()
	{
		return toPackage('.');
	}
	public String toPackage(int ch)
	{
		// Package is the path of a class, by def
		String intname = toInternalName();
		int postpos = intname.lastIndexOf('/');
		if (postpos<=0) return "";				// Not in package
		String rawpkg = intname.substring(0, postpos);

		// Make return string have right separator char
		if (ch=='/')
			return rawpkg;
		else return rawpkg.replace('/', (char)ch);
	}

	// Return qualified name of class
	public String toQualifiedName()
	{
		return fullname;
	}
	String toInternalName()
	{
		return constant(thisClass).toInternalString();
	}

	// Return string representation of class (= src dcl)
	public String toString()
	{
		// Build list of inherited interfaces
		String inherit = "";
		for (int i=0; i<interfaces.length; i++) {
			if (i == 0)
				inherit = " implements ";
			else inherit += ", ";
			inherit += constant(interfaces[i]).toString();
		}

		// Do interface vs. class processing (interface is an access modifier
		// is why classkywd is set to "" for interfaces). Interfaces
		// can only be extended by interfaces, but unfort compiler does
		// set superClass to java/lang/Object for unextended interfaces
		String classkywd = hasModifier(Trek.ACC_INTERFACE) ? "" : "class ";
		String extend = "";
		if (superClass!=0)
		{
			if (!hasModifier(Trek.ACC_INTERFACE)
				|| !toSuperClass().equals("java.lang.Object"))
			{
				String sup = constant(superClass).toInternalString();

				// If superclass and new class are both in same class, 
				// use unnested form of name in extends clause
				int dollidx = sup.indexOf('$');
				String supfx = sup.substring(0, dollidx+1);
				boolean nested = !toInternalName().startsWith(supfx);
				extend = " extends " + toNamePieces(sup, nested, null);
			}
		}

		// Return all pieces concat
		return Datum.toAccessString(accessFlags, this) + 
			classkywd + toName() + extend + inherit;
	}

	// Return qualified name of superclass
	public String toSuperClass()
	{
		if (superClass==0) return null;
		TypeRef tref = (TypeRef)constant(superClass);
		return tref.toQualifiedName();
	}

	// Return simple name of super class. Provided because super's classfile
	// does not itself have to be available
	private String toSuperClassName()
	{
		if (superClass==0) return "";
		return constant(superClass).toName();
	}

	// Write the class file out if it has changed
	void write(String outdir) throws IOException
	{
		// Unmodified (or had error and not forcing write), no work to do
		if ((Trek.debug&Trek.Db_write)==0)
		{
			if (!modified) return;
			if (hadError)
			{
				parent.logMsg("Warning - " + toQualifiedName()
					+ " not saved because an error was logged for it");
				modified = false;
				return;
			}
		}

		// If have already written case-ignored same file, dont write this
		String outfile = toInternalName().replace('/', File.separatorChar)
			+ ".class";
		String path = outdir + outfile;
		if (caseMiss)
		{
			File file = new File(path);
			if (file.exists() && !exactlySameName(file, outfile))
			{
				parent.logError(path + " not saved because there is "
					+ "file there with same name, except for its case");
				modified = false;
				return;
			}
		}

		// Tell user writing classfile if so requested
		parent.writeVerbose("Writing " + toQualifiedName());
		parent.wrote += 1;

		// Insure dirs of class's dir path actually exist under outdir.
		// Create file to write updated class to
		int sepidx = path.lastIndexOf(File.separatorChar);
		if (sepidx >= 0)
			new File(path.substring(0, sepidx)).mkdirs();
		DataOutputStream chan = new DataOutputStream(
			new BufferedOutputStream(new FileOutputStream(path), 2048));

		// Output and close the classfile
		write(chan);
		modified = false;				// Because disk = memory now
		chan.close();
	}									// End of write

	// Write the classfile to specified stream and clear modified flag.
	public void write(DataOutputStream chan, boolean clean)
	{
		try {
			write(chan);
			modified = false;
			if (clean) clean();		// Remove mods if so requested
		}
		catch (Exception e) {
			logError("C=" + chan.toString() + ";D=writing;", e);
		}
	}

	// Write the classfile to specified stream and clear modified flag.
	private void write(DataOutputStream chan) throws IOException
	{
		// Add JTrekCreated attr skeleton (if not already there)
		if (!jtrekCreated)
		{
			Attr newatts[] = new Attr[attrs.length+1];
			System.arraycopy(attrs, 0, newatts, 0, attrs.length);
			StringStorage ss =  StringStorage.add("com.compaq.JTrekCreated", this);
			newatts[attrs.length] = new Attr(ss.idx);
			attrs = newatts;
		}

		// Write the hdr
		chan.writeInt(magic);
		chan.writeShort(minorVersion);
		chan.writeShort(majorVersion);

		// Write constant pool and class props
		CpInfo.writePool(chan, this, cp.size());
		chan.writeShort(accessFlags);
		chan.writeShort(thisClass);
		chan.writeShort(superClass);

		// Write list of direct interfaces
		if (interfaces == null)
			chan.writeShort(0);				// Class implements no interfaces
		else {
			chan.writeShort(interfaces.length);
			for (int i=0; i<interfaces.length; i++) {
				chan.writeShort(interfaces[i]);
			}
		}

		// Write content of this class:
		// each field (data member), method, and attribute
		Field.writeFields(chan, this);
		Method.writeMethods(chan, this);
		Attr.writeAttrs(chan, attrs, this, null);
	}

}

