//
// THIS SOFTWARE IS FURNISHED UNDER A LICENSE AND  MAY
// BE USED OR COPIED ONLY IN ACCORDANCE WITH THE TERMS
// OF SUCH LICENSE.
//
//
// COPYRIGHT (c) 1997 BY DIGITAL EQUIPMENT CORPORATION.
// ALL RIGHTS RESERVED.
//

package dec.trek;
import java.io.*;

// Super class for all types of constant pool (cp) entries as one class.
// Contains abstract methods for all instance methods, and shared static methods.

public abstract class CpInfo
{
	protected ClassFile parent;		// The classfile this constant belongs to
	int idx;						// Its constant-pool index
	public Object userDefined;		// User-owned field

    // The various types.
	protected static final int Cp_rawstring = 1;	// A name or a String value
	protected static final int Cp_integer = 3;		// Scalar
	protected static final int Cp_float = 4;		// Scalar
	protected static final int Cp_long = 5;			// Scalar
	protected static final int Cp_double = 6;		// Scalar

	// If need to do type-based scan for any of the following types of const
	// outside this file, elim the protected keyword for that type's symbol.
	protected static final int Cp_typeref = 7;		// Entity
	protected static final int Cp_string = 8;		// Scalar (but indirect value)
	protected static final int Cp_fieldref = 9;		// Member ref
	protected static final int Cp_methodref = 10;	// Member ref
	protected static final int Cp_intmethodref = 11;// Member ref
	protected static final int Cp_nameandtype = 12;	// Entity

	// Construct a start-scan cp entry
	CpInfo(ClassFile cf)
	{
		parent = cf;
	}

	// Returns true if specified object has same type/value(s) as "this".
/*
	boolean equals(CpInfo cpi)
	{
		return isSameType(cpi) && cpi.toInternalString().equals(toInternalString());
	}
*/

	// The common work method to scan pool for entry matching data in cpi
	// key = value to lookup in pool
	// cf == classfile whose pool is to be scanned
	static CpInfo get(Object key, ClassFile cf)
	{
		return (CpInfo)cf.cphash.get(key);
	}

    // Return count of args of a method ref
	int getArgCount()
	{
		throw new TrekException("Pool entry " + idx + " is not a method ref");
	}

	// Returns true if is class constant created just to mark imported class
	protected boolean isTempItem()
	{
		return false;
	}
	
	// Returns non-0 if reffed method has no return value,
	// but -1 for constructor and 1 otherwise
	int isVoidCall()
	{
		throw new TrekException("Pool entry " + idx + " is not a method ref");
	}

	// Returns true if arg is same class as this
	abstract protected boolean isSameType(CpInfo cpi);

	// Returns unique key for specified object
	abstract Object keyOf();

	// Scan for next entry of specified type of constant
	// Returns matching constant or null
	protected CpInfo nextOf()
	{
		// Continue with next constant
		int start = idx + 1;
		if (start <= 0) start = 1;

		// Scan for entry that matches requested type of entry
		int len = parent.constants();
		for (int i=start; i<len; i++) {
			CpInfo cpi = parent.constant(i);
			if (cpi == null) continue;

			// Match relevant category only
			if (this instanceof ValueRef)
			{
				if (cpi instanceof ValueRef)
					return cpi;
				else continue;
			}
			if (isSameType(cpi))
				return cpi;
		}
		return null;
	}

	// Read entire constant pool
    static void readPool(DataInput chan, ClassFile cf) throws IOException
	{
		// Get pool size, which is 1 gtr than # of constants in pool
		// Create and read/fill-in each entry in the pool
		// Read type of constant and dispatch to fill in rest
		int len = chan.readUnsignedShort();
		cf.addPool(len);
		while (cf.cpSize() < len) {
			int tag = chan.readUnsignedByte();
			switch (tag) {

			// Read an entity
			case Cp_typeref:
				new TypeRef(chan.readUnsignedShort(), cf);
				break;
			case Cp_nameandtype:
				new NameAndType(chan.readUnsignedShort(), chan.readUnsignedShort(), cf);
				break;

			// Read a member ref
			case Cp_fieldref:
				new FieldRef(chan.readUnsignedShort(), chan.readUnsignedShort(), cf);
				break;
			case Cp_methodref:
				new MethodRef(false,
					chan.readUnsignedShort(), chan.readUnsignedShort(), cf);
				break;
			case Cp_intmethodref:
				new MethodRef(true,
					chan.readUnsignedShort(), chan.readUnsignedShort(), cf);
				break;

			// Read a string ref
			case Cp_string:
				new StringConstant(chan.readUnsignedShort(), cf);
				break;

			// Read a scalar constant
			case Cp_double:
				new DoubleConstant(chan.readDouble(), cf);
				break;
			case Cp_float:
				new FloatConstant(chan.readFloat(), cf);
				break;
			case Cp_integer:
				new IntegerConstant(chan.readInt(), cf);
				break;
			case Cp_long:
				new LongConstant(chan.readLong(), cf);
				break;

			// Read actual char data (utf is a packed repr of unicode wherein
			// ascii chars each take one byte)
			case Cp_rawstring:
				new StringStorage(chan.readUTF(), cf);
				break;
			default:
				throw new TrekException("Bad tag in constant pool");
			}
		}								// End of loop
	}

    // Return internal class name of a class or member ref
	String toClassName()
	{
		throw new TrekException("Pool entry " + idx + " is not a member reference");
	}

	// Returns descriptor or signature of pool entry.
	String toDesc()
	{
		throw new TrekException("Pool entry " + idx + " does not have a descriptor");
	}
	String toSignature()
	{
		throw new TrekException("Pool entry " + idx + " does not have a descriptor");
	}

    // Return internal class name of a class or member ref
	String toInternalClassName()
	{
		throw new TrekException("Pool entry " + idx + " is not a member reference");
	}

	// Return string(s) associated with specified constant pool entry.
	// Differs from toString in that member refs contain raw desc and /'s
	// rather than source-code signature and .'s. Also entries that dont
	// involve strings (eg. Cp_long) simply return null string.
	String toInternalString()
	{
		return "";
	}

    // Returns simple name of this entry.
	String toName()
	{
		return toString();
	}

	// Returns ref expr for this entry. Usually this is the simple name,
	// but it can be qualified name for a constructor or class ref.
	String toRef()
	{
		return toString();
	}

    // Return ref to static name. If ref to/from same class, just return name.
	// Else return the name's class, qualified as neces
	String toStaticRef()
	{
		throw new TrekException("Pool entry " + idx + "cannot have static name");
	}

	// Returns tag for dump report
	abstract protected String toTag();

	// Write a pool entry
	abstract protected void write(DataOutput chan) throws IOException;

	// Write all the entries of the pool out
	static void writePool(DataOutput chan, ClassFile cf, int len)
		throws IOException
	{
		chan.writeShort(len);

		// Write out each entry in the pool. Can be gaps in pool because of
		// long's/double's. Class/longVal chk is for class-ref's created
		// when only usage of class is in arg list.
		for (int i=1; i<len; i++) {
			CpInfo cpi = cf.constant(i);
			if (cpi==null || cpi.isTempItem()) continue;

			// Acct for 2-entry long/double hack in constant pool layout
			cpi.write(chan);
			if (cpi instanceof DoubleConstant || cpi instanceof LongConstant)
				i += 1;
		}
	}

}

