package dec.trek;
import java.util.Vector;

//
//
// Datum
//
//
final class Datum implements Cloneable
{
	String arrayInit[];				// For ArrayInit only
	String desc;					// Descriptor of this datum
	String dest;					// If opr is assign, raw target of assign
	Vector condExpr;				// For decomp of test in cond expr
	int intVal;						// For integer constants
	Local loc;						// Set if datum uses local
	int opr;						// This datum's top-level opr
	String src;						// Source of this datum (eg. a name)
	int soft;						// Datum's softness

	// Types of firmness and softness
	static final int firm = 0;	// Normal firm case
	static final int softy = 1;	// Desc completely unknown

	// Object is most broad, final classes least broad.
	static final int softobj = 2;	// Type set from unconstrained source
	static final int atmost = 3;	// Object's type at most this broad
	static final int atleast = 4;	// Object's type at least this broad

	// Minimum integer length
	static final int byplus = 8;	// Is int, and byte or longer
	static final int chplus = 15;	// Char or longer
	static final int shplus = 16;	// Short or longer
	static final int intonly = 32;	// Must be int

	// Construct a datum
	// sr = source code
	// dsc = descriptor
	// sof = softness
	// op = opr that combined terms to lead to this datum
	Datum(String sr, String dsc)
	{
		this(sr, dsc, Datum.firm, Term.op_atomic);
	}
	Datum(String sr, String dsc, int sof)
	{
		this(sr, dsc, sof, Term.op_atomic);
	}
	Datum(String sr, String dsc, int sof, int op)
	{
		src = sr;
		desc = dsc;
		soft = sof;
		opr = op;
	}

	// Construct a datum from a local
	Datum(Local loc)
	{
		this(loc.name, loc.desc, loc.soft, Term.op_atomic);
		this.loc = loc;
	}

	// Return a copy of the specified datum
	Datum copy()
	{
		try {
			return (Datum)clone();
		}
		catch (Exception e) {
			return null;
		}
	}

	// Returns # of words in datum with specified desc
	static int getDataSize(String desc)
	{
		// If empty, ret 0
		if (desc.length()==0)
			return 0;

		// Chk 1st letter of desc
		int ch0 = desc.charAt(0);
		switch (ch0) {
		case 'V':
			return 0;
		case 'D':
		case 'J':
			return 2;
		default:
			return 1;
		}
	}

	// If desc represents a class, return internal-fmt form
	// of the class (ie. strip L ; and any ['s)
	// dsc = descriptor to process
	static String getDescClass(String dsc)
	{
		// Strip any array info
		int bridx = dsc.lastIndexOf('[');
		if (bridx >= 0) dsc = dsc.substring(bridx+1);

		// Is variable's type a class or primitive?
		if (dsc.startsWith("L"))
			return dsc.substring(1, dsc.length()-1);
		else return null;
	}

    // Parses the descriptor in desc starting at pos, returning its length
	// dsc = encoded data type info -- potentially for mult variables
	// pos = where next descriptor starts within dsc
	static int getDescLength(String dsc, int pos)
	{
		// Chk if at end of dsc
		if (dsc==null || pos >= dsc.length())
			return 0;

		// Eat array dimensions
		int idx;				// Curr index into dsc
		for (idx=pos; dsc.charAt(idx)=='['; idx++) ;

		// Eat the type letter (& assoc data for L)
		switch (dsc.charAt(idx++)) {
		case 'B':
		case 'C':
		case 'D':
		case 'F':
		case 'I':
		case 'J':
		case 'V':
		case 'S':
		case 'Z':
			break;

		// The complicated case, the type is a class. Nominally output
		// all chars between the L and the terminating ;.
		case 'L':
			idx = dsc.indexOf(';', idx) + 1;
			if (idx > 0) break;
		default:
			throw new TrekException("Bad data-type info (" + dsc + ")");
		}
		return idx - pos;
    }								// End of getDescLength

	// Returns true if this is non-constant datum whose data type is soft.
	// Except for dummy datums, this means datum is assoc with local.
	// A local is soft if it has not been used in a context that exactly
	// defines its type. (Except for integers <= 32-bits, booleans,
	// and objects, all usages provide enough context).
	boolean isSoft()
	{
		return soft>=Datum.softy && opr<=Term.op_atomic;
	}

	// Promote a data type to what expressions of that type are computed in
	// or internally represented in
	static String promote(String desc)
	{
		// Normalize integer data types (U is integer constant)
		if (desc.length()==1)
		{
			int ch1 = desc.charAt(0);
			if (ch1=='C' || ch1=='B' || ch1=='S' || ch1=='U')
				desc = "I";
		}
		return desc;
	}

	// Insure datum (plus its local) is as firm as possible
	// soft = softness which may be more constrained than dtm's curr value
	void refineSoftness(int sof)
	{
		if (soft<=Datum.firm)
			return;

		// Set dtm's soft if calc-ed value further constrains it.
		// Propagate calc-ed value to local if result assoc with local
		if (sof<=Datum.firm || (sof>soft
				&& (sof!=Datum.atleast || soft!=Datum.atmost)))
		{
			soft = sof;
			if (loc!=null)
				loc.soft = sof;
		}
	}

	// Returns index of right-char, accounting for quoted substrings (and
	// nested parenthesis if left is 0) -- or -1 if a right-char not found
	// str = string being searched.
	// nest = -1 when no nested-pairs considered
	//			0 when nested parentheses skipped
	// right = set of possible chars to match on
	// start = where to start search within str
	static int smartIndexOf(String str, int nest, String rchars, int start)
	{
		int endpos = start;
		int lev = 0;
		int quote = -1;
		int len = str.length();

		// Look for rchars in str, accting for quoted strings
		while (true) {
			if (endpos>=len) return -1;

			// Skip chars after backslash within quoted string
			int ch = str.charAt(endpos++);
			if (ch=='\\' && quote!=-1)
			{
				endpos += 1;
				continue;
			}

			// Ignore all but end-quote in quoted string
			if (ch==quote)
				quote = -1;			// End quoted string

			// Else if at top-level and have match, then done.
			else if (quote<0)
			{
				if (ch=='\'' || ch=='"')
					quote = ch;		// Start quoted string
				else if (lev==0 && rchars.indexOf(ch)>=0)
					break;

				// Also take care of nested parens if nest is 0
				else if (nest==0)
				{
					if (ch==')' && lev>0)
						lev -= 1;
					else if (ch=='(')
						lev += 1;
				}
			}
		}							// End loop

		// Return pos of just matched terminator (-1 to undo the ++)
		return endpos - 1;
	}

    // Reconstruct source-fmt of name's access flags.
	static String toAccessString(int acc, Object obj) {
		StringBuffer sb = new StringBuffer(32);

		if ((acc & Trek.ACC_PUBLIC) != 0) sb.append("public ");
		if ((acc & Trek.ACC_PRIVATE) != 0) sb.append("private ");
		if ((acc & Trek.ACC_PROTECTED) != 0) sb.append("protected ");

		if ((acc & Trek.ACC_STATIC) != 0) sb.append("static ");
		if ((acc & Trek.ACC_FINAL) != 0) sb.append("final ");
		if ((acc & Trek.ACC_VOLATILE) != 0) sb.append("volatile ");
		//if ((acc & Trek.ACC_THREADSAFE) != 0) sb.append("threadsafe ");
		if ((acc & Trek.ACC_TRANSIENT) != 0) sb.append("transient ");
		if ((acc & Trek.ACC_NATIVE) != 0) sb.append("native ");
		if ((acc & Trek.ACC_STRICT) != 0) sb.append("strictfp ");

		// Interface implies abstract
		if ((acc & Trek.ACC_INTERFACE) != 0)
			sb.append("interface ");
		else if ((acc & Trek.ACC_ABSTRACT) != 0)
			sb.append("abstract ");

		// Corresponds to acc_super for classes
		if (!(obj instanceof ClassFile) && (acc & Trek.ACC_SYNCHRONIZED) != 0)
			sb.append("synchronized ");
		return sb.toString();
    }

	// Debugging aid
	public String toString()
	{
		return dest + "=" + src + "(" + desc + "," + soft + ") [" + opr + "] "
			+ (loc!=null ? loc.name : "No local");
	}

}

