//
// THIS SOFTWARE IS FURNISHED UNDER A LICENSE AND  MAY
// BE USED OR COPIED ONLY IN ACCORDANCE WITH THE TERMS
// OF SUCH LICENSE.
//
//
// COPYRIGHT (c) 1997 BY DIGITAL EQUIPMENT CORPORATION.
// ALL RIGHTS RESERVED.
//

package dec.trek;
import java.util.Vector;

//
//
// Divide a method into statements
//
//
final class Divide 
{
	private Block blk;					// Block at current level
	private int condRight;				// Pc of C in overall a?b:c if non-0
	private Instruction currInst;		// Instruction curr eating
	private int stackSize;				// Num words curr on stack
	private Statement ifLoan;			// Local-only stmt for inlined if/then/else
	private int instEnd = 0;			// End of curr inst
	private int level;					// Top-level code is level 1
	private Block levs[] = new Block[100];
	private boolean maybeLoan;			// True if extra load pending
	private Method meth;				// Its parent method
	private int methLen;				// # of bytes in method's code
	private ExceptionHandler nextTry;	// Next try block if any
	private int nextTryPost;			// Loc past its catches if non-0
	private Statement prim;				// Primord stmt being exploded
	private Statement saveLoan;			// Stmt that did a "confirmed" loan
	private int stmtStart = 0;			// Start of curr stmt in insts

	// Bump level, and create, fill-in, and return blk for that level
	// type = stmt code of end-stmt of stmt that ends new blk
	// end = pc where blk ends
	// endstr = source code for blk-end stmt, or null
	// top = stmt that starts this blk
	// Notes:
	//  The level i blk describes the blk encompassing stmts whose level is i.
	private void addBlock(int type, int end, String endstr, Statement top)
	{
		// Access blk at higher level, creating if necessary
		blk = levs[++level];
		if (blk == null)
		{
			blk = new Block();
			levs[level] = blk;
		}

		// Set contents of blk and return it
		blk.type = type;
		blk.bottom = blk.end = end;
		blk.endstr = endstr;
		blk.top = top;
		blk.postHandlers = 0;
		blk.lastTry = null;
		blk.loc = null;
		blk.endAsap = false;
		return;
	}

	// Add a case label and move to next case if any
	// Notes:
	//	switch & cases all share the same blk
	private void addCase()
	{
		// It's possible for case target to be optimized continue. If so,
		// insert codeless goto into case.
		if (blk.end <= blk.top.start)
		{
			Statement st = addEmptyStatement(Trek.ST_UNKNOWN, level+1, null);
			st.addr = blk.end;
		}

		// Get info for next case to set end of curr case.
		if (blk.casidx < blk.cases.size())
		{
			Case cas = (Case)blk.cases.elementAt(blk.casidx++);
			if (levs[level].bottom>=cas.pc)
			{
				blk.bottom = blk.end = cas.pc;
				blk.endstr = cas.toString();
			}

			// But if outer blk ends in middle of case, short-circuit rest
			// of switch. This can occur in optimized code that looks like:
			// "lp: for () {switch () {case a: i=1; break; case b: break lp;}}"
			// That is, compiler set case-b's pc is after For-loop rather
			// than generating a case-b that contained only "goto past-lp".
			// Conversely make the "straddle case" ends at bottom of outer
			// block, rather than at case-b (ie. the switch's next case).
			// See addEndStmt/addSwitchTail for how case-b's are handled.
			else {
				blk.type = Trek.ST_ENDSWITCH;
				blk.endstr = "} // End switch";
				blk.bottom = blk.end = levs[level].bottom;
			}
		}

		// No more cases, get scan ready to end overall switch stmt
		else {
			blk.type = Trek.ST_ENDSWITCH;
			blk.endstr = "} // End switch";
			blk.cases = null;			// Done with cases vector
		}
		level += 1;						// "Re-enter" switch body's level
	}									// End addCase

	// Add a catch block, terminating preceding catch if any
	// inst = the store (or pop) at start of catch
	// exc = exception handler blk for this catch
	// Returns:
	//	Updates instEnd if black-magic jsr code also at start of blk
	private int addCatchStart(Instruction inst, ExceptionHandler exc)
	{
		// The store/pop "improperly" decr stackSize since the excep-val was
		// placed on stack by the excep, not some load within this method.
		stackSize = 0;

		// Close open inner code blks, inner catches, & the open try/catch blk. 
		// Note: open inner blks include those whose endpt was set to
		// postHandlers (like the catch itself) -- eg. because if-goto was opt.
		if (withinBlock(-1)>=0)
		{							// Usual except for last Catch bef Finally
			for (int i=level-1; i>=1; i--) {
				addEndStatement(stmtStart);

				// Disting open try/catch at new catch's level from an open 
				// inner try/catch by looking for t/c chain whose start pt 
				// matches new catch's start and whose Try ends at or before it
				ExceptionHandler tryexc = levs[i].lastTry;
				if (tryexc==null)
					continue;			// At inner code blk
				if (tryexc.startPc==exc.startPc && tryexc.endPc<=exc.handlerPc)
					break;				// At new catch's level
			}
		}

		// Build stmt string, depending on catch vs finally. 
		Statement top;
		if (exc.catchType == 0)
		{
			// The jsr inst is the magic done to access Finally logic without
			// a new stack frame. It jumps to the store of the return addr
			// left on the stack by the jsr. The inst after that starts the
			// Finally body, so Finally stmt ends at target_inst + its size.
			Instruction in2 = Instruction.at(inst, instEnd);
			if (in2.opcode==Code.jsr)
			{
				instEnd = in2.getTarget();
				in2 = Instruction.at(in2, instEnd);
				instEnd += in2.size;
			}

			// Add Finally stmt and opens its blk. Set to outer blk's bottom
			// since blk.postHandlers can be set to after Catch's & bef Finish.
			// (Note: unless Finally ends in Return stmt, ret inst will 
			// explicitly mark the end of the Finally).
			top = addStatement(Trek.ST_FINALLY, "finally {");
			addBlock(Trek.ST_ENDFINALLY, blk.bottom, "} // End finally", top);
		}

		// A catch. Get type and name for "catch arg"
		else {
			CpInfo cptype = meth.parent.constant(exc.catchType);

			// Add stmt, blk, and possibly lvt entry for Catch
			top = addStatement(Trek.ST_CATCH, null);
			addBlock(Trek.ST_ENDCATCH, blk.postHandlers, "} // End catch", top);
			String typestr = cptype.toRef();
			String namstr = toCatchArg(inst, cptype);
			top.string = "catch (" + typestr + " " + namstr + ") {";
		}
		return instEnd;
	}										// End addCatchStart

	// Build loop-related stmt, or add at least start of If-then
	// ifend = where cond stmt ends
	// ifjump = where cond stmt branches to
	// Returns:
	//	where test-stmt ends
	private int addCondStatement(int ifend, int ifjump)
	{
		instEnd = ifend;
		if (ifjump < instEnd)
			addLoopStuff(ifjump, Vmi.it_if);
		else addIfStuff(ifjump);
		return ifend;
	}									// End addCondStatement

	// Add a stmt that has no code
	// type = type of this statement
	// lev = level at which to add it
	// str = the source code to assign to this stmt
	private Statement addEmptyStatement(int type, int lev, String str)
	{
		return new Statement(meth, stmtStart, type, lev, str);
	}

	// End the current block and enter next block, which is usually the
	// outer block, but is the next Case in a Switch stmt
	// end = where code for this blk ends
	private void addEndStatement(int end)
	{
		// If ended switch "prematurely" because have cases that are just
		// goto's to past end of switch, process this tail now.
		level -= 1;					// End stmt is out a level
		if (blk.type==Trek.ST_ENDSWITCH && blk.cases!=null)
			addSwitchTail((Case)blk.cases.elementAt(blk.casidx-1));

		// Add end-stmt, and link top and bottom stmts
		Statement bot;
		if (stmtStart!=end)
			bot = addStatement(blk.type, blk.endstr);
		else bot = addEmptyStatement(blk.type, level, blk.endstr);
		blk.top.mate = bot;
		bot.mate = blk.top;

		// If created lvt entry for catch blk, set its len again in case catch
		// ended sooner than blk.end predicted (eg. when next Catch start seen)
		// Also set defStmt to eob (see Lvt.buildLvt)
		if (blk.loc!=null)
		{
			blk.loc.length = end - blk.loc.startPc;
			blk.loc.defStmt = bot;
		}

		// If if-branch is to jump-less loop bottom or past a loop bottom,
		// then it is really a Break out of the loop. So convert the Then 
		// blk into code-less Break stmt (But if ending code-less blk 
		// in 1st place (then.start==endthen.start), do nothing)
		if (blk.type==Trek.ST_ENDTHEN && blk.top.start<bot.start)
		{
			int same = blk.end;
			for (int i=level; i>1; i--) {
				Block outer = levs[i];
				if (outer.type>=Trek.ST_ENDLOOP)
				{
					if ((same>0 && outer.endAsap) || blk.end>=outer.end)
						Finish.makeBlockIntoJump(blk.top);
					break;
				}

				// If larger blk in loop than blk, then blk.end not loop end
				if (outer.end!=same)
					same = 0;
			}
		}

		// Unless doing cases, update context to outer block
		if (blk.type!=Trek.ST_CASE)
			blk = levs[level];
		else addCase();
	}									// End addEndStatement

	// Append if-stmt & start of Then block. Then do more if special case:
	// ifjump = target of if stmt
	// instEnd is at end of if-stmt
	private void addIfStuff(int ifjump)
	{
		// Chk context before creating the Then blk.
		// 1) See if target is loop bottom of loop below this If.
		// 2) See if ifjump improves postHandlers of some parent of this blk.
		// (Note that brk/contin vs. invis jump past catches is handled in
		// Finish's fixupTryCatchExit).
		int upto = isLoopBottom(ifjump, instEnd, levs[level].end);
		deoptJumpPastCatches(ifjump);

		// Put out If/Then and open Then blk. If normal case, done.
		// Scan will add stmts in Then and end-then normally.
		Statement ifst = addStatement(Trek.ST_IF, null);
		ifst.addr = ifjump;
		Statement then = addEmptyStatement(Trek.ST_THEN, level, "{");
		addBlock(Trek.ST_ENDTHEN, ifjump, "} // End if", then);

		// 1) Noop Then (ie. "if (x) ; no-else"), add empty Then body
		if (ifjump == instEnd)
		{
			addEmptyStatement(Trek.ST_EMPTY, level, ";");
			addEndStatement(stmtStart);
		}

		// 2) If-stmt was optim to jump to loop bottom rather than loop top
		// of loop following if-then. De-opt end-then to top of loop.
		else if (upto>=0)
		{
			Instruction inp = Instruction.at(currInst, upto).prev();
			if (inp.getTarget()==ifjump)
				upto = inp.start;			// Goto-down part of loop start
			blk.end = blk.bottom = upto;
		}

		// 3) If if-target is past bottom of blk that Then-blk is in, if it
		// jumps up, or if a Stmt jumps into Then blk from above, then treat 
		// if-target as codeless brk or contin.
		// (This code pattern can also be caused by an optimized branch down.
		// If so, Finish.makeElse will replace the code-less branch with
		// a Then blk built using the deopt-ed target).
		else if (ifjump>levs[level-1].bottom || ifjump<instEnd
			|| levs[2].top.goesBetween(then.start, ifjump-1))
		{
			Statement jump = addEmptyStatement(Trek.ST_UNKNOWN, level, null);
			ifst.addr = jump.start;		// Since Then now 0-length
			jump.addr = ifjump;
			addEndStatement(stmtStart);
		}
	}

	// Add codeless loop end. This occurs when bottom of loop is optim away
	// because contin, brk, ret, or throw is on all paths that precede
	// the end of the blk (and there was no branch down in 1st place).
	// Also Divide cant detect that branch-up itself forces end of nested
	// block unless there is a branch down to it (ie. an unopt Contin).
	// In this case, Divide gens ugly but still ok code. For ex:
	//	lp: while (true) {switch () {case: break lp; case: xxx; continue;}}
	// instead of:
	//	lp: while (true) { switch () {case: break lp; case: xxx;}}
	private void addInvisLoopEnd()
	{
		// Go back to last real stmt preceding the end stmt.
		// (Entire Finally blk not real since it is essentially subroutine).
		Statement non0 = meth.lastStatement();
		if (non0.type==Trek.ST_ENDFINALLY) non0 = non0.mate.prev();
		while (non0.insts.length==0 || non0.type==Trek.ST_ENDSYNC) {
			non0 = non0.prev();
		}

		// If invis loop-end is fallen-into, this must mean there is code-less
		// Break before loop-end -- else there would be branch-up there
		switch (non0.type) {
		case Trek.ST_UNKNOWN:
		case Trek.ST_THROW:
		case Trek.ST_RETURN:
			break;
		default:
			addEmptyStatement(Trek.ST_BREAK, level, "break;");
		}

		// Add code-less loop end
		addLoopEnd(true, Vmi.it_branch);
	}

	// Have just seen branch up to top of loop with invis-start.
	// Returns loop-start stmt (creating it and its block if need be)
	// loopstart = addr where loop bottom branches back to
	private Statement addInvisLoopStart(int loopstart)
	{ 
		Statement succ;				// The stmt succeeding top of loop

		// Chk if already created this loop
		for (int i=level; i>=2; i--) {
			Block lev = levs[i];
			succ = lev.top.next();
			if (lev.type==Trek.ST_ENDDO && succ.start == loopstart)
				return lev.top;		// Noop
		}

		// If branch back to top of loop has been optimized to branch back
		// to bottom of loop (recall visible loops start with goto bottom),
		// then canonicalize branch target back to top of loop. Note that
		// stmt is null when code is "while (...);" That is, if entire loop is
		// the loop-end stmt, an addStatement hasnt been done for it yet.
		Statement stmt = meth.statementOf(loopstart, null);
		if (stmt!=null && stmt.type>=Trek.ST_ENDLOOP)
			loopstart = stmt.mate.start;

		// Setup for usual do {} while (...) case.
		stmt = meth.lastStatement();
		int looptype = Trek.ST_DO;
		int endtype = Trek.ST_ENDDO;

		// Find loop-top if while (...) is itself not loop top. That is,
		// while (...); OR do {;} while (...); are legal, so there may be
		// no succ at all.
		int looplev = level;
		succ = null;
		for (; stmt!=null && stmt.start>=loopstart; stmt=stmt.prev()) {

			// The st.in.len==0 test is needed to exclude empty stmts before
			// loop-start, eg. "if (...) {...} do"  and  "if (...) { do"
			// Conversely st.type chks allow for nested invis loop starts
			if (stmt.start==loopstart && stmt.insts.length==0
				&& stmt.type!=Trek.ST_FOR
				&& stmt.type!=Trek.ST_WHILE
				&& stmt.type!=Trek.ST_DO)
			{
				// Also disting between invis-loop {try} and try {invis-loop}
				// Latter case means havent seen end-try stmt yet
				if (stmt.type!=Trek.ST_TRY || stmt.mate==null)
					break;
			}

			// Bump level of loop's stmts.
			// Curr stmt could be nested continue, so level of loop cannot
			// be determined from it. Use 1st stmt of loop instead.
			succ = stmt;
			stmt.level += 1;
			looplev = stmt.level - 1;
		}

		// When: if (xxx) {... [jump/ret/throw]} while(...)/for(;...)
		// is optim, the If branches directly to loop-end rather than goto-ing
		// past Then and letting goto there do it. There are 2 sub-cases.
		// 1) the Then ends with fall thru, so a goto loopbot is after Then 
		//    blk and thus this case is handled by normal vis-loop machinery.
		// 2) it doesnt, so loop-top is invis -- but the loop is not a Do.
		if (succ!=null)
		{
			Statement stp = succ.prev();
			if (stp!=null && stp.type==Trek.ST_ENDTHEN
				&& stp.mate.prev().addr > loopstart)
			{
				looptype = Trek.ST_WHILE;
				endtype = Trek.ST_ENDWHILE;
			}
		}

		// Create loop-start stmt and return it. Dont know end of loop
		// yet, so conventionally set end to its outer limit -- the bot of
		// the block encompassing the loop
		Statement top = new Statement(
			meth, loopstart, 0, null, looptype, looplev, null);
		if (succ!=null) top.insert(succ);
		addBlock(endtype, levs[looplev].bottom, null, top);

		//Insure created blk is fully setup
		insertInvisLoopBlock(looplev+1);
		return top;
	}									// End addInvisLoopStart

	// Add loop-end stmt
	// invstart = true if loop has invis loop-start (no goto down)
	// endbranch = the type of branch-back: branch or if-inst
	private void addLoopEnd(boolean invstart, int endbranch)
	{
		// Fix goto's above loop that branch to bottom of loop. This
		// allows if-blks to end correctly & insures assign stmt preceding
		// loop is not treated as loop init. (That is, u cant directly
		// hop to bottom of loop if u need to execute init clause).
		Statement top = blk.top;
		setCanonLoopTargets(levs[level-1].top, top);

		// Build stmt, bind top and bottom, and move to outer blk
		level -= 1;					// Move to level of blk-end's stmt
		Statement bot = addStatement(blk.type, null);
		top.mate = bot;
		bot.mate = top;
		blk = levs[level];			// Make info at outer level curr
		return;
	}								// End addLoopEnd

	// At a branch up. Add an optimized continue or loop-end stmt
	// loopstart = addr where curr inst branches back to
	// endbranch = the type of branch-back: branch or if-inst
	// Notes:
	//	May create the loop-start for the continue/loop-end as well
	//	Optimized Continue is possible in 1st place only if there is
	//	no work (incr or test) to do at bottom of each loop iter.
	private void addLoopStuff(int loopstart, int endbranch)
	{
		// Add a loop-start if start of loop has been invis til now. 
		// This occurs if:
		// 1) no branch-down at top of loop, which occurs for:
		//    do...while,  while (true), for(init; guaran true 1st time; incr),
		// 2) goto-down has been optimized away (ie. if there were a goto-down,
		//	  its target would be the curr branch-up stmt. Note that this
		//    incl the optimization that the branch-up to loopstart has
		//	  been optimized to go to the bottom of an immed inner loop)
		boolean invstart = false;
		Statement top = blk.top;
		if (top.type!=Trek.ST_WHILE
			|| (blk.bottom!=stmtStart && blk.bottom!=loopstart))
		{
			top = addInvisLoopStart(loopstart);
			loopstart = top.start;
			invstart = true;
		}

		// If curr inst is before [best guess as to loc of] loop bot
		// And (there is later goto to top of loop
		//	    Or (it is nested And there is not goto
		//          to it from above start of its apparent blk)),
		// then it is Contin that goes directly to top of loop.
		boolean latergoto;
		Block loopblk = levs[top.level+1];
		if (instEnd<=loopblk.bottom
			&& ((latergoto=isLaterGoto(loopstart, instEnd, loopblk.bottom))
			  || (top.level+1<level
				&& !top.goesBetween(blk.top.start+1, stmtStart))))
		{

			// If there is no later goto, tell addEndStmt to end loop asap --
			// right after cases, catches, or if-clauses Contin is in. That
			// is, no later goto means loop is over, except syntactially.
			if (!latergoto)
				loopblk.endAsap = true;

			// Continue can be goto inst or an if-inst from an optimized
			// "if (xxx) continue" (ie. "if-inst loop-top" rather than
			// unopt "if-inst .+3; goto loop-top")
			if (endbranch==Vmi.it_branch)
			{
				Statement stmt = addStatement(Trek.ST_UNKNOWN, null);
				stmt.addr = loopstart;
			}
			else addIfStuff(loopstart);
		}

		// Else add loop bot. First tho, end any open inner blks. (This
		// happens because invis loop bottom wasnt known when their .end
		// value was set, so it was set to bottom of more top level blk)
		else {
			loopblk.bottom = stmtStart;		// Bottom info now known
			loopblk.end = instEnd;
			for (int lev=top.level+1; lev<level; lev++) {
				addEndStatement(stmtStart);
			}
			addLoopEnd(invstart, endbranch);
		}
	}										// End addLoopStuff

	// Add a Statement object at curr blk level for insts just scanned
	// type = type of this stmt
	// str = source code to assign at this time (or null)
	private Statement addStatement(int type, String str)
	{
		// Add the stmt whose ender was just seen.
		Statement stmt = new Statement(meth,
			stmtStart, instEnd-stmtStart, prim.insts, type, level, str);
		stmtStart = instEnd;

		// If this is end of in-lined method, point to load at its start
		if (stackSize<0)
		{
			if (ifLoan!=null)
			{
				stmt.loan = ifLoan;
				ifLoan = null;
			}
			else {
				stmt.loan = saveLoan;
				saveLoan = null;
			}
		}
		stackSize = 0;

		// If this is start of in-lined method, it may start with load
		// of objref that will be used to identify where to store eventual
		// result in a later stmt. If so, save it.
		if (maybeLoan)
		{
			saveLoan = stmt;
			maybeLoan = false;
		}
		return stmt;
	}

	// Add empty cases & empty goto's for optimized-away switch body
	// cas = case info for 1st short-circuited case
	private void addSwitchTail(Case cas)
	{
		// Indic bottom of this sort of Switch is already known, and
		// set 1st case to be processed this way
		blk.top.addr = Statement.ad_bound;
		int castart = cas.pc;

		// If switch is in Try or Catch, setup to improve postHandlers.
		Block outer = null;
		int tcend = 0;
		int lev = withinBlock(-1);
		if (lev>=0)
		{
			tcend = levs[lev].end;
			outer = levs[lev-1];
		}

		// Scan rest of cases. Output goto to where case's code ended up,
		// since it could have been contin or brk opted into switch vector.
		// However just 1-goto for labels with no code between them.
		while (true) {
			if (castart!=cas.pc || blk.cases==null)
			{
				Statement stmt = addEmptyStatement(Trek.ST_UNKNOWN, level+1, null);
				stmt.addr = castart;

				// If lowest cas.pc beyond end of t/c blk (and before curr 
				// postHandlers) is a better postH guess -- and doesnt start
				// jsr to chain's Finally blk, update postH
				if (outer!=null && castart>tcend && castart<outer.postHandlers)
				{
					Instruction inst = Instruction.at(currInst, castart);
					if (inst.opcode!=Code.jsr) outer.postHandlers = castart;
				}
				if (blk.cases==null) break;
			}

			// Output label for next case & move to next case if any
			addEmptyStatement(Trek.ST_CASE, level, cas.toString());
			castart = cas.pc;
			if (blk.casidx < blk.cases.size())
				cas = (Case)blk.cases.elementAt(blk.casidx++);
			else blk.cases = null;
		}
	}
			
	// Updates maybeLoan as needed
	private void checkLoan(Instruction inst)
	{
		// Inlined methods can preload ref for storing retval.
		// Ref at start of stmt can be loan to subseq stmt that does this
		// store. It can be obj ref (for put*) or element ref (for xastore).
		if (stmtStart==inst.start)
		{
			if (inst.isRefLoad() || inst.type==Vmi.it_new)
				maybeLoan = true;
			return;
		}

		// Any possibility curr inst gets rid of cached ref?
		if (!maybeLoan || stackSize>1)
			return;						// No

		// Stack empty obviously gets rid of it
		if (stackSize<=0)
			maybeLoan = false;

		// If get replaced top of stack, loan could still be active
		// if gotten thing is array or obj ref
		if (inst.type==Vmi.it_loadfld)
		{
			CpInfo cpi = (CpInfo)inst.getReferenceArg();
			String desc = cpi.toDesc();
			if (!desc.startsWith("L") && !desc.startsWith("["))
				maybeLoan = false;
		}
	}

	// Called if see astore, jsr, or monexit at top-level of Sync blk:
	// 1) Adds end-sync stmt  OR
	// 2) Does nothing because continue, break, etc is after jsr/monexit  OR
	// 3) Does nothing because not in end-sync in 1st place.
	// inst = top-level astore, monitorexit, or jsr in sync blk
	// jm = at monitorexit or jsr
	// Returns where to resume eating insts (not curr pos only for #1)
	// Note:
	//	The use of jsr to access the implicit Finally in an end-sync is 
	//	really a waste since such finally's do only an aload/monexit. But
	//	all the early compilers do this.
	private int checkAddSyncEnd(Instruction inst, boolean jm)
	{
		// If at (jsr|monexit not-goto), eat not-goto in usual way
		// (and ctl will come back here when Finally-handler reached).
		// If at something else altogether, just ret immed of course.
		// Note: hopover code seq & start of handler mutually exclusive
		Instruction inn = inst.next();
		boolean hopover = jm && inn.type==Vmi.it_branch;
		if (!hopover && isHandlerStart(stmtStart)==null)
			return instEnd;			// Not at j|m goto and not in handler

		// Scan to bottom of end-sync stmt so can:
		// 1) Add it  OR
		// 2) See that goto was contin or break, rather than hop over
		int syncend = instEnd;
		for (Instruction in=inn; ; in=in.next()) {
			if (in==null)
			{
				if (!hopover)
					break;
				else return instEnd;	// Pattern not seen by end of method
			}
			syncend += in.size;

			// jsr inst thru its ret is end-sync
			if (in.opcode==Code.ret)
				break;

			// Athrow is bot of end-sync if end-sync starts with monexit.
			// But not at that case if athrow is followed by astore that 
			// starts jsr body of a sync's implicit Finally blk. (Conversely
			// dont confuse this with astore at start of encompassing handler).
			if (in.opcode==Code.athrow)
			{
				Instruction in2 = in.next();
				if (in2==null || !in2.isRefStore()
					|| isHandlerStart(in2.start)!=null)
					break;
			}
		}							// End loop

		// Saw end-sync pattern. Setup to chk goto target if at hopover seq.
		// If are (or at handler), add end-sync stmt and return updated scan pt
		if (hopover)
		{
			int target = inn.getTarget();
			int top = deoptTarget(target, syncend, levs[level-1].end);
			if (top>=0) target = top;
			if (target!=syncend)
				return instEnd;			// Goto wasnt hop over, so eatInst it
		}
		instEnd = syncend;
		addEndStatement(instEnd);
		return instEnd;
	}

	// If target exits handler blk and it jumps to end of loop,
	// shrink postH for that try/catch chain if de-opted target is better
	// target = where an inst jumps down to
	// level = level of stmt that branches to specified target
	// Notes:
	//	Stmt could be break or contin, or it could be blk-internal jump
	//	that was optimized to jump past Catch's. Finish's fixupTryCatchExit
	//	will decide this eventually.
	private void deoptJumpPastCatches(int target)
	{
		if (level<=1) return;
		Block curr = blk;

		// Scan encompassing blks for a Try/Catch chain that target jumps 
		// out of. When reach blk target is within, there is no more to chk.
		for (int lev=level-1; ; lev--) {
			if (target<=curr.end) return;	// Not a jump past catches
			Block parent = levs[lev];

			// Make improved-posth chk if at try/catch level
			if (curr.type==Trek.ST_ENDTRY || curr.type==Trek.ST_ENDCATCH)
			{
				int top = deoptTarget(target, curr.end, parent.end);
				if (top>=0 && top<parent.postHandlers)
					parent.postHandlers = top;
			}
			curr = parent;
		}
	}

	// Returns de-optimized target if it was branch down to end of loop, else -1
	// target = addr that may be start of loop bottom's if-test/goto
	// start = where it nominally would branch back to
	// blkend = end of encompassing blk (loop end can be no further)
	private int deoptTarget(int target, int start, int blkend)
	{
		// Ret immed if not branching to loop bottom
		int top = isLoopBottom(target, start, blkend);
		if (top<0)
			return top;

		// Handle invis vs. vis loop top
		Instruction curri = Instruction.at(currInst, top);
		Instruction previ = curri.prev();
		if (previ.getTarget()==target)
			return previ.start;		// Vis top (ret loc bef down-goto at top)
		else return top;			// Invis top (ret loc bot goes back to)
	}

	// Divide the method's primordial stmt into its real stmts
	Divide(Method method, Statement primord)
	{
		if ((Trek.debug&Trek.Db_at) != 0)
			Trek.println(System.out, "Dividing " + method.toName());

		// Access and remove primord stmt
		// For effic, cache primordial stmt's insts, and position to start
		meth = method;
		prim = primord;
		methLen = prim.insts.length;
		instEnd = 0;

		// Need top & bottom of blk for level 1, so level chks work even when
		// processing top-level code. Top-level blk spans method.
		// Scan thru insts, determining which end statements.
		try {
			currInst = Instruction.cache(prim, meth.parent.parent.icache);
			Statement top = addEmptyStatement(Trek.ST_LABEL, 0, null);
			addBlock(Trek.ST_ENDLABEL, prim.insts.length, "", top);
			nextTry = findNextTry(null);	// Setup 1st try start pt if any
			while (instEnd<methLen) {

				// Output end-of-blk stmt when scan reaches curr blk's endpt.
				if (instEnd>=blk.end)
				{
					addEndStatement(stmtStart);
					continue;				// Mult levels can end at same pt
				}

				// At a loop with invis bottom, so it needed to be ended
				// as soon as loop's blk was at top of levels stack
				if (blk.endAsap)
				{
					addInvisLoopEnd();
					continue;
				}

				// If at start of next try block, create Try stmt and blk,
				// and find the next Try block to scan for.
				if (instEnd==nextTry.startPc)
				{
					top = addEmptyStatement(Trek.ST_TRY, level, "try {");
					blk.lastTry = nextTry;

					// Init postHandlers to nearer of parent's bottom and
					// best Goto seen by findNextTry. (If's & goto's eaten 
					// while try/catch open may improve it more).
					// Hack: set try eob to +1 past eob so that out-of-blk
					// chks work right, but that addCatchStart rather 
					// than top-level loop ends the Try blk.
					if (nextTryPost>=0 && nextTryPost<blk.bottom)
						blk.postHandlers = nextTryPost;
					else blk.postHandlers = blk.bottom;
					addBlock(Trek.ST_ENDTRY, nextTry.handlerPc+1,
						"} // End try", top);
					nextTry = findNextTry(nextTry);
					continue;				// Nested try's possible
				}

				// Eat first (or later) instruction of method
				currInst = Instruction.at(currInst, instEnd);
				instEnd += currInst.size;		// Position to next inst
				eatInst(currInst);
			}								// End loop

			// Term the top-level blk that implicitly spans entire method,
			// and any other blocks that were left open.
			while (level > 0) addEndStatement(stmtStart);
		}

		// Report that divide failed and set method contents to all insts
		// + the stmts that were successfully divided
		catch (Exception e) {
			meth.logMethodError("dividing", e);
			primord.insert(meth.first);
		}
	}									// End of explode

	// Decide what this branch inst corresponds to.
	// Returns where to resume processing (which is right after the
	// branch unless processing condit expr)
	private int eatBranch(Instruction inst)
	{
		Statement last;
		int target = inst.getTarget();

		// 1) A condit expression (ie. a ? b : c)? If jumping past PC of
		// top-level ":" (ie. condRight), we can now continue processing 
		// after c since b and c obviously both have to leave stack with 
		// same amount of data.
		// Conversely had to process bit by bit til thru end of b to calc 
		// that stack size in 1st place. Similarly want to skip over nested 
		// c's within top-level b. However if nested b has been optimized
		// to jump over top-level c, that is equivalent to reaching end
		// of top-level b -- so we can end cond mode for that too.
		if (condRight!=0)
		{
			if (target>condRight)
				condRight = 0;			// End cond mode
			return target;
		}

		// 3) branch back at bottom of if-less loop or optimized continue
		if (target < instEnd)
			addLoopStuff(target, Vmi.it_branch);

		// 4) branch down at loop start (ie. so exit-loop test is done
		// before 1st iter of loop). Can involve case of blk (like If) in
		// front of loop optim to branch to bottom of loop. If so, put out
		// end-stmt for that blk now. (If that blk is loop, then this is
		// simply nested loops rather than the described special case).
		else if (isLoopBottom(target, instEnd, blk.end)>=0)
		{
			if (blk.bottom>inst.start && blk.bottom<=target
				&& blk.type<Trek.ST_ENDLOOP)
				addEndStatement(stmtStart);

			// Consecutive identical goto's are really 1 stmt
			Instruction in2 = inst.next();
			if (in2.type==Vmi.it_branch && in2.getTarget()==target)
				instEnd += in2.size;
			Statement top = addStatement(Trek.ST_WHILE, null);
			addBlock(Trek.ST_ENDWHILE, blk.end, null, top);
			blk.bottom = target;
		}

		// 5) If it is noop goto before jsr, it is not end-of-stmt.
		else if (target==instEnd && inst.next().opcode==Code.jsr)
			;

		// 6) something that needs to be decided after method exploded:
		// else-bypass at end-then, break, continue, or jump at catch-end
		else {

			// Chk if also involves in-lined if/then/else special case
			if (stackSize>0  && blk.type==Trek.ST_ENDTHEN && blk.end==instEnd)
			{
				stackSize = 0;
				Statement ifst = blk.top.prev();

				// Build dcl-only local for temp that is used in the
				// inlined if (will be named if + ifLoan.start)
				ifLoan = new Statement(meth, ifst.start,
					Trek.ST_LOCAL, ifst.level, null);
				ifLoan.insert(ifst);
				ifLoan.addr = target;
				ifLoan.loan = saveLoan;
				saveLoan = null;

				// Build the stmt at end of Then that "uses" the temp local.
				// Setup src for temp local's dcl
				Statement stget = new Statement(meth,
					stmtStart, inst.start-stmtStart, prim.insts,
					Trek.ST_ASSIGN, level, null);
				stmtStart = inst.start;
				stget.loan = ifLoan;
				ifLoan.string = new Vmi().execute(stget);
			}

			// Add the goto stmt
			Statement st = addStatement(Trek.ST_UNKNOWN, null);
			st.addr = target;
			deoptJumpPastCatches(target);
		}
		return instEnd;
	}									// End eatBranch

	// Process a call instruction. If it ends stmt, build a stmt object
	// Returns (updated) point to resume scan
	private int eatCall(Instruction inst)
	{
		// A constructor (vc<0) is like a non-void if its new is followed
		// by dup inst (which occurs if "new xxx();" is not entire stmt).
		int vc = isVoidCall(inst);
		if (vc<0 && stackSize > 0)		// Constructor & new's dup on stack?
			return instEnd;				// Yes, new object will be stored

		// If used ret val, this is "function call" in middle of stmt.
		// Also if retval "replaced" loan, dont consider it a loan.
		if (vc==0)
		{								// Not void
			Instruction in2 = inst.next();
			if (in2.opcode!=Code.pop && in2.opcode!=Code.pop2)
			{
				if (stackSize<=1) maybeLoan = false;
				return instEnd;			// A used retval
			}

			// If unused ret val, incl pop in what was eaten
			stackSize += in2.getStackDelta();
			instEnd += in2.size;		// Ignored retval

			// It at new-inner-class mumbo jumbo, not ending stmt
			if (stackSize>=2)
				return instEnd;

			// End stmt since retval unused, and
			// because of lookahead, do stmt bndary chking here
			if (stackSize<=0) maybeLoan = false;
		}

		// Void or unused ret val. Build a stmt. Normally stackSize is <=0,
		// but a jtrek-instrumented method can put a call anywhere.
		if (stackSize>0) maybeLoan = true;
		addStatement(Trek.ST_CALL, null);
		return instEnd;
	}

	// Process the next instruction of the method
	// Notes:
	//	JVM stack size must be 0 at end of stmt, so stackSize is used to help
	//	detect end of stmt. A leap in instEnd is okay wrt stackSize if change
	//	is from jump inst to target, since target could be next inst executed.
	//	Else it is called method's responsibility to update stackSize.
	private void eatInst(Instruction inst)
	{
		Instruction in2;

		// Display latest inst if tracing
		if ((Trek.debug&Trek.Db_atsmall) != 0)
			Trek.println(System.out,
				inst.start + "(" + level + "): "
				+ inst.toString() + " " + stackSize);

		// Primary stmt boundary chking is based on opd stack being used up.
		// Unfort inlined methods can preload ref for storing retval.
		// Use maybeLoan to catch many (but not all) of these cases.
		stackSize += inst.getStackDelta();
		checkLoan(inst);
		switch (inst.type) {

		// Chk if dup-ing elims the loan
		case Vmi.it_dup:
			if (maybeLoan)
			{
				int totop = -1;
				switch (inst.opcode) {

				// Special case sync(xxx) use of temp local to not be
				// treated as loan of store of temp local
				case Code.dup:
					in2 = inst.next();
					if (in2.type==Vmi.it_storelv
						&& in2.next().opcode==Code.monitorenter)
						totop = 2;
					break;

				// If dup-inst replaces stk top, top wasnt a pre-cached ref.
				case Code.dup_x1:	totop = 3;	break;
				case Code.dup_x2:	totop = 4;	break;
				case Code.dup2_x1:	totop = 5;	break;
				case Code.dup2_x2:	totop = 6;	break;
				}
				if (stackSize==totop)
					maybeLoan = false;
			}
			break;

		// Update instEnd to as close to end of if-stmt as can now.
		// See details in findThen/addStatement.
		case Vmi.it_if:
			instEnd = findThen(inst);
			break;

		// Access method's return-info in constant pool to
		// determine if stand-alone call. If not, still not to chk
		// if code ignores return result (ie. next instruction is pop)
		case Vmi.it_call:
			instEnd = eatCall(inst);
			break;

		// An uncond branch can be a number of things.
		case Vmi.it_branch:
			instEnd = eatBranch(inst);
			break;

		// iinc inst (which is used for +/- lv && ++/--) is a stmt unless
		// the value of the expression is used in the stmt (eg. a=b[i++]).
		// it_loadlv case test for this by looking for: load iinc. Note
		// that iinc has stack delta of 0.
		case Vmi.it_loadlv:
			in2 = inst.next();
			if (isIincLoad(in2, inst))
				instEnd += in2.size;			// iinc doesnt end stmt
		case Vmi.it_loadconst:
		case Vmi.it_loadel:
		case Vmi.it_loadfld:
			break;

		// Add a switch stmt & stuff to recog case stmts & end of switch
		case Vmi.it_switch:
			eatSwitch(inst);
			break;

		// A store to local.
		// Close Sync blk if at sync's implicit Finally handler
		case Vmi.it_storelv:
			if (blk.type==Trek.ST_ENDSYNC)
			{
				instEnd = checkAddSyncEnd(inst, false);
				if (blk.type!=Trek.ST_ENDSYNC)
					break;			// Added Sync end
			}

			// Storing exception left on stack for Catch is only situation
			// a store-lv can start a stmt. But if saveLoan non-null,
			// this store is not the logical start of a stmt -- that
			// occurred in the loaning stmt.
			if (inst.start==stmtStart && saveLoan==null)
			{							// Worth cost of isHandlerStart call					
				ExceptionHandler exc = isHandlerStart(inst.start);
				if (exc!=null)
				{
					instEnd = addCatchStart(inst, exc);
					break;			// Added Catch start
				}
			}

			// A store to undef local when there is lvt means
			// this is temp local within a stmt, so dont end stmt.
			// Also workaround ms compiler bug that exception scopes
			// start at 0.
			if (meth.lvRead>=0)
			{
				Local loc = meth.getLocal(inst.getLocalIdx(), inst);
				if (loc==null
					|| (loc.startPc==0 && loc.length<prim.insts.length))
					break;
			}

		// Merge with other store's.
		// Generate stmt for assign unless:
		// 1) It is start of loop bottom (eg. i++<n)
		// 2) If stack not empty, result of store will be used in stmt.
		//    For ex, a=b=c will do this "at the b" so to speak.
		case Vmi.it_storel:
		case Vmi.it_storefld:
			if (isInLoopBottom())
				break;

			// If a and b are local objects, the store of b looks like loan
			// because original load was object and stk size now 1. So chk
			// if next inst is store to see if top of stack was a thing
			// rather than a ref to a thing
			int stk = stackSize;
			if (stk>1
				|| (stk==1 && (!maybeLoan || inst.next().type==Vmi.it_storelv)))
				break;

			// End the stmt. Deal with special case that final field may be
			// assigned value in constructor of its class even tho at source
			// level this is taken care of with init clause of field. That is,
			// suppress source code gen for this assign because it is illegal
			// at source level. Use full comment syntax since stmt may be
			// dumped as multiple lines
			Statement stmt = addStatement(Trek.ST_ASSIGN, null);
			if (inst.type==Vmi.it_storefld)
			{
				ClassFile cf = meth.parent;
				Field fld = inst.toMemberClass().equals(cf.toQualifiedName())
					? (Field)inst.getMember() : null;

				// But dont do the chk if this is synthetic fld for inner class
				if (fld!=null && !fld.hasModifier(Trek.ACC_STATIC)
					&& fld.toName().indexOf('$')<0)
				{
					String src = stmt.toString();
					if (fld.hasModifier(Trek.ACC_FINAL))
						stmt.string = "/* " + src + " */";
					else if (src.startsWith(fld.toName() + " = {"))
						stmt.string = "/* " + src + " */";
					else stmt.string = null;	// Too early to cache regular code
				}
			}
			break;

		// A return or throw always ends a stmt
		case Vmi.it_return:
			int type = inst.opcode==Code.athrow ? Trek.ST_THROW : Trek.ST_RETURN;
			addStatement(type, null);
			break;

		// Other insts
		case Vmi.it_other:
			switch (inst.opcode) {

			// Exception handlers that dont use generated exception
			// can pop the exception. If not that case, it is
			// probably bad generated code, but a pop is effectively
			// like a store in the sense it means "done with something"
			case Code.pop:
				ExceptionHandler exc;
				if ((exc=isHandlerStart(inst.start))!=null)
					instEnd = addCatchStart(inst, exc);	// Catch or Finally
				else if (stackSize<=0)
					addStatement(Trek.ST_EMPTY, ";");
				break;

			// Iinc is nominally treated as an Assign. Thus we gen 2 stmts
			// (i++; x=i;) rather than (x=++i;) unless:
			// 1) iinc is start of loop bottom (eg. ++i<n). Note that for
			// incr clause, blk.bottom would follow not precede the iinc).
			// Also see addInvisLoopStart & "end-then assign" special case.
			// 2) iinc is ++i in middle of stmt (ie. iinc load).
			case Code.iinc:
				if (isInLoopBottom())
					break;
				if (stmtStart!=inst.start && isIincLoad(inst, inst.next()))
					break;
				addStatement(Trek.ST_ASSIGN, null);
				break;

			// Process synchronize stmt. Output it and open its blk.
			// Also ignore the Try blk that is implicitly part of it. 
			// Note that sync () { try {} catch {}} looks like a single 
			// try with both a catch and a finally. So sync block has its
			// own nextTry only if nextTry contains just a Finally.
			case Code.monitorenter:
				Statement top = addStatement(Trek.ST_SYNC, null);
				addBlock(Trek.ST_ENDSYNC, blk.bottom, "} // End sync", top);
				if (nextTry.catchType==0)
					nextTry = findNextTry(nextTry);
				break;

			// Exiting sync mode, but only at end-sync stmt if at top-level.
			// However if return stmt at this pt, end-sync stmt will not
			// be added until after Return has been processed
			case Code.monitorexit:
				if (blk.type==Trek.ST_ENDSYNC)
					instEnd = checkAddSyncEnd(inst, true);
				break;

			// End the open Finally blk (sync bottom detected with monexit).
			// Ret always means end of Finally blk. But Finally blks can
			// also end with unconditional Return stmt.
			case Code.ret:
				addEndStatement(instEnd);
				break;

			// Jsr's only exist for Finally/Sync. If at top-level jsr in sync,
			// add end-sync unless need to delay til after return eaten
			case Code.jsr:
			case Code.jsr_w:
				if (blk.type==Trek.ST_ENDSYNC)
					instEnd = checkAddSyncEnd(inst, true);
				break;
			}										// End Other's switch
		}											// End type switch

		// End stmt because it ends else of inlined if/then/else
		if (ifLoan!=null && ifLoan.addr==instEnd)
		{
			Statement st = addStatement(Trek.ST_ASSIGN, null);
			st.loan = ifLoan;
			stackSize = 0;
		}
	}

	// Add statement for switch & build table to identify case-label stmt's
	private void eatSwitch(Instruction inst)
	{
		Statement stmt = addStatement(Trek.ST_SWITCH, null);
		Vector cases = Case.makeCases(inst);

		// Identify end of blk. Initially this is 1st case. Subseq,
		// it is each successive case.
		Case cas = (Case)cases.firstElement();
		addBlock(Trek.ST_CASE, cas.pc, cas.toString(), stmt);
		blk.cases = cases;
		blk.casidx = 1;

		// Until see break to end of switch, assume start of last case is
		// end of switch -- which is where it will be if no break ever seen.
		// Note that at inst-level, one case is always the out-of-range case
		cas = (Case)cases.lastElement();
		stmt.addr = cas.pc;
	}

	// Returns 1st except block of next Try chain in method,
	// or dummy one past end of program if no more left.
	// The main explode loop uses this block's startPc field to decide when
	// to start next Try stmt
	private ExceptionHandler findNextTry(ExceptionHandler nexc)
	{
		// Start scan with upper bound that loses to anything
		// Note: current blk at this point is block Try is in, not
		// the block its contents will be in.
		ExceptionHandler bexc = new ExceptionHandler();
		bexc.startPc = methLen;
		bexc.endPc = methLen;
		bexc.catchType = 0;
		ExceptionHandler fexc = null, lexc = null;

		// Identify where next one is. Combined start/end check insures outer
		// levels to inner levels at same start pt are eaten in that order
		for (int i=0; i<meth.excepts.length; i++) {
			ExceptionHandler exc = meth.excepts[i];

			// Exclude all the already-processed Try chains, if any.
			// Note that when nexc was for a try/catch/finally, the Finally
			// is skipped this time since its endPc is further down the
			// Catch's (which was what nexc was set to last time)
			if (nexc!=null)
			{
				if (exc == nexc) continue;
				if (exc.startPc < nexc.startPc) continue;
				if (exc.startPc == nexc.startPc && exc.endPc >= nexc.endPc)
					continue;
			}

			// At new try-chain if at earlier not-yet-processed try-start PC,
			// or if at same start and further-down end (= outer try-chain).
			if (exc.startPc < bexc.startPc
				|| (exc.startPc==bexc.startPc && exc.endPc>bexc.endPc))
			{

				// Save this exception blk as curr best first and
				// last Catch of this chain. Finally-blks with same startPc
				// excluded since their end-try is after all Catch's. But
				// set fexc to distingush below between Try/Catch/Finally and
				// try {try/catch}/finally. Also note that in former case,
				// Catch's blk guaran before Finally's blk in excep array.
				if (exc.catchType==0)
					fexc = exc;
				if (exc.startPc < bexc.startPc || exc.catchType!=0)
					bexc = lexc = exc;
			}

			// See if at earlier or later Catch of this chain. (Save last 
			// excep blk so can set nextTryPost & chk if at catch-less chain)
			else if (exc.startPc==bexc.startPc && exc.endPc==bexc.endPc)
			{
				if (exc.handlerPc>lexc.handlerPc)
					lexc = exc;
				else if (exc.handlerPc<bexc.handlerPc)
					bexc = lexc;
			}

		}								// End loop

		// If past all handlers, return dummy bexc
		// If only blk at this startpt was a Finally, return it
		if (lexc==null)
			return bexc;
		if (bexc==fexc)
		{
			nextTryPost = fexc.handlerPc;
			return fexc;
		}

		// Find loc past chain by scanning for branches. If Finally not
		// involved, just return bexc. Else if Try/Catch is nested in 
		// Try/Finally, Return the Finally's excep blk. (Is nested if Post
		// loc is before the Finally)
		nextTryPost = goesPastChain(bexc.startPc, lexc.handlerPc, fexc);
		if (fexc==null || nextTryPost<0 || nextTryPost>=fexc.handlerPc)
			return bexc;			// Normal case
		else return fexc;			// Nested in Try/Finally case
	}

	// Determine where this test-stmt ends. (Test-stmt means an
	// actual if-stmt, exit-test in a loop stmt, test-clause in a?b:c).
	// inst = 1st if* instruction in curr stmt
	// Returns:
	//	Where to resume top-level eat-inst scan.
	//	If in middle of condexpr, that is next inst.
	//	If recog test-stmt, builds it and returns where test-stmt ends.
	//  If recog condexpr, returns end of its 1st AND clause and sets
	//	condRight to pc after b-part of a?b:c.
	// Notes:
	//	Does not support loops that have loop-test at the top of the loop.
	//	Not known to be used, but... Anyway the chk would be for forward branch
	//	in an if-inst whose target branches back.
	private int findThen(Instruction inst)
	{
		// If in middle of condit expression, if-inst's not relevant
		// to finding end of stmt
		if (condRight!=0)
			return inst.start + inst.size;

		// A branch-out if-inst always ends test-stmt, but it will also appear
		// in the middle of the test-stmt when only top-level ANDs follow it).
		int andend = 0;				// Addr of 1st inst past last-seen AND clause
		int ifjump = 0;				// Addr where if-stmt jumps to:
									//   after then-stmt or start of loop

		// Hop thru if-insts until see one that branches out of the test-stmt.
		while (true) {
			int pt_start = -1;				// Presume not relevant
			int target = inst.getTarget();	// Target of curr if-inst
			int ifend;						// End of curr if-inst
			while (true) {
				ifend = inst.start + inst.size;
				if (target >= methLen)
					throw new TrekException("If-instruction branches past end of method");

				// For end-loop-stmt, branching out of test-stmt simply means
				// a backwards branch. For if-stmt and a?b:c, the simplest chk
				// is that target is past the bottom of the encompassing blk.
				// (This happens for if-stmt in: loop { if (...) break; ...}
				// if the compiler gens the break as the target of the if-inst.
				// Also chk patho case of jumping to next inst (if (x) ; ...)
				if (target <= inst.start) break;
				if (inst.start < blk.bottom && target > blk.bottom) break;
				if (target == ifend) break;
				
				// Otherwise br-out from if/a? is indic by targi.prev NOT
				// being another test in the test-stmt. The main chk to
				// determine this is to see if targi.prev is not an if-inst.
				Instruction targi = Instruction.at(inst, target);
				Instruction ptargi = targi.prev();
				pt_start = ptargi.start;
				if (ptargi.type!=Vmi.it_if)
				{							// Have a branch-out unless in a?b:c

					// a?b:c does not apply or not relevant yet if:
					// 1) andend non-0 -- 1st iter already decided
					// 2) targi.prev is not a Goto
					// 3) The Goto starts the Then (ie. it is brk or contin)
					// Note: a?b:c in middle of If invis because if-scan hops
					// from if-inst target to if-inst target to ...
					if (andend>0 || ptargi.type!=Vmi.it_branch || ifend==pt_start)
						break;

					// The other key requirement is that the code before
					// the Goto not be a stmt (it is usually simply a load).
					// However when at A of "if (A && B) break/contin",
					// hasStatements would return false because the && B has
					// no stmts of course, so exclude this case explicitly.
					Instruction pptargi = ptargi.prev();
					if (pptargi.type==Vmi.it_if
						|| hasStatements(ifend, pt_start)!=null)
						break;

					// Is conditional expr. Enter cond-expr mode
					condRight = targi.start;
					return ifend;
				}

				// 1) Chk if targi is past end of encompassing blk. This says
				// targi must also be past curr If's then-blk. So the prev
				// if-inst must be the last test of the curr If.
				// 2) Chk if at If of: if (...) {loop...end-loop}. This case
				// applies if pt_target is branch up to after start of if-stmt.
				if (targi.start>=blk.end) break;
				int pt_target = ptargi.getTarget();
				if (pt_target < pt_start && pt_target > stmtStart) break;

				// Finally, are there stmts between inst and target. That is,
				// is inst an if-inst that directly breaks or continues?
				if (hasStatements(ifend, pt_start)!=null) break;
				inst = ptargi;
				target = pt_target;
			}										// End loop

			// Targi is after Then-stmt, or it is start of a loop.
			// Ways to detect end of test-stmt, 1st 4 can be chked quickly:
			// 1) If processing a later AND clause and its jump-out loc (ie.
			// ifjump) is diff than the jump-out loc of the curr if-inst
			// (ie. target), then andend is the end of the 1st stmt. For ex,
			// this will happen if method contains 2 consec loop bottoms).
			// BUT in optim code you can appear to get:
			// 1a) branch-back to outer loop followed branch-back to littler loop.
			// This does not indic 2 stmts because it would mean start and end of
			// littler loop are straddling end of outer loop (ie. xcrossing blks).
			// 1b) branch way past bottom of loop followed by branch-back to top.
			// This does not indic 2 stmts if 1st if-insts are within bottom
			// of loop -- just means if's that were supposed to jump just
			// past end-loop stmt have been optim to jump to a later stmt.
			if (andend>0 && target!=ifjump
				&& !((target<andend && ifjump<target)
					|| (target<andend && andend>blk.bottom && ifjump>blk.bottom)))
				return addCondStatement(andend, ifjump);

			// 2) current loop must end by this location (eg. an outer blk
			//    cant end in the middle of a loop-end stmt). By convention,
			//    when loop is created, its blk.end is set to outer block's.
			// 2a) visible loop start. If chk wasnt here, end of outer blk
			//    and loop-end would crisscross.
			// 2b) invis loop start. If chk wasnt here, outer block would
			//    be ended before code realized had even seen invis loop.
			// 3) then-clause just one inst (ie. inst before target's inst
			//    is also the inst right after the end of the if-stmt).
			// 4) If branches to next inst, which happens for:
			//		if (xxx) { ;// code this later } anything-but-else
			// 5) Start of then is start of next Try block
			if (ifend==blk.end
				|| ifend == pt_start
				|| ifend == target
				|| ifend == nextTry.startPc)
				return addCondStatement(ifend, target);

			// 6) If there are stmts after this and-clause, it also ends whole If
			inst = hasStatements(ifend, -1);
			if (inst.type!=Vmi.it_if)
				return addCondStatement(ifend, target);

			// Save end of curr AND clause, and look for next
			andend = ifend;
			ifjump = target;
		}									// End outer loop
	}										// End findThen

	// Returns nearest post-bot address that Try chain jumps to, or -1
	// top = start of Try chain (ie. where to start scan)
	// bot = last Catch blk (ie. where to stop scan/chk branch past)
	// fexc = Finally handler, if any
	private int goesPastChain(int top, int bot, ExceptionHandler fexc)
	{
		int best = -1;

		// Bound the scan to inside the current block and
		// scan for actual inst containing branch back
		Instruction curri = Instruction.at(currInst, top);
		for (; curri!=null && curri.start<bot; curri=curri.next()) {
			int target = curri.getTarget();
			if (target<=bot)
				continue;					// Doesnt exit or no branch

			// If have Finally blk, target could be inst that is common
			// code for invoking Finally blk (jsr or jump to Finally body).
			// So if target inst is a jumping inst that jumps to start of
			// of Finally handler (or later), use start of Finally handler
			if (fexc!=null && target<fexc.handlerPc)
			{
				Instruction targi = Instruction.at(curri, target);
				int retarget = targi.getTarget();
				if (retarget>=fexc.handlerPc)
					target = fexc.handlerPc;
			}

			// Decide if curr target better postHandler
			if (target<best || best<0)
				best = target;				// Found better candidate
		}
		return best;
	}

	// Returns null if goal not achieved. Inst that met goal otherwise.
	// If end>0, goal is seeing stmt. If end<0, goal is that or if-inst.
	// start = where to start scan
	// end = where to end chks (or -1 to stop on if-inst)
	private Instruction hasStatements(int start, int end)
	{
		int stacksize = 0;

		// Scan insts between start and end
		Instruction curri = Instruction.at(currInst, start);
		for (; curri.start<end || end<0; curri=curri.next()) {
			stacksize += curri.getStackDelta();

			// Chk next inst. Branch can be part of condit expr only
			// if something on stack when reach goto. Conversely if past
			// 1st load of condit expr, clear stack size since both loads
			// cant be executed at same time.
			switch (curri.type) {
			case Vmi.it_branch:
				if (stacksize<=0 || curri.getTarget()<=curri.start)
					return curri;
				stacksize = 0;
				break;

			// Branch up cant be part of conditional expr
			case Vmi.it_if:
				if (end<0 || curri.getTarget()<=curri.start)
					return curri;
				else break;
 
			// Calls that dont return a value and  call sites that dont use
			// use the ret val (or dont use a constructed object) end a stmt
			case Vmi.it_call:
				int vc = isVoidCall(curri);
				if (vc<0 && stacksize > 0)
					break;				// Constructed obj used
				if (vc==0)
				{						// There is return val
					Instruction in2 = curri.next();
					if (in2.opcode!=Code.pop && in2.opcode!=Code.pop2)
						break;			// And it's used
				}
				return curri;			// Void method, unused ret val,
										// or new objref not used
			case Vmi.it_storel:
			case Vmi.it_storefld:
			case Vmi.it_storelv:
				if (stacksize-- > 0)
					break;

			// These always end stmts
			case Vmi.it_switch:
			case Vmi.it_return:
				return curri;

			// Other insts. If default or fall thru, keep scanning
			case Vmi.it_other:
				switch (curri.opcode) {
				case Code.ret:
				case Code.jsr:
				case Code.jsr_w:
				case Code.monitorenter:
				case Code.monitorexit:
					return curri;

				// Unless have imbedded i++, ++i, etc, iinc ends stmt
				case Code.iinc:
					if (!isIincLoad(curri, curri.next()))
						return curri;
				}
			}
		}								// End loop
		return null;
	}

	// Insert new invis loop's blk into proper spot in levs
	// blklev = level to insert it at
	private void insertInvisLoopBlock(int blklev)
	{
		Block loopblk = blk;
		int delta = level - blklev;
		if (delta > 0)
		{
			System.arraycopy(levs, blklev, levs, blklev+1, delta);
			levs[blklev] = loopblk;
			blk = levs[level];
		}

		// Move parent-type info from old parent of blklev+1 to loopblk.
		// Note that blklev+1 may not curr exist (eg. between end-try & catch)
		Block oldpar = levs[blklev-1];
		loopblk.postHandlers = oldpar.postHandlers;
		loopblk.lastTry = oldpar.lastTry;
		oldpar.postHandlers = 0;
		oldpar.lastTry = null;
	}

	// Returns its Except if specified loc is catch/finally-start
	// offset = loc to check
	private ExceptionHandler isHandlerStart(int offset)
	{
		// Check exception vector
		// Find the catch/finally block at this stmt address
		for (int i=0; i<meth.excepts.length; i++) {
			ExceptionHandler exc = meth.excepts[i];
			if (exc.handlerPc == offset)
				return exc;
		}
		return null;
	}

	// Returns true if: later inst does branch to loopstart OR to an inner
	// loop's start that is after loopstart and before next
	// loopstart = start of outermost loop that stmt could be Continue for
	// next = loc of inst just past the potential Continue
	// loopbot = furthest down loopstart's bottom could be
	// Notes:
	//	If isLaterGoto is true or currInst branches above an inner block,
	//	currInst is a Continue stmt
	private boolean isLaterGoto(int loopstart, int next, int loopbot)
	{
		// Scan insts between "next" and "loopbot"
		Instruction curri = Instruction.at(currInst, next);
		for (; curri!=null && curri.start<loopbot; curri=curri.next()) {
			int target = curri.getTarget();
			if (target>=loopstart && target<next)
				return true;
		}
		return false;
	}

	// Returns true if iinc inst is next to load of same lv
	// iinc = inst that may be iinc
	// load = adjacent inst that may be load of same lv
	private boolean isIincLoad(Instruction iinc, Instruction load)
	{
		return iinc.opcode==Code.iinc
			&& load.isIntegerLoad()
			&& iinc.getLocalIdx()==load.getLocalIdx();
	}

	// If current inst is in end-loop stmt, return true
	private boolean isInLoopBottom()
	{
		if (blk.type>=Trek.ST_ENDLOOP && stmtStart==blk.bottom)
			return true;
		else return false;
	}

	// Returns true if If-stmt itself does a Break from loop with invis start
	// ifst = if-stmt being checked
	// Returns true if ifst jumps right after jump-inst that looks like a 
	// loop bottom (ie. it goes above If's start && is last place to do this)
	// AND the supposed loop bottom is:
	// 1) Not in an inner blk (which would mean it was an optimized Continue)
	// 2) Not right after ifst itself (which would mean it was a Then that 
	//	  consists of only a Break or Continue stmt).
	// Notes:
	//	If really just jumping beyond optimized continue, makeElse
	//	in Finish will rectify falsely treating this as break-out
	private boolean isInvisJumpOut(Statement ifst)
	{
		Instruction targi = Instruction.at(currInst, ifst.addr);
		Instruction ptargi = targi.prev();
		int ptarget = ptargi.getTarget();
		int blktop = levs[ifst.level].top.start;
		return ptarget>=0 && ptarget<=ifst.start
			&& ptargi.start!=ifst.start+ifst.insts.length
			&& ptarget>=blktop
			&& !isLaterGoto(ptarget, targi.start, levs[1].bottom);
	}

	// Returns addr branched up to if chunk of code at bottom branches
	// up to top of specified (or inner) loop. Else returns -1.
	// bottom = addr that may be start of loop bottom's if-test/goto
	// start = where it nominally would branch back to
	// blkend = end of encompassing blk
	// Notes:
	//	When goto's are optimized, the branch back at bottom of outer loop
	//	could go up to end of inner loop, rather than to its top (ie. start).
	//	So this code checks if branch up is between start and bottom. This
	//	is all that is really required since start is the top of a block and
	//	bottom is approx the end of a block. Iow, a branch before bottom
	//	would be crisscrossing blocks unless we are at start's loop bottom
	private int isLoopBottom(int bottom, int start, int blkend)
	{
		// Bound the scan to inside the current block and
		// scan for actual inst containing branch back
		Instruction curri = Instruction.at(currInst, bottom);
		for (; curri!=null && curri.start<blkend; curri=curri.next()) {
			int target = curri.getTarget();
			if (target>=0)					// A branch inst?
			{								// Is it branch "to" loop top
				if (target < start)
					return -1;				// Up too much (end of outer loop)
				if (target < bottom)
					return target;			// Up to right place
				if (target <= curri.start)
					return -1;				// Up, but to wrong place
				if (curri.type==Vmi.it_branch)
					return -1;				// Down-goto in loop-bot impos
			}

			// Else is non-branching inst or forward-branching if-inst,
			// which are both possible in a complex loop-exit condition
		}
		return -1;							// If/then ended method
	}

	// Returns void-call info for specified instruction
	private int isVoidCall(Instruction inst)
	{
		int refidx = inst.readUnsignedShort(inst.start+1);
		ClassFile cf = meth.parent;
		return cf.constant(refidx).isVoidCall();
	}

	// Replace/extend last statement object to include insts just scanned
	// stmt = stmt to replace
	// start = identifies inst of primordial stmt it should start with
	// str = source code to assign at this time (or null)
	// Returns the new combined stmt
	private Statement replaceStatement(Statement stmt, int start, String str)
	{
		// Do the work
		stmt.remove();
		stmtStart = start;
		Statement stnew = addStatement(stmt.type, str);

		// If replacing stmt that bounds a blk, hook in its replacement
		if (stmt.mate!=null)
		{
			stmt.mate.mate = stnew;
			stnew.mate = stmt.mate;
		}
		return stnew;
	}

	// Replace each goto target that has been optimized to go directly to bottom
	// of loop with canonical target (ie. top of loop).
	// levtop = 1st stmt of blk that loop as whole is within
	// looptop = top stmt of loop
	private void setCanonLoopTargets(Statement levtop, Statement looptop)
	{
		// Loop-bottom's stmt not created yet, so want past end of last stmt
		Statement st = meth.lastStatement();
		int botstart = st.start + st.insts.length;

		// Scan looking for branch to bottom stmt
		for (st=levtop.next(); st!=looptop; st=st.next())
		{
			if (st.type != Trek.ST_UNKNOWN) continue;
			int target = st.addr;		// Get goto's target
			if (target==botstart)
				st.addr = looptop.start;
		}
	}

	// Return arg of catch (eg. Exception foo)
	// inst = 1st inst of Catch stmt
	// cptype = pool info for catch's exception type
	// Notes:
	//	Catch blk & stmt must already exist
	private String toCatchArg(Instruction inst, CpInfo cptype)
	{
		Local loc;

		// If no local (ie. pop), just create dummy name
		String desc = "L" + cptype.toInternalString() + ";";
		int locidx = inst.getLocalIdx();
		if (locidx < 0)
			return "e" + level;

		// Catch stmts define a local for their throwable. If lvt not
		// in classfile, create the local now. However since havent
		// seen end of catch blk yet, set it to best guess of blk length.
		// If blk terminated sooner in addEndStmt, loc len will be fixed then
		else if (meth.lvRead<0)
		{
			int stend = inst.start + inst.size;
			int len = blk.end - stend;
			loc = new Local(stend, len,	null, desc, locidx, meth);
			blk.loc = loc;			// So addEndStatement can fixup its len
			return loc.name;
		}

		// Lvt exists. Just use name defined there.
		while (true) {
			loc = meth.getLocal(locidx, inst, desc);
			if (loc==null)
				return inst.toLocal(locidx);

			// Workaround ms compiler bug that start of local is 0
			// rather than Catch stmt and javac compiler bug that
			// totally bogus def can be in lvt. Note: javac1.3 starts
			// local at catch rather than after inst that stores excep
			int delta = instEnd - loc.startPc;
			if (delta==0)
				return loc.name;
			if (loc.startPc==0 || loc.startPc==inst.start)
			{
				loc.startPc = instEnd;
				loc.length -= delta;
				return loc.name;
			}
			else loc.remove();			// Totally bogus
		}
	}

	// Returns nearest level of specified type(s) of blk, or -1
	// blktype = Trek.ST_END*, or -1 for endtry or endcatch
	private int withinBlock(int blktype)
	{
		for (int lev=level; lev>=2; lev--) {
			int ctype = levs[lev].type;
			if (blktype==-1)
			{
				if (ctype==Trek.ST_ENDTRY || ctype==Trek.ST_ENDCATCH)
					return lev;
			}
			else if (ctype==blktype)
				return lev;
		}
		return -1;
	}

}
