//
// THIS SOFTWARE IS FURNISHED UNDER A LICENSE AND  MAY
// BE USED OR COPIED ONLY IN ACCORDANCE WITH THE TERMS
// OF SUCH LICENSE.
//
//
// COPYRIGHT (c) 1997 BY DIGITAL EQUIPMENT CORPORATION.
// ALL RIGHTS RESERVED.
//

package dec.trek;
import java.io.*;

//
//
// ExceptionHandler. Processes a method's exception handler info (try/catch's)
//
//
public class ExceptionHandler
{
	private Method parent;
	int startPc;		// Start of try blk
	int endPc;			// End of try blk
	int handlerPc;		// Start of catch/finally blk
	int catchType;		// CP index of exception's data type(eg. Throwable)
						//   or 0 for a Finally

	// Returns exception handler of specified catch/finally stmt
	public static ExceptionHandler find(Statement stmt)
	{
		if (stmt.type!=Trek.ST_CATCH && stmt.type!=Trek.ST_FINALLY)
			throw new TrekException("Statement not a catch or finally statement");
			
		// Scan for it
		ExceptionHandler excepts[] = stmt.parent.excepts;
		for (int i=0; i<excepts.length; i++) {
			ExceptionHandler ex = excepts[i];
			if (stmt.start==ex.handlerPc)
				return ex;
		}									// End loop
		throw new TrekException("Could not find handler for specified statement");
	}

	// Returns catch/finally stmt of exception handler
	public Statement getHandlerStatement()
	{
		Statement st = parent.statementOf(handlerPc, null);
		if (st==null)
			throw new TrekException("Could not find handler statement");
		return st;
	}

	// Returns try stmt of exception handler
	// Notes:
	//	Tricky because try's 0-len and mult can be together,
	//	So use level of handler-stmt to disambiguate
	public Statement getTryStatement()
	{
		Statement hand = getHandlerStatement();
		for (Statement stmt=hand.prev(); stmt!=null; stmt=stmt.prev()) {
			if (stmt.type==Trek.ST_TRY && hand.level==stmt.level)
				return stmt;
		}
		throw new TrekException("Could not find try statement");
	}

	// Read the specified method's exception table
	// chan = where to write to
	// method = obj whose exceptions are to be read
	static void readExcepts(DataInput chan, Method method)
		throws IOException
	{
		int i = -1;

		try {
			int len = chan.readUnsignedShort();		// # of exceptions
			method.excepts = new ExceptionHandler[len];

			// Read each exception entry
			for (i=0; i<len; i++) {
				ExceptionHandler ex = new ExceptionHandler();
				method.excepts[i] = ex;
				ex.parent = method;
				ex.startPc = chan.readUnsignedShort();
				ex.endPc = chan.readUnsignedShort();
				ex.handlerPc = chan.readUnsignedShort();
				ex.catchType = chan.readUnsignedShort();
			}									// End loop
		}										// End try

		// Gen msg for error if any
		catch (Exception e) {
			method.parent.logError(toContext("reading", method, i), e);
		}
		return;
	}

	// Create context string for error during processing of exceptions
	// doing = action being performed at time of exception
	// method = method whose exceptions are being processed
	// i = index of exception when problem occurred
	static private String toContext(String doing, Method meth, int i)
	{
		// Setup content of error and combine with context
		String mname = "M=" + meth.toName() + ";";
		String elem = "E=exception[" + i + "];";
		return "D="+doing+";" + mname + elem;
	}

	// Returns dump presentation string for an exception handler
	public String toString()
	{
		String str;
		if (catchType == 0)
			str = "<finally>";
		else str = parent.parent.constant(catchType).toName();
		return startPc + "-" + endPc + " " + handlerPc + " " + str;
	}

	// Returns data type of exception
	public String toType()
	{
		if (catchType == 0)
			return null;
		TypeRef ref = (TypeRef)parent.parent.constant(catchType);
		return ref.toQualifiedName();
	}

	// Write the current object's exceptions
	// chan = where to write to
	// method = obj whose exceptions are to be written
	static void writeExcepts(DataOutput chan, Method method)
		throws IOException
	{
		int i = -1;

		try {
			int len = method.excepts.length;
			chan.writeShort(len);				// # of exceptions
			if (len <= 0) return;

			// Write each exception entry after insuring PC's uptodate
			for (i=0; i<len; i++) {
				ExceptionHandler ex = method.excepts[i];
				chan.writeShort(Statement.adjustPc(method, ex.startPc));
				chan.writeShort(Statement.adjustPc(method, ex.endPc));
				chan.writeShort(Statement.adjustPc(method, ex.handlerPc));
				chan.writeShort(ex.catchType);
			}									// End loop
		}										// End try

		// Gen msg for error if any
		catch (Exception e) {
			method.parent.logError(toContext("writing", method, i), e);
		}
		return;
	}

}

