//
// THIS SOFTWARE IS FURNISHED UNDER A LICENSE AND  MAY
// BE USED OR COPIED ONLY IN ACCORDANCE WITH THE TERMS
// OF SUCH LICENSE.
//
//
// COPYRIGHT (c) 1997 BY DIGITAL EQUIPMENT CORPORATION.
// ALL RIGHTS RESERVED.
//

package dec.trek;
import java.io.*;
import java.util.Vector;

//
//
// Field. The data member of a class. Also contains static routines that
// manipulate things that all names (incl classes & methods) have in common.
//
//
public class Field extends Member 
{
	int initIdx;					// Index of its initial value
	String initString;

	// Dump a field
	void dump(PrintWriter chan, int flags)
	{
		Trek.printPretty(chan, "    ", toString() + ";");

		// Print its attrs. Except for ConstantValue which is done via toString
		if ((flags & Trek.DUMP_ATTR)!=0)
		{
			for (int j=0; j<attrs.length; j++) {
				Attr attr = attrs[j];
				if (attr.value != null)
					attr.dump(chan, parent, "      ");
			}
		}
	}

	// Create Field object that represents field ref
	Field(ClassFile cf, int nameidx, int descidx)
	{
		parent = cf;
		nameIdx = nameidx;
		descIdx = descidx;
	}

	// Construct field and insert in its classfile
	private Field(ClassFile cf)
	{
		parent = cf;
		idx = -1;
	}

	// Return size of field. Ref fields are considered size of the ptr, not
	// the underlying object that may eventually point to
	public int getSize()
	{
		String desc = toDesc();
		switch (desc.charAt(0)) {
		case 'B':
			return 1;
		case 'C':
		case 'S':
			return 2;
		case '[':
		case 'L':
		case 'I':
		case 'F':
		case 'Z':
			return 4;
		case 'J':
		case 'D':
			return 8;
		}
		throw new TrekException("Data type not recognized");
	}

	// Returns true if this field is an array
	public boolean isArray()
	{
		String desc = toDesc();
		return desc.charAt(0)=='[' ? true : false;
	}

	// Navigate to next field of classfile. If at end, ret null instead
	public Field next()
	{
		if (idx >= parent.fields.size()-1)
			return null;
		Field field = (Field)parent.fields.elementAt(idx + 1);
		return field;
	}

	// Navigate to prev field of classfile. If at start, ret null instead
	public Field prev()
	{
		if (idx <= 0)
			return null;
		Field field = (Field)parent.fields.elementAt(idx - 1);
		return field;
	}

    // Read all the fields from the classfile
	static void readFields(DataInput chan, ClassFile cf) throws IOException
	{
		// Get # of fields and alloc/read-in all of them.
		int len = chan.readShort();
		cf.fields = new Vector(len);	// Get cnt & alloc fields array
		cf.rawFields = new Vector(len);
		for (int i=0; i<len; i++) {
			Field fld = new Field(cf);
			fld.read(chan);

			// Need orig order to avoid possib of illegal forward refs in
			// dumping static initializers (eg. a=b+1; b=0;).
			// Need sorted order for lookups & all other field displays
			cf.rawFields.addElement(fld);
			Trek.insertSorted(cf.fields, fld);
		}
	}

	// Setup init string for specific source dcl
	// src = src of fld's statement
	// name = name of fld
	String toInitString(String src, String name)
	{
		if (initString!=null)
			return initString;

		// Setup initial value if any. Just copy constant value unless fld
		// is boolean. In that case, map it to true/false
		String initval = "";
		if (initIdx > 0)
		{								// Fld has init val
			ValueRef vref = (ValueRef)parent.constant(initIdx);
			if (!src.startsWith("boolean "))
				initval = " = " + vref.toString();
			else {
				Integer iobj = (Integer)vref.getObject();
				if (iobj.intValue() == 0)
					initval = " = false";
				else initval = " = true";
			}
			initString = initval;
			return initval;
		}

		// Do complicated cases by chking for stmt that starts with "name ".
		// Note that field dcl-ed Final must have an initializer clause and
		// arrays do not use initializer clauses even when array is init-ed
		boolean isfinal = hasModifier(Trek.ACC_FINAL);
		boolean initcode = isfinal || toDesc().startsWith("[");
		String namesp = name + " ";
		Statement st;

		// 1) Chk for static init value by finding stmt in clinit.
		if (hasModifier(Trek.ACC_STATIC))
		{
			if (parent.clinit!=null && (parent.isClinitSimple() || initcode))
			{
				// The -1 is trailing semi-colon. The "in-line init" comment 
				// is hack to prevent stmt from going into static {} block
				for (st=parent.clinit.firstStatement(); st!=null; st=st.next()) {
					String str = st.toString();
					if (str.startsWith(namesp))
					{
						initval = str.substring(name.length(), str.length()-1);
						st.string += " // in-line init";
						break;
					}
				}
			}
		}

		// 2) Chk for instance init value by finding stmt in constructor
		else if (initcode)
		{
			Method meth = parent.lookupMethod(parent.toName(), null);
			if (meth!=null)
			{
				// Constructor must start with super-construct() [fld-inits]
				// Can rely on explode of constructor already having been done
				// by time namesp tested because firstStatement is called
				// The -1 is space before =, the -4 is trailing ; */
				namesp = "/* " + namesp;	// Fld inits src is commented out
				for (st=meth.firstStatement(); st!=null; st=st.next()) {
					if (st.type!=Trek.ST_ASSIGN && st.type!=Trek.ST_CALL)
						break;
					String str = st.toString();

					// Found a fld init, capture right side of assign
					if (str.startsWith(namesp))
					{
						initval = str.substring(namesp.length()-1, str.length()-4);
						break;
					}
				}
			}
		}

		// Store and return calculated init string
		initString = initval;
		return initval;
	}

	// Return string that uniquely identifies this fld (see Trek.getObject)
	public String toObjectId()
	{
		// Construct and return qualclass.field
		String fldname = toName();
		return parent.toQualifiedName() + "." + fldname;
	}

	// Return string representation of field: access-flags type name[=val],
	// but arg names themselves are not available, so just use a<n>.
	public String toString() {return toString(null);}

	String toString(String classname)
	{
		// Setup name of field, its type info, and its initial value
		String name = toRef(classname);
		String dcl = parent.toDef(name, toDesc(), 0);
		String initval = toInitString(dcl, name);

        // If field is this$xxx
		// Setup access flags & concat with fld/initval and return.
		// However suppress source code for final field with no init
		// value -- which is how outer class's this is encoded
		String access = Datum.toAccessString(accessFlags, this);
		//if (hasModifier(Trek.ACC_FINAL) && initval.length()==0)
		//	return "/* " + access + dcl + initval + " */";
        //else 
			return access + dcl + initval;
	}

	// Returns ext-form type of the field
	public String toType()
	{
		return parent.toDef(null, toDesc(), 1);
	}

    // Write all the fields to the classfile in orig order (see readFields)
	static void writeFields(DataOutput chan, ClassFile cf) throws IOException
	{
		// Write # of fields and each field
		chan.writeShort(cf.fields.size());
		for (int i=0; i<cf.fields.size(); i++) {
			Field fld = (Field)cf.rawFields.elementAt(i);
			fld.write(chan);
		}
	}

}

