//
// THIS SOFTWARE IS FURNISHED UNDER A LICENSE AND  MAY
// BE USED OR COPIED ONLY IN ACCORDANCE WITH THE TERMS
// OF SUCH LICENSE.
//
//
// COPYRIGHT (c) 1997 BY DIGITAL EQUIPMENT CORPORATION.
// ALL RIGHTS RESERVED.
//
package dec.trek;
import java.util.Vector;

//
// Finish exploding method into its statements.
//
final class Finish 
{
	// Top-level code is level 1. levs[n] is top-stmt of code at level n.
	// Thus levs[0] undefined and levs[1] is dummy-stmt above 1st real stmt.
	private int level;
	private Statement levs[] = new Statement[100];

	private int labelIdx;					// # of manual blocks
	private Method meth;				// Its parent method
	private Statement prim;
	private Vmi vm;						// For executing loop stmts

	// Add a level to open blks, setting new level and data for it.
	private void addBlock(Statement top)
	{
		levs[++level] = top;
	}

	// Just discovered blk, bump level of stmts in it and add end-blk stmt
	// base = stmt at top of block, or end-blk to move down if type -1
	// target = addr where to insert bottom stmt
	// type = type to assign bottom, or -1 to move end-blk down
	// endstr = source code of bottom (ignored if type is -1)
	static void bumpBlock(Statement base, int target,
		int type, String endstr)
	{
		Statement top, bot;
		Statement succ;

		// Move end-blk down or create it
		if (type<0)
		{
			bot = base;
			succ = base.next();			// Must be done before remove
			bot.remove();
			bot.start = target;
			top = bot.mate;
		}
		else {
			top = base;
			succ = base.next();
			bot = new Statement(base.parent, target, type, top.level, endstr);
		}

		// Find end of blk and bump level of all intervening stmts.
		for (; succ!=null; succ=succ.next()) {

			// Handle special case of:
			//   loop {if () stmts1; else {if () stmts2; else contin;} stmts3;}
			// The 2nd If appeared to have Then-blk that went past stmts3
			// during Divide. But since 1st If's end-else is before stmts3,
			// 2nd's If is really an optim continue (or break)
			if (succ.type==Trek.ST_THEN && succ.mate.start>target)
				makeBlockIntoJump(succ);

			// Look for end of bump region. Usually this is easy, it is
			// the 1st actual code past target. It's tricky though if
			// 0-length stmts are at target. In that case, you want to
			// include end of blks whose bodies are in region (even a
			// code-less Then blk since its If is already in). Conversely
			// a free-standing top-level code-less action stmt like a goto
			// is "logically" actual code past target, and should end region.
			if (succ.start==target
				&& (succ.level<bot.level
					|| succ.insts.length>0
					|| (succ.type<Trek.ST_END
						&& succ.type!=Trek.ST_THEN
						&& succ.type!=Trek.ST_CASE
						&& succ.level==bot.level)))
				break;
			succ.level += 1;
		}
		if (succ==null)
			throw new TrekException("Could not find end of block to bump");

		// Position the end-blk stmt and link it up
		bot.insert(succ);
		top.addr = target;
		top.mate = bot;
		bot.mate = top;
	}

	// Set type of unbound jumps, set bottom of switches, and build
	// blks around else's & manual blks for entire method
	private void classify()
	{
		// Scan the method. Note that toString calls execute, which
		// also displays stuff for Db_at
		for (Statement st=meth.firstStatement(); st!=null; st=st.next()) {
			if ((Trek.debug&Trek.Db_atsmall) != 0)
			{
				String str = st.toString();
				Trek.println(System.out, "-->[" + str + "]");
			}

			// Locate end of switch block.
			switch (st.type) {
			case Trek.ST_SWITCH:
				setSwitchBottom(st);

			// Disting labeled goto's by whether goto it comes from
			// same level or an applicable inner block.
			case Trek.ST_THEN:
			case Trek.ST_ELSE:
			case Trek.ST_FINALLY:
			case Trek.ST_SYNC:
			case Trek.ST_TRY:
			case Trek.ST_CATCH:
			case Trek.ST_LABEL:
				addBlock(st);
				break;

			case Trek.ST_DO:
			case Trek.ST_FOR:
			case Trek.ST_WHILE:
				st = classifyLoop(st);
				addBlock(st);
				break;

			case Trek.ST_ENDTHEN:
			case Trek.ST_ENDELSE:
			case Trek.ST_ENDFINALLY:
			case Trek.ST_ENDSYNC:
			case Trek.ST_ENDTRY:
			case Trek.ST_ENDCATCH:
			case Trek.ST_ENDLABEL:
			case Trek.ST_ENDDO:
			case Trek.ST_ENDFOR:
			case Trek.ST_ENDWHILE:
			case Trek.ST_ENDSWITCH:
				levs[level] = null;
				level -= 1;
				break;

			// Only interested in intra-block jumps
			case Trek.ST_UNKNOWN:
				st = classifyJump(st);
			}
		}							// End loop
	}								// End classify

	// Bind an unbound branch to structure of method & set its source code
	// st = stmt that is an unbound branch
	// Returns stmt to continue processing with
	private Statement classifyJump(Statement st)
	{
		boolean inloop = false;
		boolean inswit = false;

		// Chk if physically at jump at end of try/catch blk (ie. there
		// could be a bunch of inner blks that also end at this pt)
		boolean endtc = false;
		for (Statement stn=st.next(); stn!=null; stn=stn.next()) {
			if (stn.start>st.start+st.insts.length)
				break;
			else if (stn.type==Trek.ST_ENDTRY || stn.type==Trek.ST_ENDCATCH)
			{
				endtc = true;
				break;
			}
		}

		// Deopt jumps to end of try/catch chain. Note that this causes
		// jump at end of a try/catch blk to target its End stmt
		Statement resume = fixupTryCatchExit(st);
		if (resume!=st)
			return resume;

		// See which level it goes to
		int target = st.addr;
outer:	for (int lev=st.level; lev>=1; lev--) {
			Statement top = levs[lev];
			Statement bot = top.mate;
			Statement prebot = bot.prev();

			// Loops can be broken or contin.
			switch (bot.type) {
			case Trek.ST_ENDDO:
			case Trek.ST_ENDFOR:
			case Trek.ST_ENDWHILE:

				// Chk if branching past loop. Brk chk must be 1st since
				// jumping to 0-length end-of-loop (ie. start==end) is
				// a Break, not a Continue.
				if (isShortCut(target, bot.start+bot.insts.length))
				{
					st.type = Trek.ST_BREAK;
					st.string = "break" + toLabel(top, inswit|inloop) + ";";
					return st;
				}

				// Nominally Continue goes to blk-end stmt of loop.
				// But optimized Continue can go directly to top of loop,
				// or branch can be to prebot in for-loop (because incr
				// clause not merged with bot until finishLoops).
				if (target==bot.start
					|| target==top.start+top.insts.length
					|| (target==prebot.start && prebot.addr==Statement.ad_incr))
				{
					st.type = Trek.ST_CONTINUE;
					st.string = "continue" + toLabel(top, inloop) + ";";
					return st;
				}
				inloop = true;
				break;

			// Switch can only be broken
			case Trek.ST_ENDSWITCH:
				if (isShortCut(target, bot.start))
				{
					st.type = Trek.ST_BREAK;
					st.string = "break" + toLabel(top, inswit|inloop) + ";";
					return st;
				}
				inswit = true;
				break;

			// If if-target was encoded as dummy goto, maybe this is case of
			// inner blk being optimized to go directly to end of outer Else.
			// It is if inner-target goes beyond outer if's target (ie. there
			// cant be a 0-size outer else) and inner-target goes to within
			// outer-outer-block (eg. it is not a break).
			case Trek.ST_ENDTHEN:
			case Trek.ST_ENDELSE:
				Statement nearbot = bot.type==Trek.ST_ENDTHEN ? prebot : bot.next();
				Statement outer = levs[lev-1];
				if (lev==st.level || st.insts.length>0
					|| outer.type!=Trek.ST_THEN
					|| target<=outer.mate.start || target>levs[lev-2].mate.start)
					break;

				// If inner-if and outer blk both branch to end of outer-outer
				// Else OR outer blk ends with return or throw (and from above
				// we know it goes between outer & outer-outer blk), extend
				// inner-if's Then blk until end of outer blk, and setup
				// to later recognize outer-outer Else blk.
				if ((nearbot.type==Trek.ST_RETURN || nearbot.type==Trek.ST_THROW)
					|| (nearbot.type==Trek.ST_UNKNOWN && nearbot.addr==target))
				{
					Statement topthen = st.prev();
					Statement endthen = st.next();

					// Move end-then down to end of next outermost blk
					// (or the end-then-goto right before it)
					st.remove();
					Statement end;
					for (end=endthen.next(); end!=null; end=end.next()) {
						if (end.type>=Trek.ST_END || end.type==Trek.ST_UNKNOWN)
							break;
					}
					bumpBlock(endthen, end.start, -1, null);

					// If no goto at nearbot, move pseudo-goto down so makeElse
					// will see usual structure of "goto end-else" "end-then"
					// at end of outer Then blk. 
					if (nearbot.type!=Trek.ST_UNKNOWN)
					{
						st.insert(bot);
						st.start = bot.start;
						st.level = bot.level + 1;
					}

					// Resume bind scan with stmt after where psuedo-goto
					// was (skipping any stmts could bypass nested blocks).
					return topthen.next();
				}
			}							// End switch

			// If jump is an opt Contin, or to bot.start and bot is also end of
			// outer blk that is loop, later iter of main loop will process it
			if (target <= st.start)
				continue;
			for (Statement end=bot; end!=null && target==end.start; end=end.next())
			{
				if (end.type>=Trek.ST_ENDLOOP)
					continue outer;
			}

			// If (deopted) jump goes past eob, later iter of main loop 
			// will process it.
			boolean thenbot = st.next().type==Trek.ST_ENDTHEN;
			if (thenbot) target = fixupEndThenTarget(st);
			if (target > bot.start)
				continue;

			// This is branch preceding Else if currently it is followed
			// by then-end and goto doesnt go past end of blk that If is in
			if (thenbot && st.level-1==lev)
				return makeElse(st, target);

			// If de-opted jump at end of try/catch into jump to next inst,
			// then it is simply part of the endtry/catch.
			if (target==st.start+st.insts.length && endtc)
				return mergeWithEnd(st);

			// If noop or jump past end-sync, just make it empty stmt
			Statement targst = meth.statementOf(target, st);
			if (target==st.start+st.insts.length
				|| (top.type==Trek.ST_SYNC
					&& targst.prev().type==Trek.ST_ENDSYNC))
			{
				st.type = Trek.ST_EMPTY;
				st.string = ";";
				return st;
			}

			// Getting past target>bot.start means targst is nominally at
			// level lev. So if targst at nested level, code must be jumping
			// into middle of blk, which is illegal -- unless this is
			// goto at end of try/catch blk
			if (targst.level>lev)
			{
				if (endtc)
					return mergeWithEnd(st);

				// Jump to 1st stmt in blk that has codeless blk-start is 
				// really jumping to the blk-start, which is okay.
				// Else have error, report it by insuring method wont compile.
				targst = targst.prev();
				if (targst.insts.length>0 || targst.level>lev)
				{
					st.type = Trek.ST_EMPTY;
					st.string = "Illegal jump into middle of a block;;";
					return st;
				}
			}

			// Must be break out of "manually inserted" blk.
			// If encompassing blk is switch, make blk within curr case
			if (bot.type==Trek.ST_ENDSWITCH)
				for (top=st; top.type!=Trek.ST_CASE; top=top.prev()) ;
			makeLabeledBlock(st, top);
			return st;
		}							// End loop

		// Jump was not bound. Show place of error in source code
		st.type = Trek.ST_EMPTY;
		st.string = "Jump could not be equated to legal source statement;;";
		return st;
	}								// End classifyJump

	// Classify the current loop
	// top = top of loop to process
	// Returns stmt to continue processing with
	private Statement classifyLoop(Statement top)
	{
		// See if there is assigment stmt that might be loop's incr clause.
		// goesTo() prevents assign stmt that a Continue branches
		// after from being treated as incr clause.
		Statement bot = top.mate;
		Statement prebot = bot.prev();
		String incvar = null;
		if (prebot.type == Trek.ST_ASSIGN && !goesToLoose(top, bot))
		{
			vm.execute(prebot);

			// If source is constant, very unlikely intended as incr
			int ch = vm.ropd.charAt(0);
			if (ch!='\"' && ch!='\'')
			{
				try {
					Integer.parseInt(vm.ropd);
				}
				catch (Exception e) {
					incvar = vm.lopd;
				}
			}
		}

		// If loop-start preceded by asmt stmt, chk if init clause.
		// goesTo() prevents b=1 in "if a=1 else b=1 loop-start"
		// from being treated as loop initializer
		Statement pretop = top.prev();
		String initvar = null;
		String initval = null;
		if (pretop!=null && pretop.type==Trek.ST_ASSIGN
			&& !goesTo(levs[top.level], top))
		{
			vm.execute(pretop);		// The initializer

			// Array initial=value clause cant be in loop initializer
			if (vm.ropd.lastIndexOf("}")<0)
			{
				initvar = vm.lopd;
				initval = vm.ropd;
			}
		}

		// pretop or prebot could be random assignment stmt. Chk if targets
		// of init stmt and incr stmt are the same variable. If not, neither
		// is definitely the winner... Use heuristic that incvar
		// wins unless it is array element and initvar isnt.
		if (initvar!=null && incvar!=null && !initvar.equals(incvar))
		{
			if (initvar.lastIndexOf(']')<0 && incvar.lastIndexOf(']')>=0)
				incvar = null;
			else initvar = null;
		}

		// Set type of loop.
		// Cant be do...while unless bottom of loop contains if-test.
		// If both init & incr, gen for(init; ; incr). If incr that is
		// target of continue, gen for(;; incr). Else gen while(true).
		// (Note: a loop-bottom can be codeless if loop bottom is
		// unreachable -- eg. has return at bottom)
		Instruction inst = bot.firstInstruction();
		if (inst==null || inst.type==Vmi.it_branch)
		{
			if (incvar!=null && (initvar!=null || goesTo(top, prebot)))
				bot.type = Trek.ST_ENDFOR;
			else bot.type = Trek.ST_ENDWHILE;
		}

		// There is if-test at loop-bottom. If goto at top, this is not Do.
		// If there is init or incr, gen a For stmt. Else gen a While stmt.
		else if (top.insts.length>0)
		{
			if (initvar!=null || incvar!=null)
				bot.type = Trek.ST_ENDFOR;
			else bot.type = Trek.ST_ENDWHILE;
		}

		// Is if-test but no goto-down. Nominally this is:
		// do {...} while ();  But it can also be:
		// 1) An optim-away goto after end-then (see addInvisLoopStart),
		// so either while(xxx) or for(; xxx; incr)
		// 2) for (...; var opr constant; ...) whose 1st iter
		// evals to true at compile-time. Iow, the compiler could tell
		// that the if-test was not needed before the loop's 1st iteration.
		// 3) do {empty} while ();, which is equiv to: while () empty;. So
		// classify it as latter because it is more readable.
		else {
			if (bot.type==Trek.ST_ENDWHILE)
			{							// Optim-away goto-down case
				if (incvar!=null)
					bot.type = Trek.ST_ENDFOR;
			}
			else if (initvar!=null && firstIterTrue(initvar, initval, bot))
				bot.type = Trek.ST_ENDFOR;
			else if (top.start==bot.start)
				bot.type = Trek.ST_ENDWHILE;
			else bot.type = Trek.ST_ENDDO;
		}

		// Build the source code assoc with insts at end of loop. Note
		// that for For/While this source code is placed at top of loop
		switch (bot.type) {
		case Trek.ST_ENDDO:
			top.type = Trek.ST_DO;
			top.string = "do {";
			break;

		// Process while-loop.
		case Trek.ST_ENDWHILE:
			top.type = Trek.ST_WHILE;
			bot.string = "} // End while";
			break;

		// Process for-loop, prebuilding merged blk bndary stmts if applic
		// Indic pretop will become init clause of For, if applic
		// Indic prebot will become incr clasue of For, if applic
		case Trek.ST_ENDFOR:
			top.type = Trek.ST_FOR;
			if (initvar != null)
			{
				pretop.addr = Statement.ad_init;
				int forlen = pretop.insts.length + top.insts.length;
				pretop.mate = new Statement(meth, pretop.start,
					forlen, prim.insts, Trek.ST_FOR, top.level, null);
			}
			if (incvar != null)
			{
				prebot.addr = Statement.ad_incr;
				int forlen = prebot.insts.length + bot.insts.length;
				prebot.mate = new Statement(meth, prebot.start,
					forlen, prim.insts, Trek.ST_ENDFOR, top.level, null);
			}
		}							// End switch
		return top;
	}								// End classifyLoop

	// Do 2nd phase of exploding method's primordial stmt into its real stmts
	// method = method being finished
	// primord = single stmt object containing all its code
	//
	// Current state of method-contents is:
	// 1) dummy block-start stmt
	// 2) Divided stmts, with jumps not yet bound to source types and loops not
	//    fully classified
	// 3) dummy block-end stmt
	// Final state is dummy start/end removed, jumps bound, loops classified,
	// and { and }'s of single-statement then/else blocks removed
	Finish(Method method, Statement primord)
	{
		meth = method;
		prim = primord;
		vm = new Vmi();
		if ((Trek.debug&Trek.Db_at) != 0)
			Trek.println(System.out, "Finishing " + meth.toName());

		// Do stuff that was postponed until full struct of method seen,
		// and get rid of dummy start/end for top block (created by Divide).
		try {
			classify();
			new Lvt().declareLocals(meth);
			finishLoops();
			meth.first.remove();
			meth.last.remove();
			trim();
		}
		catch (Exception e) {
			meth.logMethodError("finishing", e);
		}
	}									// End of Finish

	// Return merged stmt (with pre's source code imbedded)
	private Statement finishForEnd(Statement pre, Statement forend)
	{
		// Get source code of pre-stmt (the reason kept sep til now).
		// Set level in case loop blk has been bumped in interim
		Statement new_fe = pre.mate;
		new_fe.string = vm.execute(pre);
		new_fe.level = forend.level;

		// Replace pre & forend with new for-end stmt
		forend.remove();
		new_fe.insert(pre);
		pre.remove();
		return new_fe;
	}

	// Finish loops. In partic, generate source code now that locals
	// have been declared. Also for For loops, combine
	// pretop/prebot with top/bot of each loop
	private void finishLoops()
	{
		// Non-null top.string only if label was created for loop
		for (Statement top=meth.firstStatement(); top!=null; top=top.next()) {
			String ifstr;
			String label = top.string;
			Statement bot = top.mate;

			// Do work if at loop
			switch (top.type) {
			case Trek.ST_DO:
				ifstr = vm.execute(bot);
				bot.string = "} while (" + ifstr + ");";
				break;

			case Trek.ST_WHILE:
				ifstr = vm.execute(bot);
				if (ifstr.length()==0) ifstr = "true";
				top.string = "while (" + ifstr + ") {";
				if (label!=null) top.string = label + top.string;
				break;

			case Trek.ST_FOR:
				// Setup bndary info of loop.
				Statement pretop = top.prev();
				Statement prebot = bot.prev();
				ifstr = vm.execute(bot);

				// If For has init-clause, need to incl assignment stmt in it now
				// 1) If no goto at top, simply convert assign to For-stmt
				//    and get rid of the empty placeholder loop-start
				// 2) Else build stmt that contains both assign and loop-start's
				//    goto, and remove both old stmts
				String initstr;
				if (pretop!=null && pretop.addr==Statement.ad_init)
				{
					top = finishForEnd(pretop, top);
					initstr = top.string;
				}
				else initstr = ";";

				// If there is incr stmt, merge prev stmt with loop bottom and
				// build a for-stmt that has an incr clause.
				String incstr;
				if (prebot!=null && prebot.addr==Statement.ad_incr)
				{
					bot = finishForEnd(prebot, bot);
					incstr = bot.string;
					incstr = incstr.substring(0, incstr.length()-1);
				}
				else incstr = "";

				// Setup source and (because of removes/inserts) rebind top to bottom
				top.string =
					"for (" + initstr + " " + ifstr + "; " + incstr + ") {";
				bot.string = "} // End for";
				top.mate = bot;
				bot.mate = top;
				if (label!=null) top.string = label + top.string;
			}								// End switch
		}									// End loop
	}										// End finishLoops

	// Returns true if 1st iter guaran true
	// init* = loop var & its init val
	// if* = opds & opr of loop's ifstr
	private boolean firstIterTrue(String initvar, String initval, Statement bot)
	{
		// Get info for loop's test-clause
		vm.execute(bot);
		String iflopd = vm.lopd;
		int ifopr = vm.opr;
		String ifropd = vm.ropd;

		// Could compare initval to ifropd too.
		// But people who code backwards deserve what they get
		if (ifopr>=Term.op_assign)
			return false;			// Not relational operator
		if (!initvar.equals(iflopd))
			return false;			// No way to chk
		
		// Need numeric constants to simulate compile-time chk
		int ival, tval;
		try {
			ival = Integer.parseInt(initval);
			tval = Integer.parseInt(ifropd);
		}
		catch (Exception e) {
			return false;			// Either term is not numeric
		}

		// Finally do the comparison
		switch (ifopr) {
		case Trek.DOIF_EQ:	return ival==tval;
		case Trek.DOIF_NE:	return ival!=tval;
		case Trek.DOIF_GE:	return ival>=tval;
		case Trek.DOIF_LE:	return ival<=tval;
		case Trek.DOIF_GT:	return ival>tval;
		case Trek.DOIF_LT:	return ival<tval;
		}
		return false;
	}										// End firstIterTrue

	// Returns canonical value of st's target:
	// May need to deopt st.addr if it is within inner If and branches
	// between end of outer-if and end of the blk encompassing that.
	// st = goto stmt to process
	// Creates pseudo-goto stmt if necessary.
	private int fixupEndThenTarget(Statement st)
	{
		// If this is code-less goto, insufficient nesting, other than
		// Then, or interior to outer blk, then no work.
		if (st.insts.length==0) return st.addr;
		if (st.level<=2) return st.addr;		// At most 1-level of If
		Statement bot = levs[st.level-1].mate;
		if (bot.type!=Trek.ST_ENDTHEN) return st.addr;
		Statement bot2 = levs[st.level-2].mate;
		if (st.addr<=bot.start) return st.addr;

		// If end of inner-then branches to end of outer Else, then set
		// its target to its canon value -- end of inner Else
		Statement prebot = bot.prev();	// Last stmt in outer Then
		if (prebot.type==Trek.ST_UNKNOWN && prebot.addr==st.addr)
			return prebot.start;

		Statement postbot = bot.next();	// Last stmt in outer Then
		if (postbot.type==Trek.ST_UNKNOWN && postbot.addr==st.addr)
			return postbot.start;

		// Outer Then doesnt end with goto because inner Else to-be ends in
		// return, throw, brk, or contin. Insert pseudo-goto to st.addr
		// there so makeElse will see usual structure of "goto past-else-blk"
		// "end-then" at end of outer Then blk.
		if (st.addr>=bot2.start) return st.addr;
		if (prebot.type==Trek.ST_RETURN || prebot.type==Trek.ST_THROW
			|| (prebot.type==Trek.ST_UNKNOWN && prebot.addr>st.addr))
		{
			Statement pseudo = new Statement(meth, bot.start, Trek.ST_UNKNOWN,
				prebot.level, null);
			pseudo.insert(bot);
			pseudo.addr = st.addr;
			return bot.start;
		}
		return st.addr;
	}

	// If jump in try or catch blk goes just past all handlers (or to
	// common jsr after last Catch and before Finally),
	// de-opt st.addr to bottom of its own try/catch blk.
	// Return st unless If/codeless-goto special case applies.
	// In that case, return its successsor because st destroyed.
	// st = jump stmt to chk
	private Statement fixupTryCatchExit(Statement st)
	{
		// Scan for st being in a try/catch blk
		for (int lev=st.level; lev>=1; lev--) {
			Statement top = levs[lev];
			Statement bot = top.mate;
			if (st.addr<=bot.start)
				return st;
			if (top.type!=Trek.ST_TRY && top.type!=Trek.ST_CATCH)
				continue;

			// It is. Find end of this t/c chain.
			Statement postbot = bot.next();
			for (; postbot!=null; postbot=postbot.mate.next()) {
				if (postbot.type!=Trek.ST_CATCH	&& postbot.type!=Trek.ST_FINALLY)
					break;
			}


			// 1) If st jumps further down than bottom of t/c chain (unless
			// it is to jsr of goto-jsr-goto of outer t/c chain), then no deopt
			if (postbot!=null)
			{
				for (Statement coded=postbot; coded!=null; coded=coded.next()) {
					int stlen = coded.insts.length;
					if (stlen==0) continue;
					int pastpost = coded.start + stlen;
					if (coded.type==Trek.ST_ENDSYNC && st.addr==pastpost)
						break;				// Jumps to end of Sync blk
					if (st.addr>=pastpost)
						return st;			// Past 1st non-empty post stmt
					else break;				// At or to midde of it
				}

				// 2) If there is code-less loop bottom there (meaning jump 
				// is break out of this loop), then no deopt
				if (postbot.type>=Trek.ST_ENDLOOP && postbot.insts.length==0)
					return st;
			}

			// Handle special case that if-inst was opt to exit
			// catch's and so Divide put out code-less goto.
			Statement then = st.prev();
			if (st.insts.length==0 && then.type==Trek.ST_THEN)
			{
				// Get rid of spurious code-less goto, find real
				// end of then. 2 cases: past-then jumps to same
				// loc as If or does not. For former, exclude jump
				// from expanded Then blk.
				st.remove();
				int thenend;
				Statement eob = then.getCodeEob();
				Statement peob = eob.prev();
				if (peob.type==Trek.ST_UNKNOWN && st.addr==peob.addr)
					thenend = peob.start;
				else thenend = eob.start;
				then.prev().addr = thenend;
				bumpBlock(then.next(), thenend, -1, null);
				return then.next();
			}

			// The usual case, just deopt
			st.addr = bot.start;
			break;
		}
		return st;
	}

	// Returns true if stmt between (top, bot) goes to bot
	private boolean goesTo(Statement top, Statement bot)
	{
		return top.goesBetween(bot.start, bot.start);
	}

	// If code between top & bottom of loop either goes to bottom or is
	// optimized to go to 1st stmt of loop, return true.
	private boolean goesToLoose(Statement top, Statement bot)
	{
		Statement ntop = top.next();

		// Scan for Continue
		for (Statement st=ntop; ; st=st.next()) {
			if (st==null || st.start>=bot.start)
				return false;
			if (st.type==Trek.ST_UNKNOWN || st.type==Trek.ST_IF)
			{
				int target = st.addr;		// Get goto's target
				if (target==bot.start)
					return true;

				// For optimized Contin, exclude goto top of nested loop
				// that starts at same pc as top's loop
				if (target!=top.start+top.insts.length)
					continue;				// Not opt Contin
				if (st.level<=top.level+1
					|| (ntop.type!=Trek.ST_FOR
						&& ntop.type!=Trek.ST_WHILE
						&& ntop.type!=Trek.ST_DO))
					return true;
			}
		}
	}

	// Returns true if target is along optimize-chain of bottom
	private boolean isShortCut(int target, int bottom)
	{
		if (target==bottom)
			return true;
		else return false;
	}

	// Turn bogus Then blk into a code-less goto that branches
	// to if-inst's target and decr level of stmts that were in it.
	// For ex: if () then {xxx;} --> if () {code-less jump} xxx;
	static void makeBlockIntoJump(Statement then)
	{
		// Bump level of stmts that used to be in blk
		Statement body1 = then.next();
		Statement endthen = then.mate;
		Statement st;
		for (st=body1; st!=endthen; st=st.next()) st.level -= 1;

		// Move endthen up next to then
		endthen.remove();
		endthen.insert(body1);
		endthen.start = then.start;

		// Insert code-less jump and update target addresses
		Statement brkconn = new Statement(
			then.parent, then.start, Trek.ST_UNKNOWN, then.level+1, null);
		brkconn.insert(endthen);
		Statement ifst = then.prev();
		brkconn.addr = ifst.addr;
		ifst.addr = brkconn.start;	// Since Then blk 0-length
	}

	// Re-end the Then blk and add start and end of Else blk
	// stmt = the unbound goto-stmt being converted to Else stmt
	// target = where the branch at end of Then goes to
	// Returns the last processed statement after all is said and done.
	// 1) In normal case, this is stmt prev to end-then (iow, the goto was
	//	  prev to end-then when processing started)
	// 2) In fix-divide case, it is then-stmt because goto deleted.
	// Notes:
	//	stmt will be a code-less goto when Then ends with return or throw.
	private Statement makeElse(Statement stmt, int target)
	{
		if (stmt.type!=Trek.ST_UNKNOWN)
			throw new TrekException("Bad End-then/Else statement pattern");

		// If goto-inst first (and thus only) code in Then blk, 2 cases:
		Statement endthen = stmt.next();
		if (stmt.prev().type==Trek.ST_THEN)
		{
			// If code-less goto only code in Then blk, then Divide falsely
			// categorized if-inst as going directly to break/contin. Fix now
			// by deleting code-less goto and moving end-then down to target.
			if (stmt.insts.length==0)
			{
				Statement then = stmt.prev();
				stmt.remove();			// The code-less goto
				bumpBlock(endthen, target, -1, null);
				return then;
			}

			// If real goto-inst only code in Then blk, need an empty stmt.
			else {
				Statement st = new Statement(meth, stmt.start,
					Trek.ST_EMPTY, stmt.level, ";");
				st.insert(stmt);
			}
		}

		// Move then-end put after unbound goto to before it (initially
		// it was thought to be break/contin -- now know it is Else).
		// Finally convert unbound goto into else's start.
		endthen.remove();
		endthen.insert(stmt);
		endthen.start = stmt.start;
		endthen.string = "}";
		stmt.string = "else {";
		stmt.level = endthen.level;
		stmt.type = Trek.ST_ELSE;

		// Bump level of stmts in the else block, and insert blk-end stmt
		bumpBlock(stmt, target, Trek.ST_ENDELSE, "} // End if");
		return endthen.prev();
	}									// End makeElse

	// Build start and end of manual block if havent yet.
	// brk = jumping stmt that necessitated creating labeled blk
	// top = top of block encompassing brk
	// Returns label of block
	private void makeLabeledBlock(Statement brk, Statement top)
	{
		String label;
		int target = brk.addr;			// Put bottom of labeled blk here

		// If this block already exists, return its label
		if (top.type == Trek.ST_LABEL && top.mate.start==target)
		{
			String str = top.string;
			label = str.substring(0, str.length()-3); // Strip ": {"
		}

		// New block, build its start and stmts and bump level of contents
		else {
			Statement succ = brk.getInsertPoint(top.level+1);
			label = "label" + labelIdx;
			labelIdx += 1;
			top = new Statement(meth, succ.start,
				Trek.ST_LABEL, succ.level, label + ": {");
			top.insert(succ);
			bumpBlock(top, target, Trek.ST_ENDLABEL, "} // End label");

			// Create and move this block to its proper level
			addBlock(top);
			int blklev = top.level + 1;
			int delta = level - blklev;
			if (delta > 0)
			{
				System.arraycopy(levs, blklev, levs, blklev+1, delta);
				levs[blklev] = top;
			}
		}

		// Create source code of break stmt
		brk.type = Trek.ST_BREAK;
		brk.string = "break " + label + ";";
	}

	// Merge st with code-less End stmt of try/catch following it.
	// Returns st.prev since want orig st.next to be next classified
	// Notes:
	//	There may be code-less stmts between st and this End.
	//	Slave of classifyJump... only called if endtc set
	// 
	private Statement mergeWithEnd(Statement st)
	{
		for (Statement stn=st.next(); stn!=null; stn=stn.next()) {
			if (stn.start>st.start+st.insts.length)
				break;					// Inconsistency
			stn.start = st.start;		// "Move" codeless stuff above merge

			// If at the end-try/catch known to follow st, merge st
			if (stn.type==Trek.ST_ENDTRY || stn.type==Trek.ST_ENDCATCH)
			{
				Statement stp = st.prev();
				st.remove();
				stn.insts = st.insts;
				return stp;
			}
		}
		throw new TrekException("A call to mergeWithEnd failed");
	}

	// Finish building bottom of switch block
	// top = switch stmt
	private void setSwitchBottom(Statement top)
	{
		if (top.addr==Statement.ad_bound) return;	// Bottom already setup

		// Scan cases looking for a goto that could be "break;".
		// To be this, its target must be start of last case or gtr. Look
		// for such a target till hit empty case(s) at bottom of switch
		Statement st;
		int best = -1;
		for (st=top.next(); st.start<top.addr||st.type!=Trek.ST_CASE; st=st.next())
		{
			if (st.type==Trek.ST_UNKNOWN && st.addr >= top.addr
				&& (best<0 || st.addr<best))
				best = st.addr;
		}
		Statement hicase1 = st;


		// It must also be before the end of the nearest encompassing blk.
		// If no such best, this means switch logically extends to end of
		// blk. But this can be unreadable if it makes last case too big.
		// So alternately, pull last case out of switch using arbitrary
		// heuristic of limiting last case to 100 inst bytes.
		int outerbot = levs[top.level].mate.start;
		if (best<0 || best>outerbot)
		{
			if (outerbot - top.addr < 100)
				best = outerbot;
			else best = -1;
		}

		// If couldnt find break or break goes to start of empty cases,
		// leave end of switch where it is.
		if (best<0 || best==top.addr)
		{
			for (st=hicase1; st.type!=Trek.ST_ENDSWITCH; st=st.next()) {

				// Get rid of empty cases there if they include default:.
				// Otherwise removing cases would add cases to the default path.
				Statement stn;
				if (st.string.equals("default:"))
				{
					// Anything besides case labels is a bug, so catch this
					for (st=hicase1; st.type!=Trek.ST_ENDSWITCH; st=stn) {
						if (st.type!=Trek.ST_CASE)
							throw new TrekException
								("Other statement mixed in with empty cases");
						stn = st.next();
						st.remove();
					}
					break;
				}
			}
		}

		// Found a break to end of switch, move switch end down to that pt.
		// Iow, insert stmts between last case and "best" in a case blk.
		else {
			Statement endsw;
			for (endsw=hicase1; ; endsw=endsw.next()) {
				if (endsw.type==Trek.ST_ENDSWITCH && endsw.level==top.level)
					break;
			}

			// Remove source-superflous return at end of method.
			// Either way, insert identified stmts in last case
			Statement retchk = meth.lastStatement();
			if (best>=retchk.start)
			{
				retchk = retchk.prev();
				if (retchk.type==Trek.ST_RETURN && retchk.insts.length==1)
					best = retchk.start;
			}
			bumpBlock(endsw, best, -1, null);
		}
		top.addr = Statement.ad_bound;	// Bottom now set
	}									// End setSwitchBottom

	// Create string for break/continue label & label top if necessary
	// top = stmt at top of blk that is being branched out of
	// outer = true if doing label for outer blk
	private String toLabel(Statement top, boolean outer)
	{
		// Do unlabeled case 1st
		if (!outer)
			return "";

		// Return a label, but 1st insure top of blk has the label.
		// However, source code for top of blk hasnt been created yet.
		String label = "lev" + top.level;
		if (top.type==Trek.ST_SWITCH)
		{
			top.toString();
			if (!top.string.startsWith("lev"))
				top.string = label + ": " + top.string;
		}

		// Loop src code cant be created until finishLoops, so it has to take
		// care of label merging there. Conversely if top.string non-null, an
		// earlier break must have set it to the label already.
		else if (top.string==null)
			top.string = label + ": ";
		return " " + label;
	}

	// Elim unneeded noop stmts (ie. {}'s for single stmt then/else's)
	// Also set stmt number of each stmt
	private void trim()
	{
		Statement next1, next2;

		// Scan the method's stmts
		int num = 1;
		for (Statement stmt=meth.firstStatement(); ; stmt=next1) {
			stmt.mate = null;
			stmt.delta = 0;
			stmt.addr = num++;
			next1 = stmt.next();
			if (next1==null) break;

			// Chk if at then/else block (and not at method-end)
			if (stmt.type!=Trek.ST_THEN && stmt.type!=Trek.ST_ELSE)
				continue;
			next2 = next1.next();
			if (next2==null)
				continue;

			// If at "then {one-stmt;}", remove "then {" and "}" stmts
			if (next2.type==Trek.ST_ENDTHEN)
			{
				next1.addr = stmt.addr;
				next1 = next2.next();
				stmt.remove();
				next2.remove();
			}

			// Replace "else {" with "else", and remove "}"
			else if (next2.type==Trek.ST_ENDELSE)
			{
				next1.addr = num++;
				next1 = next2.next();
				stmt.string = "else";
				next2.remove();
			}
		}								// End of loop
	}									// End trim

}

