package dec.trek;
import java.io.*;

//
//
// Linum -- line number table for a method
//
//
final class Linum 
{
	// The fields in a classfile
	int startPc;				// Where usage of this line starts
	int lineNumber;				// Source line number assoc with this startPc

	// Construct a line number table entry
	// startpc = where this line numbers exists in its method
	// linum = its line number within its overall class file
	// idx = entry index
	// meth = line's method
	Linum(int startpc, int linum, int idx, Method meth)
	{
		startPc = startpc;
		lineNumber = linum;
		meth.linums[idx] = this;
	}

	// Returns number of bytes in local-vars table
	static int attrLength(Method method)
	{
		return 2 + method.linums.length*4;
	}

	// Read the specified method's line-number table
	// chan = where to read from
	// method = obj whose line nums are to be read
	static void readLinums(DataInput chan, Method method)
		throws IOException
	{
		int i = -1;

		try {
			int len = chan.readUnsignedShort();		// # of entries in table
			method.linums = new Linum[len];

			// Read each line's entry
			for (i=0; i<len; i++) {
				new Linum(
					chan.readUnsignedShort(),		// Start pc
					chan.readUnsignedShort(),		// Line number
					i, method);						// Entry to set
			}									// End loop
		}										// End try

		// Gen msg for error if any
		catch (Exception e) {
			method.parent.logError(toContext("reading", method, i), e);
		}
		return;
	}

	// Create context string for error during processing of line nums
	// doing = action being performed at time of exception
	// method = method whose exceptions are being processed
	// i = index of exception when problem occurred
	static private String toContext(String doing, Method meth, int i)
	{
		// Setup content of error and combine with context
		String mname = "M=" + meth.toName() + ";";
		String elem = "E=line[" + i + "];";
		return "D="+doing+";" + mname + elem;
	}

	// Write the current object's line-num table
	// chan = where to write to
	// method = obj whose lines are to be written
	static void writeLinums(DataOutput chan, Method method)
		throws IOException
	{
		int i = -1;

		try {
			int len = method.linums.length;
			chan.writeInt(2 + len*4);			// Size of this attr
			chan.writeShort(len);				// # of entries

			// Write each line-num entry after insuring PC's uptodate
			for (i=0; i<len; i++) {
				Linum linum = method.linums[i];

				// Write out each field
				int startpc = Statement.adjustPc(method, linum.startPc);
				chan.writeShort(startpc);
				chan.writeShort(linum.lineNumber);
			}									// End loop
		}										// End try

		// Gen msg for error if any
		catch (Exception e) {
			method.parent.logError(toContext("writing", method, i), e);
		}
		return;
	}

}

