//
// THIS SOFTWARE IS FURNISHED UNDER A LICENSE AND  MAY
// BE USED OR COPIED ONLY IN ACCORDANCE WITH THE TERMS
// OF SUCH LICENSE.
//
//
// COPYRIGHT (c) 1997 BY DIGITAL EQUIPMENT CORPORATION.
// ALL RIGHTS RESERVED.
//

package dec.trek;
import java.io.*;

// Abstract super class for method and field refs

public abstract class MemberRef extends ThingRef
{
	// Member references are identified by their class and nat info (ie.
	// name and data type info). These are represented as other cp entries.
	protected int classIdx;			// Cp entry index of ref's class
	protected int natIdx;			// Cp entry index of ref's name-and-type

	// Create ref constant
	MemberRef(int clidx, int natidx, ClassFile cf)
	{
		super(cf);
		classIdx = clidx;
		natIdx = natidx;
		cf.addConstant(this);
	}

	MemberRef(ClassFile cf)
	{
		super(cf);
	}

	// Add a method or field ref, if it doesnt already exist.
	// Either way, return a matching pool entry.
	// tag = Cp_field/methodref
	// name = member's name
	// desc = member's descriptor
	// cf = classfile want to know if ref is defined within
	protected static MemberRef add(int tag, String name, String desc, ClassFile cf)
	{
		MemberRef rc = get(tag, name, desc, cf);
		if (rc!=null) return rc;

		// Separate member and class names
		int sepidx = name.lastIndexOf('/');
		String classname = name.substring(0, sepidx);
		name = name.substring(sepidx+1);

		// Add subsid elements as needed and then member ref
		TypeRef tref = TypeRef.add(classname, cf);
		NameAndType nat = NameAndType.add(name, desc, cf);
		if (tag==Cp_fieldref)
			return new FieldRef(tref.idx, nat.idx, cf);
		else return new MethodRef(false, tref.idx, nat.idx, cf);
	}

	// Add a method or field ref, if it doesnt already exist.
	// Either way, return a matching pool entry.
	// mem = the member need ref for
	// cf = classfile want to know if ref is defined within
	static MemberRef add(Member mem, ClassFile cf)
	{
		String qname = mem.parent.toInternalName() + "/" + mem.toName();
		int tag = mem instanceof Field ? Cp_fieldref : Cp_methodref;
		return add(tag, qname, mem.toSignature(), cf);
	}

	// Add a method or field ref, if it doesnt already exist.
	// Either way, return a matching pool entry.
	// mref = the ref (from possibly another classfile) to chk/add
	// cf = classfile want to know if ref is defined within
	static MemberRef add(MemberRef mref, ClassFile cf)
	{
		String qname = mref.toInternalClassName() + "/" + mref.toName();
		int tag = mref instanceof FieldRef ? Cp_fieldref : Cp_methodref;
		return add(tag, qname, mref.toSignature(), cf);
	}

	// Scan for matching ref entry
	// tag == what to search for (method or field ref)
	// name = its fully qualified internal name
	// desc = its internal data-type info
	// cf = classfile to search
	static MemberRef get(int tag, String name, String desc, ClassFile cf)
	{
		return (MemberRef)CpInfo.get((char)tag + name + ":" + desc, cf);
	}

	// Scan for constant that is ref to this member
	// mem = member obj to search for
	// cf = classfile to search
	static MemberRef get(Member mem, ClassFile cf)
	{
		// Do the lookup of qualified mem name + its full descriptor
		int tag = mem instanceof Field ? Cp_fieldref : Cp_methodref;
		String qname = mem.parent.toInternalName() + "/" + mem.toName();
		return get(tag, qname, mem.toSignature(), cf);
	}

	// Return member reffed by this pool entry, if accessible
	// dummy = so signature of MemberRef dif than MethodRef/FieldRef
	Member getObject(int dummy)
	{
		// Get to classfile of the reffed object
		ClassFile refcf = parent;
		String classname = toInternalClassName();
		ClassFile cf = refcf.parent.getClassFile(classname);

		// Now look in fields or methods, depending on whether inst was call
		Member mem;
		if (this instanceof MethodRef)
			mem = cf.lookupMethod(toName(), toSignature());
		else mem = (Member)Trek.lookup(cf.fields, toName());
		return mem;
	}

	// Returns the TypeRef of the class pointed to by this member ref
	public TypeRef getTypeRef()
	{
		return (TypeRef)parent.constant(classIdx);
	}

    // Return unqualified external class name of member ref
	String toClassName()
	{
		return getTypeRef().toName();
	}

	// Returns data type of pool entry
	// Notes:
	//	For fields, sig & desc mean same thing.
	//	For methods, its retval+arglist vs. just retval
	String toDesc()
	{
		String desc = parent.constant(natIdx).toDesc();
		if (this instanceof FieldRef)
			return desc;
		else return desc.substring(desc.indexOf(')') + 1);
	}

    // Return internal class name of member ref
	String toInternalClassName()
	{
		return getTypeRef().toInternalString();
	}

	// Return string(s) associated with specified constant pool entry.
	// Differs from toString in that member refs contain raw desc and /'s
	// rather than source-code signature and .'s.
	String toInternalString()
	{
		return toInternalClassName() + "/" +
			parent.constant(natIdx).toInternalString();
	}

    // Returns simple external name of method
	public String toName()
	{
		String name = parent.constant(natIdx).toName();
		if (name.startsWith("<init>"))
			return toClassName();
		else return name;				// Method not a constructor
	}

    // Return unqualified external class name of member ref
	String toQualifiedClassName()
	{
		return getTypeRef().toQualifiedName();
	}

    // Returns source-code string assoc with this entry.
	String toRef()
	{
		String name = parent.constant(natIdx).toName();
		if (name.startsWith("<init>"))
			return "new " + getTypeRef().toString();
		else return name;
	}

	// Returns signature of pool entry
	// Notes:
	//	For fields, sig & desc mean same thing.
	//	For methods, its retval+arglist vs. just retval
	String toSignature()
	{
		return parent.constant(natIdx).toDesc();
	}

    // Return ref to static name. If ref to/from same class, just return name.
	// Else return the name's class, qualified as neces
	String toStaticRef()
	{
		String name = parent.constant(natIdx).toName();
		if (parent.thisClass == classIdx
			|| parent.superClass == classIdx)
			return name;
		String classname = getTypeRef().toString();
		return classname + "." + name;
	}

	// Write type of constant and its class and data type info
	protected void write(DataOutput chan) throws IOException
	{
		chan.writeShort(classIdx);
		chan.writeShort(natIdx);
	}

}

