//
// THIS SOFTWARE IS FURNISHED UNDER A LICENSE AND  MAY
// BE USED OR COPIED ONLY IN ACCORDANCE WITH THE TERMS
// OF SUCH LICENSE.
//
//
// COPYRIGHT (c) 1997,1998 BY DIGITAL EQUIPMENT CORPORATION.
// ALL RIGHTS RESERVED.
//
package dec.trek;
import java.io.*;
import java.util.Date;

//
//
// RunLog -- parent class for instrumentation classes
// There are 2 models of instrumentation flow:
// 1) There is a method that the instrumented program must call first,
//		and presumably a method that does post-processing after all
//		calls have occurred. (Profile is like this)
// 2) Any inserted code might be the 1st call. (Watch is like this).
//
//
public class RunLog implements RunLogSink
{
	// Should be set non-0 after log system has been initialized. After
	// it is set, instrumentation calls in jdk become active.
	public static int sysInited;

	// If your log system uses any instrumented methods -- in particular
	// if you instrument any jdk methods, you must prevent infinite recursion
	// by returning if already active. Otherwise you should set "active" on
	// entry and clear "active" on exit.
	protected static boolean active;

	// Log streams accessible so instrumenter can add writeItem methods
	protected DataOutputStream dataStream;		// For binary log entries
	protected PrintWriter textStream;			// For text log entries
	protected int logType;						// Type of log

	// Info about underlying log
	private ByteArrayOutputStream bufferStream;	// Underlying stream
	private RunLogSink sink;					// Owner of log

	// Types of log
	public static final int LOG_DEFAULT = -1;	// sink.open decides
	public static final int LOG_TEXT = 0;		// Log is text
	public static final int LOG_DATA = 1;		// Log is raw data types

	// Types of data in log
	public static final byte ITEM_BOOLEAN = 0;
	public static final byte ITEM_DOUBLE = 1;
	public static final byte ITEM_FLOAT = 2;
	public static final byte ITEM_INT = 3;
	public static final byte ITEM_LONG = 4;
	public static final byte ITEM_STRING = 5;
	public static final byte ITEM_DATE = 6;

	public static final byte ITEM_SOURCE = 10;
	public static final byte ITEM_EXTEND = 15;

	// Close the log
	public void closeLog()
	{
		if (sink!=null)
			sink.close();
		sink = null;
	}

	protected void finalize()
	{
		closeLog();
	}

	// Return type of log (data or text)
	public int getLogType()
	{
		return logType;
	}

	// Pad a numeric string to specified width
	// str = numeric value to pad
	// width = how wide to make it
	private final static String spaces
		= "                                            ";
	private final static int numSpaces = spaces.length();
	private static String leftPad(String str, int width)
	{
		int strlen = str.length();
		int pad = width - strlen;
		if (pad<=0)
			return str;
		else return spaces.substring(0, pad) + str;
	}

	// Constructors with default args
	//public RunLog() {this("run", LOG_DEFAULT);}
	public RunLog(String sug) {this(sug, LOG_DEFAULT);}

	// Sets logger to class on cmd-line, or defaults to RunLog
	// sug = a suggestion for name of log
	public RunLog(String sug, int logtype)
	{
		sink = this;				// Default to RunLog's implementation
		logType = logtype;

		// If explicit sink specified, create an object for it
		String sinkname = System.getProperty("RunLogSink");
		if (sinkname!=null)
		{
			try {
				Class cl = Class.forName(sinkname);
				sink = (RunLogSink)cl.newInstance();
			}
			catch (Exception e) {
				System.out.println(e.toString() + 
					sinkname + " not a valid class, "
					+ "will use RunLog's implementation");
			}
		}


		// Open the log and the logging buffer. The jdk1.1-specific way to
		// create the text stream is to use PrintWriter rather than PrintStream
		logType = sink.open(sug, this);
		try {
			bufferStream = new ByteArrayOutputStream();
			if (logType==LOG_TEXT)
				textStream = new PrintWriter(bufferStream, true);
			else dataStream = new DataOutputStream(bufferStream);
		}
		catch (RuntimeException e) {
			e.printStackTrace();
			throw new RuntimeException("Could not open logging buffer");
		}
	}

	// Convert number to string and format as specified
	// val = non-neg val to convert
	// whole = min number of whole digits
	private static String toString(double val, int whole, int frac)
	{
		double rnder = Math.pow(10, frac);
		double rnded = ((int)(val*rnder  + 0.5)) / rnder;
		return leftPad(new Double(rnded).toString(), whole+frac+1);
	}
	private static String toString(int val, int whole)
	{
		return leftPad(new Integer(val).toString(), whole);
	}
	private static String toString(long val, int whole)
	{
		return leftPad(new Long(val).toString(), whole);
	}

	// Pad and/or truncate val to specified field layout
	// left = if doing text, amt of padding to left of string
	// width = if doing text, width to put item into (plus lpad).
	//			0 says no limit
	private static String toString(String val, int left, int width)
	{
		// Setup left padding, and chk for simple case
		String lpad = spaces.substring(0, Math.min(numSpaces, left));
		if (width<=0)
			return lpad+val;

		// Pad or truncate on right
		int len = val.length();
		if (len>=width)
			return lpad + val.substring(0, width);
		else return lpad + val
			+ spaces.substring(0, Math.min(numSpaces, width-len));
	}

	// Convenience routine to write text and terminate entry in one call
	public void writeEntry(String item)
	{
		writeItem(item);
		writeEntry();
	}

	// Write previously built entry
	public void writeEntry()
	{
		// Terminate text bufferstream with eol
		if (logType==LOG_TEXT)
			textStream.println();

		// Output buffer stream to sink & clear it
		byte buf[] = bufferStream.toByteArray();
		if (buf.length<=0) return;
		sink.write(buf);
		bufferStream.reset();
	}

	// Append boolean data item to curr log entry
	public void writeItem(boolean item, int left, int width)
	{
		if (logType==LOG_TEXT)
			writeItem(item ? "true" : "false", left, width);
		else writeItem(item);
	}
	public void writeItem(boolean item)
	{
		try {
			if (logType==LOG_TEXT)
				textStream.print(item);
			else {
				dataStream.writeByte(ITEM_BOOLEAN);
				dataStream.writeBoolean(item);
			}
		}

		catch (Exception e) {
			e.printStackTrace();
			throw new RuntimeException("Could not log item");
		}
	}

	// Append Date data item to curr log entry
	public void writeItem(Date item, int left, int width)
	{
		if (logType==LOG_TEXT)
			writeItem(item.toString(), left, width);
		else writeItem(item);
	}
	public void writeItem(Date item)
	{
		try {
			if (logType==LOG_TEXT)
				textStream.print(item.toString());
			else {
				dataStream.writeByte(ITEM_DATE);
				dataStream.writeDouble(item.getTime());
			}
		}

		catch (Exception e) {
			e.printStackTrace();
			throw new RuntimeException("Could not log item");
		}
	}

	// Append double data item to curr log entry
	public void writeItem(double item, int whole, int frac)
	{
		if (logType==LOG_TEXT)
			writeItem(toString(item, whole, frac));
		else writeItem(item);
	}
	public void writeItem(double item)
	{
		try {
			if (logType==LOG_TEXT)
				textStream.print(item);
			else {
				dataStream.writeByte(ITEM_DOUBLE);
				dataStream.writeDouble(item);
			}
		}

		catch (Exception e) {
			e.printStackTrace();
			throw new RuntimeException("Could not log item");
		}
	}

	// Append float data item to curr log entry
	public void writeItem(float item, int whole, int frac)
	{
		if (logType==LOG_TEXT)
			writeItem(toString(item, whole, frac));
		else writeItem(item);
	}
	public void writeItem(float item)
	{
		try {
			if (logType==LOG_TEXT)
				textStream.print(item);
			else {
				dataStream.writeByte(ITEM_FLOAT);
				dataStream.writeFloat(item);
			}
		}

		catch (Exception e) {
			e.printStackTrace();
			throw new RuntimeException("Could not log item");
		}
	}

	// Append integer data item to curr log entry
	public void writeItem(int item, int whole)
	{
		if (logType==LOG_TEXT)
			writeItem(toString(item, whole));
		else writeItem(item);
	}
	public void writeItem(int item)
	{
		try {
			if (logType==LOG_TEXT)
				textStream.print(item);
			else {
				dataStream.writeByte(ITEM_INT);
				dataStream.writeInt(item);
			}
		}

		catch (Exception e) {
			e.printStackTrace();
			throw new RuntimeException("Could not log item");
		}
	}
	public void writeItem(long item, int whole)
	{
		if (logType==LOG_TEXT)
			writeItem(toString(item, whole));
		else writeItem(item);
	}
	public void writeItem(long item)
	{
		try {
			if (logType==LOG_TEXT)
				textStream.print(item);
			else {
				dataStream.writeByte(ITEM_LONG);
				dataStream.writeLong(item);
			}
		}

		catch (Exception e) {
			e.printStackTrace();
			throw new RuntimeException("Could not log item");
		}
	}


	// Write string
	// item = string to write
	// left = if doing text, amt of padding to left of string
	// width = if doing text, width to put item into (plus lpad).
	//			0 says no limit
	public void writeItem(String item, int left, int width)
	{
		if (item==null) item = "null";
		if (logType==LOG_TEXT)
			writeItem(toString(item, left, width));
		else writeItem(item);
	}
	public void writeItem(String item)
	{
		if (item==null) item = "null";
		try {
			if (logType==LOG_TEXT)
				textStream.print(item);
			else {
				dataStream.writeByte(ITEM_STRING);
				dataStream.writeUTF(item);
			}
		}

		catch (Exception e) {
			e.printStackTrace();
			throw new RuntimeException("Could not log item");
		}
	}

	// *** Default implementation of the Sink interface *** //
	private OutputStream sinkStream;

	// Close the log sink
	public void close()
	{
		try {
			sinkStream.close();
		}
		catch (Exception e) {
			e.printStackTrace();
			throw new RuntimeException("Could not close log");
		}
	}

	// Open the sink
	// usage = may be used as suggestion for name of log. null is
	//		considered sug for System.out.
	// log = a log-entry writing envir (can be saved by logger that wants
	//		to write log content itself -- eg. timestamps)
	public int open(String usage, RunLog log)
	{
		if (usage==null)
			sinkStream = System.out;

		// Open a file
		else {
			String name = usage;
			if (name.indexOf('.')<0) name += ".log";
			try {
				sinkStream = new FileOutputStream(name);
			}
			catch (Exception e) {
				e.printStackTrace();
				throw new RuntimeException("Could not open " + name);
			}
		}

		// Set type of log
		if (log.logType==LOG_DEFAULT)
			return LOG_TEXT;
		else return log.logType;
	}

	// Log an entry
	public void write(byte[] entry)
	{
		try {
			sinkStream.write(entry);
		}
		catch (Exception e) {
			e.printStackTrace();
			throw new RuntimeException("Could not write to log");
		}
	}

}

