//
// THIS SOFTWARE IS FURNISHED UNDER A LICENSE AND  MAY
// BE USED OR COPIED ONLY IN ACCORDANCE WITH THE TERMS
// OF SUCH LICENSE.
//
//
// COPYRIGHT (c) 1997 BY DIGITAL EQUIPMENT CORPORATION.
// ALL RIGHTS RESERVED.
//

package dec.trek;
import java.io.*;
import java.util.*;

//
//
// Statement
//
//
public class Statement 
{
	// Info needed to make its context self-contained
	Method parent;					// Identifies Statement's method
	Statement next,prev;			// Its neighbors in its method

	// Size of statement is minimized because stmts are the major
	// storage requirement in a trek. That is, some fields do double duty.
	//
	// During explode (ie. in Divide, Finish, and Lvt):
	// 1) Addr is target-addr of if/goto-stmts (>= 0) or a coded val, see ad_*
	// 2) Delta is cnt of soft locals buildLvt pass1 marked for this stmt
	// 3) Mate is stmt at other end of blk (ie. top of loop pts to bottom,
	//	  and vice versa)
	//
	// After explode:
	// 1a) Addr in a regular stmt is its Statement # 
	// 1b) Addr in ST_PIECE STMT is its size change wrt its original size
	// 2)  Delta is # of bytes that ins/del's of code in earlier stmts have
	//	   moved this stmt
	// 3)  Mate of modified orig stmt pts to its last split-off piece if any
	//
	// The loan field:
	// 1) In regular stmt, it is non-null when an obj ref for this stmt was
	//	  loaded in an earlier stmt. This can occur when methods are in-lined.
	// 2) In ST_PIECE, it is a stmt containing the inst of a deleted-inst piece
	//
	// The string field:
	// 1) In regular stmt, it is cached toString value
	// 2) In inserted-code stmt, it is the add's change ID
	int addr;
	int delta;
	byte insts[];					// Bytes of statement's instructions
	int level;						// Indent level of statement (1 = top)
	Statement loan;
	Statement mate;
	int start;						// Offset of stmt's 1st inst in method
	String string;
	public Object userDefined;		// User-owned field
	int type;						// Type of statement (ST_*)

	// Addr coded values
	final static int ad_undef = -1;	// Not in use
	final static int ad_bound = -2;	// Says switch stmt is alr bnd to bottom
	final static int ad_init = -3;	// Stmt is For-loop's init clause
	final static int ad_incr = -4;	// Stmt is For-loop's incr clause (see
									//	classifyLoops & finishForLoops)

	// Priority level for inserted code (-10's go 1st, etc)
	final static int plev_beftrek = -10;
	final static int plev_befclass = -9;
	final static int plev_befmethod = -8;
	final static int plev_else = -7;
	final static int plev_normal = -6;
	final static int plev_return = -5;
	final static int plev_aftmethod = -4;
	final static int plev_aftclass = -3;
	final static int plev_afttrek = -2;

	// Return the relative jump that now corresponds to target
	// inst = instruction that jump is part of
	// target = orig absolute location of instruction that inst branches to
	// Notes:
	//	May only be used during writing-stmts phase.
	//	Switch inst & its stmt contain the new switch padding, but target
	//	is with respect to original padding
	int adjustJump(Instruction inst, int target)
	{
		// If have inter-stmt jump, combine target's adjustment with
		// adjust of curr stmt to see if jump delta needs to be mod.
		// (Could always do adjustPc starting with 1st stmt. Starting
		// with curr stmt in forward jump is just an optimization)
		int stend = start + insts.length;
		int oldjump = target - start - inst.start;
		int targadj;
		if (target < start)						// To earlier stmt
			targadj = Statement.adjustPc(parent, target) - target;
		else if (target > stend					// To later stmt
			|| (target == stend && next.type==Trek.ST_AFTER)
												// (incl next if curr has Aft chunk)
			|| inst.type==Vmi.it_switch)		// (By def if switch inst)
			targadj = adjustPc(target) - target;
		else return oldjump;					// Inside same piece

		// Instructions actually contain relative jumps.
		// Chk for patho case of "trying to switch" to wide jump
		int srcadj = delta + preSize();
		int adjust = targadj - srcadj;
		int newjump = oldjump + adjust;
		if (Math.abs(oldjump)<32767 && Math.abs(newjump)>=32767)
			throw new TrekException(
				"Increasing jump delta to >32766 is not supported");
		return newjump;
	}

	// Returns an updated value for pc that reflects the code ins/del
	// for the stmts that precede pc's stmt -- and that stmt's
	// addBefore's if pc is in the middle of a stmt.
	// pc = method-absolute of where an inst in "this" jumps to
	// May only be used during writing-stmts phase
	static int adjustPc(Method method, int pc)
	{
		return method.firstStatement().adjustPc(pc);
	}
	int adjustPc(int pc)
	{
		// Scan from access pt to specified pc, ignoring inserted code
		for (Statement stmt=this; stmt!=null; stmt=stmt.next) {
			if (stmt.type<=Trek.ST_INSERTED) continue;

			// Ignore 0-length source artifacts like }'s and try's.
			// Reached stmt where this pc is, so future deltas wont move it.
			if (stmt.insts.length==0) continue;
			if (pc == stmt.start)
				return pc + stmt.delta;

			// Because adjusted switch pad is in insts.length, discount
			// this when chking if pc is in middle of stmt
			int stend = stmt.start + stmt.insts.length;
			if (stmt.type==Trek.ST_PIECE)
				stend -= stmt.addr;

			// If pc is in middle of an unsplit stmt, then stmt's pre-piece
			// also pushes it forward. Else pre-piece is part of delta
			// of the split piece
			if (pc < stend)
			{
				if (stmt.type==Trek.ST_PIECE)
					return pc + stmt.delta;
				else if (stmt.mate==null)
					return pc + stmt.delta + stmt.preSize();
			}
		}							// End loop

		// To here means local's scope ends past all code. So adjust by
		// last statement's delta and any pre-pieces it has. (Note that
		// there cant be code past last stmt -- it would be inaccessible).
		Statement last = parent.last;
		return pc + last.delta + last.preSize();
	}

	// Returns inst that directly (or via subclass overriding) can call
	// the specified member, or null
	// meth = method that may be callable from this stmt
	public Instruction canCall(Method meth)
	{
		ClassFile cf = meth.parent;
		String defstr = meth.toName() + meth.toSignature();

		// Search stmt's insts for this ref
		for (Instruction inst=firstInstruction(); inst!=null; inst=inst.next()) {
			if (inst.type!=Vmi.it_call)
				continue;
			Member refmem = inst.getMember();
			if (refmem==null || refmem instanceof Field)
				continue;

			// Chk if it is directly or indirectly the desired method
			Method refmeth = (Method)refmem;
			if (refmeth==meth)
				return inst;
			if (!refmeth.hasModifier(Trek.ACC_PRIVATE)
				&& !meth.hasModifier(Trek.ACC_PRIVATE)
				&& defstr.equals(refmeth.toName() + refmeth.toSignature())
				&& refmeth.getClassFile().isAncestor(cf))
				return inst;
		}
		return null;
	}

	// Returns # of bytes in method's statements & instrumentation
	// Also sets total amount (delta) that start of each stmt has moved
	// Notes:
	//	Directly accesses stmts since does not require exploded method
	//	and wants length of inserted code.
	//	Last stmt cant have anything inserted after it because last stmt must
	//	be either a return or a branch-up.
	static int codeLength(Method method)
	{
		int len = 0;
		int predelta = 0;		// Stuff before curr stmt
		int tdelta = 0;			// Total delta from prev stmts
		Statement stswit = null;

		// Scan to set deltas and length of method. Note that len of
		// switch stmt can be changed in loop itself.
		for (Statement stmt=method.first; stmt!=null; stmt=stmt.next) {
			switch (stmt.type) {

			// Inserted code
			case Trek.ST_BEFORE:
				predelta += stmt.insts.length;
				break;
			case Trek.ST_AT:
			case Trek.ST_AFTER:
				tdelta += stmt.insts.length;
				break;

			// A (possibly modified) piece of the original stmt.
			case Trek.ST_PIECE:
				stmt.delta = tdelta;

				// Orig stmt has switch in it. Keep calling fixSwitchPad until
				// hit piece (and fix it) that has the switch
				if (stswit!=null)
				{
					Statement st = stmt.fixSwitchPad(stswit);
					if (st!=null)
					{
						stswit = null;
						stmt = st;		// fixSP replaced stmt
					}
				}

				// stmt.addr is the diff in size (if any) from the original
				// size of this piece of code -- so it is non-0 only if:
				// 1) User deleted inst
				// 2) Stmt contains switch inst whose padding needed change
				tdelta += stmt.addr;
				break;

			// At (start of) a real stmt, set its delta.
			// If stuff inserted before this stmt, acct for its delta.
			// Not done earlier because start of stmt must be at any
			// inserted before-code so that a branch to stmt's start
			// executes this code.
			default:
				stmt.delta = tdelta;
				if (predelta>0)
				{
					stmt.prev.delta = predelta;	// For preSize()
					tdelta += predelta;
					predelta = 0;
				}

				// Unfort the size of a switch inst depends on its alignment.
				// So adjust 0-padding as necessary -- now if stmt hasnt been
				// split, or later when hit piece that got switch.
				if (stmt.type==Trek.ST_SWITCH)
				{
					if (stmt.mate!=null)
						stswit = stmt;
					else stmt.fixSwitchPad(stmt);
				}
			}

			// If stmt is not an orig-stmt supplanted by pieces (ie. mate
			// non-null), its length is part of the length of the method.
			if (stmt.mate==null)
				len += stmt.insts.length;
		}								// End loop
		return len;
	}

	// Delete stmt from a method
	// chid = change ID to make deletion with
	public void delete(String chid)
	{
		// Cleanup any interior changes
		if (mate!=null)
		{
			Statement stn = null;
			for (Statement st=next; ; st=stn) {
				stn = st.next;
				if (st.type<=Trek.ST_PIECE)
					st.remove();

				// If to after-inserts or last piece, done
				if (st.type==Trek.ST_AFTER || st==mate)
					break;
			}
		}

		// Insert empty piece after main stmt
		mate = new Statement(parent, 0, 0, null);
		mate.string = chid!=null ? chid : "";
		mate.addr = -insts.length;
		mate.insert(next);
		parent.parent.modified = true;
	}

	// Dump instructions of this statement
	void dump(PrintWriter chan)
	{
		// Scan thru bytes of statement's instructions
		// Assigning instruction byte to int causes it to be sign-extended.
		// so must any away these high-order 1 bits
		for (int i=0; i<insts.length; ) {
			Instruction inst = new Instruction(this, i);
			int offset = start+i;
			Trek.println(chan, offset + ": " + inst.toString());
			i += inst.size;
		}
	}

	// Does this statement fetch the specified local
	Instruction fetches(Local loc)
	{
		return references(loc, 1);
	}

	// Return 1st instruction of statement
	public Instruction firstInstruction()
	{
		if (insts.length==0)
			return null;
		else return new Instruction(this, 0);
	}

	// Return 1st chunk of this statement. Iow, if stmt has any Before
	// chunks, want 1st of those instead of orig stmt.
	Statement firstChunk()
	{
		Statement first = this;
		for (Statement pr=prev; pr!=null && pr.type==Trek.ST_BEFORE; pr=pr.prev) {
			first = pr;
		}
		return first;
	}

	// Return 1st stmt of Then, if any
	public Statement firstOfThen()
	{
		if (type != Trek.ST_IF)
			throw new TrekException("Specified statement is not an IF");

		// Get to body of Then
		Statement st = next();
		if (st.type == Trek.ST_THEN)
			st = st.next();

		// If If-stmt optimized to branch directly to Then, there is no
		// real Then or Else path, so return null
		if (st.insts.length==0
			&& (st.type==Trek.ST_BREAK || st.type==Trek.ST_CONTINUE))
			return null;
		return st;
	}

	// Return 1st stmt of Else, if any
	public Statement firstOfElse()
	{
		// Get to body of Then
		Statement st = firstOfThen();
		if (st==null) return null;

		// Scan past Then-blk. If Else is there, ret its 1st stmt
		for (; st.level>level || st.type==Trek.ST_ENDTHEN; st=st.next()) ;
		if (st.type != Trek.ST_ELSE)
			return null;
		return st.next();
	}

	// If this switch stmt's alignment has changed, fix its 0-padding
	// orig = orig stmt (may still be active, ie. not divided into pieces)
	// Returns stmt that finally contains switch inst, or null.
	private Statement fixSwitchPad(Statement orig)
	{
		// Find where the switch inst is, so know its offset and
		// where the padding ends (the potential split point)
		Instruction curri;
		Instruction inst = null;
		for (int i=0; i<insts.length; i+=curri.size) {
			curri = new Instruction(this, i);
			if (curri.opcode==Code.tableswitch || curri.opcode==Code.lookupswitch)
			{
				inst = curri;
				break;
			}
		}
		if (inst == null) return null;

		// See what's happened to padding
		int oldpad = inst.getPad(-1);
		int newpad = inst.getPad(1);
		int adjpad = newpad - oldpad;
		if (adjpad == 0) return this;

		// Create piece with adjusted padding
		Statement piece = new Statement(parent, 0, insts.length+adjpad, null);
		piece.start = start;
		piece.delta = delta;
		piece.insert(next);

		// Adding mate or replacing piece
		if (this!=orig)
		{
			piece.addr = addr + adjpad;
			remove();
			if (this==orig.mate) orig.mate = piece;
		}
		else {
			mate = piece;
			piece.addr = adjpad;
		}

		// Gen the instructions from orig insts -- stretching or
		// squeezing the padding via loop or overlaid 2nd arraycopy
		int split = inst.start + 1 + oldpad;	// offset past end of old padding
		System.arraycopy(insts, 0, piece.insts, 0, split);
		for (int i=0; i<adjpad; i++) piece.insts[split+i] = 0;
		System.arraycopy(insts, split, piece.insts, split+adjpad, insts.length-split);
		return piece;
	}										// End fixSwitchPad

	// Get the nearest end of blk that encompasses both stmts
	// this = stmt in blk1 (can be its eob)
	// st2 = stmt in blk2 (can be its eob)
	Statement getCommonEob(Statement st2)
	{
		// Get to same level (only one or the other while can be entered)
		Statement st1 = this;
		while (st1.level > st2.level) st1 = st1.getEob();
		while (st2.level > st1.level) st2 = st2.getEob();

		// Find common end of blk
		if (st1!=st2)
		{
			st1 = st1.getEob();
			st2 = st2.getEob();
		}
		return st1;
	}

	// Return statement that ends block that "this" is in, handling
	// special case of init clause of not-finished For loop.
	// Case stmts do not implicitly start/end a a declarative scope,
	// but they do end If's and the like.
	// cases = true says consider case-stmts as the potential eob
	Statement getCodeEob() {return getEob(true);}
	Statement getEob() {return getEob(false);}
	Statement getEob(boolean cases)
	{
		Statement st = this;
		int inlevel = level;		// Level just in blk

		// If name in (still-separate) init clause of For, set
		// logical context to be inside loop
		if (addr==ad_init)
		{
			inlevel += 1;
			st = next().next();
		}

		// Scan to end of block, considering case-stmts if cases true
		while (st.level>=inlevel || (st.type==Trek.ST_CASE && !cases)) {
			Statement stn = st.next();
			if (stn==null) return st;
			st = stn;
		}
		return st;
	}

	// Scan forward to nearest stmt before which a blk-ending stmt at the
	// specified level can be inserted, and return this stmt
	Statement getEobInsertPoint(int lev)
	{
		for (Statement inspt=this; ; inspt=inspt.next()) {
			if (inspt.level>lev) continue;
			switch (inspt.type) {
			case Trek.ST_FOR:
				if (inspt.prev().addr!=Statement.ad_init)
					return inspt;
			case Trek.ST_ENDFOR: case Trek.ST_ENDWHILE: case Trek.ST_ENDDO:
			case Trek.ST_THEN: case Trek.ST_ENDTHEN:
			case Trek.ST_ELSE: case Trek.ST_ENDELSE:
			case Trek.ST_CATCH: case Trek.ST_ENDCATCH:
			case Trek.ST_FINALLY: case Trek.ST_ENDFINALLY:
				continue;
			}
			return inspt;
		}
	}

	// Scan back to nearest stmt before which a stmt at the
	// specified level can be inserted, and return this stmt
	Statement getInsertPoint(int lev)
	{
		for (Statement inspt=this; ; inspt=inspt.prev()) {
			if (inspt.level>lev) continue;
			switch (inspt.type) {
			case Trek.ST_FOR:
				if (inspt.prev().addr!=Statement.ad_init)
					return inspt;
			case Trek.ST_ENDFOR: case Trek.ST_ENDWHILE: case Trek.ST_ENDDO:
			case Trek.ST_THEN: case Trek.ST_ENDTHEN:
			case Trek.ST_ELSE: case Trek.ST_ENDELSE:
			case Trek.ST_CASE: case Trek.ST_ENDTRY:
			case Trek.ST_CATCH: case Trek.ST_ENDCATCH:
			case Trek.ST_FINALLY: case Trek.ST_ENDFINALLY:
				continue;
			}
			return inspt;
		}
	}

	public Method getMethod()
	{
		return parent;
	}

	public int getLevel()
	{
		return level;
	}

	public int getNumber()
	{
		return addr;
	}

	public int getPc()
	{
		return start;
	}

	// Returns the line number of this statement in source file if:
	// stmt is non-declarative [else 0] and LineNumber table exists [else -1]. 
	// Note that line number of loop's } is start line of the loop
	public int getSourceNumber()
	{
		// Not-applic chks
		if (parent.linums==null) return -1;
		if (insts.length==0) return 0;

		// Scan table for this stmt's start pc
		int len = parent.linums.length;
		for (int i=0; i<len; i++) {
			Linum linum = parent.linums[i];
			if (linum.startPc==start)
				return linum.lineNumber;
		}
		return 0;
	}

	public int getType()
	{
		return type;
	}

	// Returns true if there is stmt after "this" whose start is before blktop
	// and whose target is between blktop & botstart inclusive. For ex, this 
	// can clarify that bot is a loop bot rather than an optimized Continue
	boolean goesBetween(int blktop, int botstart)
	{
		// Scan looking for branch to range between blktop and bot
		for (Statement st=next(); ; st=st.next()) {
			if (st==null || st.start>=blktop)
				return false;
			if (st.type==Trek.ST_UNKNOWN || st.type==Trek.ST_IF)
			{
				int target = st.addr;		// Get goto's target
				if (target>=blktop && target<=botstart)
					return true;
			}
		}
	}

	// Return true if matching code changes in this stmt
	public boolean hasChange(String chid)
	{
		return Change.handleChange(chid, this, true);
	}

/*
	// Chk if stmt had goto (and call) added before itself
	// this = the stmt that might have goto
	boolean hasAddedGoto()
	{
		// Set stmt to 1st chunk of stmt If branches to
		// Does the stmt that If branches to start with inserted goto?
		Statement stmt = firstChunk();
		if (stmt.type!=Trek.ST_BEFORE) return false;
		Instruction inst = new Instruction(stmt, 0);
		return inst.opcode==Vmi.goto0;
	}
*/

	// Insert an item before succ
	void insert(Statement succ)
	{
		if (succ==null)
		{
			insert();
			return;
		}

		// Insure not already in.
		if (next!=null || prev!=null || parent.first==this)
			throw new TrekException("Statement already linked into a method");

		// There is a successor
		next = succ;
		prev = succ.prev;
		succ.prev = this;
		if (succ==parent.first)
			parent.first = this;
		else prev.next = this;
	}

	// Insert an item at end of list
	void insert()
	{
		// Insure not already in.
		if (next!=null || prev!=null || parent.first==this)
			throw new TrekException("Statement already inserted in method");

		// First in list
		if (parent.last==null)
		{
			parent.first = parent.last = this;
			next = prev = null;
		}

		// Append
		else {
			next = null;
			prev = parent.last;
			prev.next = this;
			parent.last = this;
		}
	}

	// Navigate to last instruction of stmt.
	public Instruction lastInstruction()
	{
		if (insts.length==0)
			return null;

		// Find last by brute force forward scan
		Instruction curri = new Instruction(this, 0);
		while (curri.start+curri.size < insts.length) curri = curri.next();
		return curri;
	}

	// Does this statement modify the specified member or local
	public Instruction modifies(Field fld)
	{
		return references(fld, true);
	}
	public Instruction modifies(Local loc)
	{
		return references(loc, -1);
	}

	// Navigate to next statement of method. If at end, ret null instead
	public Statement next()
	{
		for (Statement st=next; st!=null; st=st.next) {
			if (st.level >= 0) return st;
		}
		return null;
	}

	// Navigate to prev statement of method. If at start, ret null instead
	public Statement prev()
	{
		for (Statement st=prev; st!=null; st=st.prev) {
			if (st.level >= 0) return st;
		}
		return null;
	}

	// Returns the size of this stmt's ST_BEFORE's
	// Notes:
	//	May only be used during writing-stmts phase.
	//	Because ST_BEFORE's can only directly precede an orig stmt, guaran
	//	to return 0 if "this" is not an orig stmt
	int preSize()
	{
		if (prev!=null && prev.type==Trek.ST_BEFORE)
			return prev.delta;
		else return 0;
	}

	// Does this statement reference the specified member or local
	public Instruction references(Member mem)
	{
		return references(mem, false);
	}
	public Instruction references(Local loc)
	{
		return references(loc, 0);
	}

	// Returns inst that does the ref, or null if no such inst
	// Notes:
	//	This is a direct ref. If stmt refs the analogous member
	//	from a superclass of mem's class, this isnt a match.
	private Instruction references(Member mem, boolean modify)
	{
		// If classfile doesnt ref member, obviously none of its stmts do
		ClassFile cf = parent.parent;
		MemberRef ref = MemberRef.get(mem, cf);
		if (ref==null)
			return null;

		// Search stmt's insts for this ref
		Instruction inst;
		for (int i=0; i<insts.length; i+=inst.size) {
			inst = new Instruction(this, i);
			if (modify && !inst.modifies())
				continue;
			if (ref.idx == inst.getCpIdx())
				return inst;
		}
		return null;
	}

	// Returns offset of inst that does the ref, or null if no such inst
	// loc = identifies local var to chk
	// filter = 0 says any kind of ref, 1 says reads, -1 says modifies
	private Instruction references(Local loc, int filter)
	{
		if (loc.parent!=parent || !loc.isActive(start+insts.length))
			return null;
		int locidx = loc.idx;

		// Search stmt's insts for this ref
		Instruction inst;
		for (int i=0; i<insts.length; i+=inst.size) {
			inst = new Instruction(this, i);

			// Do write-only & read-only chks
			if (filter!=0)
			{
				boolean modifies = inst.modifies();
				if (filter<0 && modifies)
					;					// Want and is write
				else if (filter>0 && !modifies)
					;					// Want and is read
				else continue;
			}

			// Does inst ref this local?
			if (locidx == inst.getLocalIdx())
				return inst;
		}
		return null;
	}

	// Remove statement from method
	void remove()
	{
		// Hook predecessor (if any) to successor
		if (prev!=null)
			prev.next = next;
		else parent.first = next;

		// Hook successor (if any) to predecessor
		if (next!=null)
			next.prev = prev;
		else parent.last = prev;

		// Indic no longer in chain
		next = prev = null;
	}

	// Split the statement at the point indicated by code object.
	// this = orig stmt
	// code = info about code being inserted
	// Returns 2nd piece of split. If no split needed, returns succ of scanned stmt
	static final int spl_after = -2;
	static final int spl_later = -1;
	Statement split(Change code)
	{
		// If inserting Before, sort by priority. Note that orig stmt's level
		// has higher value than any plev_*.
		if (code.offset==Trek.ST_BEFORE)
		{
			Statement st;
			for (st=firstChunk(); code.priority>=st.level; st=st.next) ;
			return st;
		}
	
		// Scan is orig stmt itself or its pieces (next thru mate)
		Statement stmt = (mate==null) ? this : next;
		Statement succ = (mate==null) ? next : mate.next;

		// Scan for desired ins pt
		int residue[] = {code.offset};
		for (; stmt!=succ; stmt=stmt.next) {
			if (stmt.type <= Trek.ST_INSERTED)
				continue;				// Split occurs in real code only

			// Chk if at split-point yet
			int curr = stmt.splitNeeded(code, residue);
			switch (curr) {
			case spl_after:
				return succ;			// A simple addAfter
			case spl_later:
				continue;				// Not there yet

			// If mult inserts here, sort by priority
			case 0:
				for (stmt=stmt.prev; code.priority<stmt.level; stmt=stmt.prev) {
					if (stmt.type > Trek.ST_INSERTED) break;
				}
				return stmt.next;		// Put at beginning of successor
			}

			// Found the ins pt. Split stmt here
			Statement prefix = new Statement(parent,
				0, curr, stmt.insts);
			prefix.start += stmt.start;
			Statement suffix = new Statement(parent,
				curr, stmt.insts.length-curr, stmt.insts);
			suffix.start += stmt.start;

			// Update stmt chain and return suffix of split
			suffix.insert(stmt.next);
			prefix.insert(suffix);
			if (stmt.type==Trek.ST_PIECE)
			{
				if (mate==stmt) mate = suffix;
				stmt.remove();
			}
			else mate = suffix;
			return suffix;
		}								// End loop

		// Fell thru, shouldnt happen unless offset exhausted just as got
		// to succ -- which can happen for iinc that ends stmt
		if (residue[0]==0)
			return succ;
		throw new TrekException(
				"Could not find insert point in: " + toLocation());
	}

	// Returns spl_after if inserted code goes after stmt. Else calcs
	// where to split stmt by reducing residue of inspt offset:
	// 1) If not to right chunk yet, spl_later returned
	// 2) Else returns where to split current chunk
	// this = current chunk
	// code = describes inspt
	// residue = curr residue of inspt offset (or coded offset value)
	private int splitNeeded(Change code, int residue[])
	{
		int resid = residue[0];

		// Get original length
		int origlen = insts.length;
		if (type==Trek.ST_PIECE)
			origlen -= addr;

		// User said exactly which inst to split at. Problem is this offset
		// is with respect to orig stmt.
		if (resid>=0)
		{
			if (origlen > resid)
				return resid;
			residue[0] -= origlen;
			return spl_later;
		}

		// Switch key, return val, or throw exception?
		else if (resid==Trek.ST_AT)
		{
			Instruction inst;
			for (int i=0; i<insts.length; i+=inst.size) {
				inst = new Instruction(this, i);
				if (inst.type==Vmi.it_switch || inst.type==Vmi.it_return)
					return i;
			}
			return spl_later;
		}

		// Vanilla case of after whole stmt
		else return spl_after;
	}

	// Construct primordial statement.
	Statement(Method method, int len)
	{
		statement(method, 0, len, null, Trek.ST_PRIMORD, 0, null);
	}

	// Construct regular in-method statement
	Statement(Method method, int start, int len, byte insts[],
		int type, int lev, String str)
	{
		statement(method, start, len, insts, type, lev, str);
	}

	// Construct regular in-method code-less statement
	Statement(Method method, int start, int type, int lev, String str)
	{
		statement(method, start, 0, null, type, lev, str);
	}

	// Break off piece of statement
	Statement(Method method, int start, int len, byte insts[])
	{
		statement(method, start, len, insts, Trek.ST_PIECE, -1, null);
	}

	// Construct inserted statement
	// prilev = <0 for inserted (see plev_*)
	Statement(Method method, int len, byte insts[], int prilev, int type)
	{
		statement(method, 0, len, insts, type, prilev, null);
	}

	// Do work of constructor
	// method = parent method
	// start = offset of stmt in method (0 if a code modification)
	// len = # of bytes of code in this stmt (may be 0)
	// insts = the code itself (or null to set/read-in later)
	// type = type of stmt
	// lev = indent level. Normal top=1. Primord=0. Before,after<0 (plev_*)
	// str = source code of stmt or null
	private void statement(Method method,
		int start, int len, byte insts[],
		int type, int lev, String str)
	{
		parent = method;
		level = (byte)lev;
		this.type = (byte)type;
		this.start = start;				// Offset of stmt from top of method
		string = str;					// Use any stmt-text built now

		// Setup insts if passed by caller
		this.insts = new byte[len];
		if (insts != null && len > 0)
			System.arraycopy(insts, start, this.insts, 0, len);

		// If new stmt past all others, append it to method's stmts
		if (level < 0) return;
		Statement last = method.lastStatement(); 
		if (last==null || last.start+last.insts.length==start)
			insert();
	}

	// Return statement's logical location as a string: class.method(num)
	public String toLocation()
	{
		String from = parent.getClassFile().toName() + "." + parent.toName();
		String frnum = "(" + getNumber() + ")";
		return from + frnum;
	}

	// Return string that uniquely identifies this stmt (see Trek.getObject)
	public String toObjectId()
	{
		return parent.toObjectId() + "#" + getNumber();
	}

	// Convert Statement to source-code string (if havent already) and indent
	// prepad = # of spaces for level 1 code
	// levpad = amt to indent per each higher level
	// maxpad = max chars of indent allowed
	private static final String padstr = "                              " +
		"                                                                      ";
	private static final int maxPad = 100;
	public String toIndent(int prepad, int levpad, int maxpad)
	{
		int rawpad = prepad + levpad*(level-1);
		int padlen = Math.min(rawpad, Math.min(maxpad, maxPad));
		padlen = Math.max(0, padlen);			// Must be positive
		return padstr.substring(0, padlen);
	}

	// Convert Statement to source-code string (if havent already)
	public String toString()
	{
		return toString(null);
	}
	String toString(Vmi vm)
	{
		if (string==null)
		{
			if (vm==null) vm = new Vmi();
			string = vm.execute(this);
		}
		return string;
	}

	// Remove the matching code changes from this stmt
	public boolean undoChange(String chid)
	{
		return Change.handleChange(chid, this, false);
	}

	// Write jump-adjusted instructions of stmt
	void write(DataOutput chan) throws IOException
	{
		Instruction inst;
		int done = 0;
		int len = insts.length;
		if (addr==-len)
			return;					// User deleted inst

		// Scan stmt, looking for jumps to adj. If inst has no jump, ignore it
		for (int curr=0; curr<len; curr+=inst.size) {
			inst = new Instruction(this, curr);
			int target = inst.getTarget();
			if (target==Instruction.tar_none) continue;

			// Write stuff before jump(s), the jump(s), and reset done offset.
			// 1) Switch stmt. Adjust each jump in its branch table
			chan.write(insts, done, curr+1-done);
			if (target==Instruction.tar_multi)
				Case.writeSwitch(chan, inst);

			// 2) Simple jumping instruction.
			else {
				int newjump = adjustJump(inst, target);
				if (inst.opcode==Code.goto_w || inst.opcode==Code.jsr_w)
					chan.writeInt(newjump);			// Wide jump
				else chan.writeShort(newjump);		// Normal 2-byte target
			}
			done = curr + inst.size;
		}

		// Write what hasnt been written yet
		if (len-done > 0)
			chan.write(insts, done, len-done);
	}									// End of write

	// Write code of method, combining statements into stream of bytes
	// chan = where to write code
	// method = method code is part of
	// Directly accesses stmts since does not require exploded method
	// and wants to write out inserted stuff
	static void writeStmts(DataOutput chan, Method method)
		throws IOException
	{
		// Write instructions of method as a stream of bytes. For user's
		// code, may need to adjust jumps to acct for inserted code
		for (Statement stmt=method.first; stmt!=null; stmt=stmt.next) {
			if (stmt.type>Trek.ST_INSERTED)
			{
				if (stmt.mate==null)
					stmt.write(chan);		// Still orig code or at a PIECE
			}
			else chan.write(stmt.insts);
		}
	}

}
