package dec.trek;
import java.util.Vector;

//
//
// Term. A term in an if-expression.
//
//
final class Term
{
	private boolean trueJump;		// Term does jump on a src-level true
	private int ifStart;			// Start of this term's if-inst
	Datum opd1;						// Left operand
	Datum opd2;						// Right operand
	int opr;						// Trek.DOIF_*
	private int subStart;			// # of sub-expr starting at this term
	private int subEnd;				// # of sub-expr ending at this term
	private int target;				// Where this expr branches to:
									// Addresses are method-relative.

	// How else-branch is inverted back to src-level Then orientation
	private static final int invertop[] = {
		Trek.DOIF_NE, Trek.DOIF_EQ,
		Trek.DOIF_GE, Trek.DOIF_LE, 
		Trek.DOIF_GT, Trek.DOIF_LT
	};

	// Types of operators (1st 6 from Trek.DOIF_*)
	// Operator precedence in increasing order
	static final int op_assign = 6;
	static final int op_cond = 7;
	static final int op_boolor = 8;
	static final int op_booland = 9;

	// Next 3 true jvm binop insts -- but can be applied to booleans
	static final int op_or = 10;
	static final int op_xor = 11;
	static final int op_and = 12;
	static final int op_comp = 13;
	static final int op_instance = 14;

	// True jvm binop insts with numeric result (and int32 if integers)
	static final int op_lshift = 15;
	static final int op_rshift = 16;
	static final int op_rushift = 17;
	static final int op_plus = 18;
	static final int op_minus = 19;
	static final int op_mult = 20;
	static final int op_div = 21;
	static final int op_rem = 22;

	// Unary operators
	static final int op_not = 23;
	static final int op_negate = 24;
	static final int op_cast = 25;

	// Free-standing. Iconst says intVal of datum represents value of datum
	static final int op_atomic = 26;
	static final int op_iconst = 27;		// Int constant
	static final int op_pconst = 28;		// Ptr constant
	static final int op_absent = 29;		// No-opr has min priority

	// String representation of operators
	static final String oprs[] = {
		" == ", " != ", " < ", " > ", " <= ", " >= ",
		" = ", " ? ",
		" || ", " && ",
		" | ", " ^ ", " & ",
		" @ ", " instanceof ",
		" << ", " >> ", " >>> ",
		" + ", " - ", " * ", " / ", " % ",
		" ! ", " - ", "",
		"", "", "", ""
	};

	// Operator precedence in increasing order
	static final int pr_absent = -1;
	static final int pr_assign = 0;
	static final int pr_cond = 1;
	static final int pr_boolor = 2;
	static final int pr_booland = 3;
	static final int pr_or = 4;
	static final int pr_xor = 5;
	static final int pr_and = 6;
	static final int pr_relat1 = 7;
	static final int pr_relat2 = 8;
	static final int pr_shift = 9;
	static final int pr_add = 10;
	static final int pr_mult = 11;
	static final int pr_unary = 12;
	static final int pr_none = 13;

	// Precedence of each operator
	static final int precs[] = {
		pr_relat1, pr_relat1, pr_relat2, pr_relat2, pr_relat2, pr_relat2,
		pr_assign, pr_cond,
		pr_boolor, pr_booland,
		pr_or, pr_xor, pr_and,
		pr_relat1, pr_relat1,
		pr_shift, pr_shift, pr_shift,
		pr_add, pr_add, pr_mult, pr_mult, pr_mult,
		pr_unary, pr_unary, pr_unary,
		pr_none, pr_none, pr_none, pr_absent
	};

	// Figure out and/or's and parenthesis in if-expr
	private void analyzeExpr(Vector ifexpr, int ifend)
	{
		int size = ifexpr.size();
		Term nterm = null;

		// Scan backwards to understand why term branches to its target
		for (int i=size-1; i>=0; i--) {
			Term term = (Term)ifexpr.elementAt(i);

			// If not bound, bind now. Note: last term already bound
			if (term.target<ifend)
			{
				if (nterm.target>=ifend)
					term.trueJump = !nterm.trueJump;
				else if (term.target!=nterm.target)
					term.trueJump = !nterm.trueJump;
				else term.trueJump = nterm.trueJump;
			}

			// Jump to false-path means if-inst polarity does not match
			// source model (which is always if (true) then), so invert.
			if (!term.trueJump)
				term.opr = invertop[term.opr];
			nterm = term;
		}

		// Look for groups of terms that are skipped
		Term lterm = (Term)ifexpr.lastElement();
		for (int i=0; i<size-2; i++) {
			Term term = (Term)ifexpr.elementAt(i);

			// Skip to a term. Find it via its ifStart addr
			if (term.target<ifend)
			{
				term.subStart += 1;
				for (int j=i+2; j<size; j++) {
					nterm = (Term)ifexpr.elementAt(j);
					if (nterm.ifStart>term.target)
					{
						Term pterm = (Term)ifexpr.elementAt(j-1);
						pterm.subEnd += 1;
						break;
					}
				}
			}

			// Skip to end. Do parends to end if both OR's and AND's in
			// term range. However if this would cross existing right parend,
			// put parends from curr term to that term.
			else {
				int rpidx = analyzeParends(ifexpr, i, size-2);
				if (rpidx>size-2)
				{
					nterm = (Term)ifexpr.elementAt(i+1);
					nterm.subStart += 1;
					lterm.subEnd += 1;
				}
				else if (rpidx>0)
				{
					term.subStart += 1;
					nterm = (Term)ifexpr.elementAt(rpidx);
					nterm.subEnd += 1;
				}
			}
		}
	}								// End analyzeExpr

	// Analyzes range of terms idx1/idx2 inclusive.
	// If interm right parend, returns its index (early right parend needed).
	// If all terms end with same kind of jump, returns 0 (no parends needed)
	// Else returns idx2+1 (right parend needed after idx2+1)
	private int analyzeParends(Vector ifexpr, int idx1, int idx2)
	{
		boolean trues = false;
		boolean falses = false;

		// Scan the range
		for (int i=idx1; i<=idx2; i++) {
			Term term = (Term)ifexpr.elementAt(i);
			if (i>idx1 && term.subEnd>0)
				return i;
			if (term.trueJump)
				trues = true;
			else falses = true;
		}
		return trues!=falses ? 0 : idx2+1;
	}

	// Append source form of relational term. Also update opd1 as needed
	// vm = jvm context
	// ifsrc = src string gen so far
	// normal = false says invert relational operators
	private void append(Vmi vm, StringBuffer ifsrc, boolean normal)
	{
		int topr = normal ? opr : invertop[opr];

		// If it's boolean test vs. true/false, elim explicit relational opr.
		// If bool==false or bool!=1, then source is !opd1
		if (opd1.desc.equals("Z") && opd2.opr==Term.op_iconst)
		{
			if ((topr==Trek.DOIF_EQ && opd2.intVal==0)
				|| (topr==Trek.DOIF_NE && opd2.intVal==1))
			{
				opd1.src = "!" + vm.isolate(opd1.src, opd1.opr, Term.op_not);
				opd1.opr = Term.op_not;
			}
			ifsrc.append(opd1.src);
			opr = opd1.opr;
			opd2.src = "";
		}

		// Otherwise have explicit terms
		else {
			ifsrc.append(vm.isolate(opd1.src, opd1.opr, topr));
			ifsrc.append(oprs[topr]);
			ifsrc.append(vm.isolate(opd2.src, opd2.opr, topr));
			opd1.opr = topr;
		}
	}

	// Returns method-relative end of if-expr that inst is part of
	// top = top of opd stack at this time
	// condinfo = where to return cond expr and cond-end info, if applic
	// vm.inst = curr IF instruction of if-expr being analyzed
	int findIfEnd(Vmi vm, int top, int condinfo[])
	{
		Instruction curri = vm.inst;
		Statement stmt = curri.parent;
		int stend = stmt.start + stmt.insts.length;

		// Find 1st branch that goes outside of curr if-clause.
		// This means outside of curr stmt except when processing a?b:c
		while (curri!=null) {
			int targ = curri.getTarget();
			if (targ<0)
			{
				curri=curri.next();
				continue;
			}

			// Done if have branch outside of whole stmt
			if (targ<=stmt.start || targ>=stend)
				return stend;

			// Could be in middle of as-yet-unrecog a?b:c. Look for pattern
			// of: goto being inst before the curr if-target.
			Instruction targi = new Instruction(stmt, targ-stmt.start);
			Instruction ptargi = targi.prev();
			if (ptargi.type!=Vmi.it_branch)
				curri = ptargi;

			// Now find end of "a?". Because condit-exprs can
			// be nested, this is just past last if-inst that jumps
			// past the goto found above -- which could be curri.
			else {							// At such a pattern

				int lastend = stmt.start + curri.start + curri.size;
				Instruction endi;
				for (endi=curri; endi.start<ptargi.start; endi=endi.next()) {
					if (endi.getTarget()==targ)
						lastend = stmt.start + endi.start + endi.size;
				}

				// Return condit info and end of if (condEnd)
				condinfo[0] = ptargi.getTarget();
				condinfo[1] = top;
				return lastend;
			}
		}										// End loop
		throw new TrekException("Could not find If end");
	}

	// Store a term in an if-clause
	// inst = if-inst of curr term
	// ifexpr = accumulated if terms
	// ifend = address in method where if-clause ends
	// Notes:
	//	Final+ize target/opr if dest known now. Else do them in toIfResult.
	//	That is, set target to -1 if if-inst jumps to true, set it to 0 if
	//	if-inst jumps to false (and invert opr), or save target for later.

	void store(Instruction inst, Vector ifexpr, int ifend)
	{
		ifexpr.addElement(this);

		// If interior jump, cant analyze till have later if-insts
		Statement stmt = inst.parent;
		ifStart = stmt.start + inst.start;
		target = inst.getTarget();
		if (target>stmt.start && target<ifend)
			return;

		// 3 cases: a?b:c (ifend within stmt), loop-end, if-stmt
		// True-jump in a? is branch to "Then"
		if (ifend != stmt.start+stmt.insts.length)
			trueJump = ifend==target;

		// True-jump in loop if-clause is a branch up.
		// Canonicalize target to be like for if/a?
		else if (stmt.type>=Trek.ST_ENDLOOP)
			trueJump = target<=stmt.start;

		// True-jump in If is branch to "Then" (which can show up as
		// if-inst in if-stmt going directly to break/contin addr)
		// (Note: stn null & ST_UNKNOWN chks are done to allow
		// debugging calls to stmt.toString during explode of method)
		else {
			Statement stn = stmt.next();
			trueJump = ifend==target || (stn!=null && stn.insts.length==0
				&& (stn.type==Trek.ST_CONTINUE 
					|| stn.type==Trek.ST_BREAK
					|| stn.type==Trek.ST_UNKNOWN));
		}

		// For loops and optimized if (xxx) continue, true branch goes up.
		// Canonicalze target to be like normal if or a?b:c
		target = trueJump ? ifend : ifend+1;
	}											// End store

	// Create a term
	Term(Datum o1, int op, Datum o2)
	{
		opd1 = o1;
		opr = op;
		opd2 = o2;
	}

	// Analyze terms of if to gen if-expr source with proper &&, ||, and ()
	// vm = Vmi context to use while generating if-source
	// ifexpr = terms comprising the test expr
	// ifend = pc at which if expr ends
	// normal = false if want to generate inverted test expression
	String toIfResult(Vmi vm, Vector ifexpr, int ifend, boolean normal)
	{
		// Finally analyze where && and ||'s go.
		analyzeExpr(ifexpr, ifend);
		int size = ifexpr.size();

		// Convert terms to string
		StringBuffer ifsrc = new StringBuffer(100);
		for (int i=0; ; i++) {
			Term term = (Term)ifexpr.elementAt(i);

			// Parenthesize sub-expressions and output term
			while (term.subStart-- > 0) ifsrc.append("(");
			term.append(vm, ifsrc, normal);
			while (term.subEnd-- > 0) ifsrc.append(")");

			// Put and/or after this term (except for last term obviously)
			if (i>=size-1) break;
			if (!term.trueJump ^ !normal)
				ifsrc.append(" && ");
			else ifsrc.append(" || ");
		}

		// Done with terms and return src gen from them
		return ifsrc.toString();
	}

}

