//
// THIS SOFTWARE IS FURNISHED UNDER A LICENSE AND  MAY
// BE USED OR COPIED ONLY IN ACCORDANCE WITH THE TERMS
// OF SUCH LICENSE.
//
//
// COPYRIGHT (c) 1997 BY DIGITAL EQUIPMENT CORPORATION.
// ALL RIGHTS RESERVED.
//

package dec.trek;
import java.io.*;
import java.util.*;

//
// Trek models code as a hierarchy of objects. A trek is a journey thru
// a list of classfiles. The list of classfiles can be all those in a
// package, or all those referenced from the starting class (usually the
// module that contains the main program)
//
public class Trek implements VerboseSink
{
	// Version of the Trek class library. Version is incr for each new kit.
	// Version string is T or V major.minor.fix-kit. T means preliminary.
	private static final int Ver_num = 7;
	private static final String Ver_string = "V1.1";

	// The trek with trek instance data: a main trek points at itself,
	// a sub-trek points at a main trek. This allows a developer to
	// combine multiple doTrek-based trek apps into one larger app.
	// This feature requires that each public method in Trek class
	// operate on "mainTrek" data rather than on "this" data.
	// Coding-wise for Trek class this means the following for methods
	// that access Trek instance data explicitly:
	// 1) Very simple methods can specify mainTrek.data in-line.
	// 2) Other post-initialization methods must use a private analog
	//		xxx_ to do the real work (ie. they call mainTrek.xxx_(same))
	// 3) Methods that can be called before initTrek (ie. setXxx methods)
	//		have cases 1 & 2, but must deal with fact that mainTrek may
	//		still be null (via mainTrek!=null ? mainTrek : this)
	// 4) Main/sub specific methods such as initTrek, joinTrek, doXxx
	//		must be coded on a case-by-case basis.
	// 5) More generally, by time any non-public method in Trek class is
	//		reached, "this" must be a main trek.
	// 6) atXxx methods are called with sub-trek if applicable, but because
	//		of 1)-5) they will effectively operate on "mainTrek" data
	//		except for skipXxx
	Trek mainTrek;

	// Instruction cache for better performance while exploding method.
	// See Instruction.cache, at, next, prev for details.
	Vector icache;

	// Scan properties, from cmd line or setXxx
	private String inClasses;	// Internal qualified names of specified classes
	private String inClassesRaw = "<empty>";
	String inDirs = getDefaultClassPath();
								// Classpath for inClasses. "." means current dir
	private String inDirsRaw = "<default>";
								// For toString
	String outDir = "";			// Dir to output updated classes of prog to
								// "" means current dir
	private int scope;			// Recursion scope
	private String scopeRaw = "";
								// For toString
	int sortBy = Trek.CLORD_BYCLASS;
								// Sort by class@pkg or pkg.class
	private boolean verbose;	// Output progress msg's if set
	private VerboseSink verboseSink;	// Object to write verbose msg's to

	// Data for trek
	Vector cfiles;					// Classfiles of this trek
	private Vector cfiles2;			// Classfiles read via getClassFile
	private PrintWriter logChan;	// Logfile for per-classfile errors
	private int loggedError;		// # of errors logged
	private String packages;		// All packages in classList (;p1;...;pn;)
	private ClassFile progCf;		// the main class of trek if any
	private int skipObject;			// Skip rest of object at this level
	int wrote;						// In verbose mode, # of classfiles written
	
	static String lineSep;

	// Debug mode session-wide & pkg-visible
	static int debug;
	static final int Db_trace = 1;	// Do stacktrace on error
	static final int Db_stop = 2;	// Stop immediately on error
	static final int Db_divonly = 4;// Only divide method (dont Finish it)
	static final int Db_atbig = 8;	// Show major processing steps
	static final int Db_atsmall = 16;	// Show minor processing steps
	static final int Db_bldloc = 32;	// Build LVT always
	static final int Db_currloc = 64;	// Dump curr lvt (built or cleaned up)
	static final int Db_write = 128;	// Write classfile even if not modified
	static final int Db_raw = 256;		// Dump inst args non-symbolically
	static final int Db_at = Db_atbig + Db_atsmall;


	// Recursion scopes
	private static final int Sc_none = 0;	// Only specified classes
	private static final int Sc_pkg = 1;	// Only classes in specified pkg's
	private static final int Sc_p1 = 2;		// Pkg + 1 level outside spec pkg's
	private static final int Sc_user = 3;	// All non-java.* pkg's
	private static final int Sc_u1 = 4;		// That + 1 level
	private static final int Sc_all = 5;	// All referenced classes

	// Access modifier flags
	// Classes: public, final, super, interface, abstract
	// Methods: public, private, protected, ...
	// Fields: ...
	public static final int ACC_PUBLIC = 0x1;
	public static final int ACC_PRIVATE = 0x2;
	public static final int ACC_PROTECTED = 0x4;
	public static final int ACC_STATIC = 0x8;
	public static final int ACC_FINAL = 0x10;
	public static final int ACC_SUPER = 0x20;		// Note overloading
	public static final int ACC_SYNCHRONIZED = 0x20;
	public static final int ACC_THREADSAFE = 0x40;	// Note overloading
	public static final int ACC_VOLATILE = 0x40;
	public static final int ACC_TRANSIENT = 0x80;
	public static final int ACC_NATIVE = 0x100;
	public static final int ACC_INTERFACE = 0x200;
	public static final int ACC_ABSTRACT = 0x400;
	public static final int ACC_STRICT = 0x800;

	// doIf comparison options
	public static final int DOIF_EQ = 0;
	public static final int DOIF_NE = 1;
	public static final int DOIF_LT = 2;
	public static final int DOIF_GT = 3;
	public static final int DOIF_LE = 4;
	public static final int DOIF_GE = 5;

	// Dump-classfile options
	public static final int DUMP_ATTR=1;
	public static final int DUMP_INST=2;
	public static final int DUMP_POOL=4;
	public static final int DUMP_STMT=8;

	// Types of instructions.
	public static final int IN_BINOP = 0;
	public static final int IN_DUP = 1;
	public static final int IN_LOADELEMENT = 2;
	public static final int IN_LOADCONSTANT = 3;
	public static final int IN_LOADFIELD = 4;
	public static final int IN_LOADLOCAL = 5;
	public static final int IN_NEW = 6;
	public static final int IN_STOREELEMENT = 7;
	public static final int IN_STOREFIELD = 8;
	public static final int IN_STORELOCAL = 9;
	public static final int IN_UNOP = 10;
	public static final int IN_BRANCH = 11;
	public static final int IN_CALL = 12;
	public static final int IN_IF = 13;
	public static final int IN_RETURN = 14;
	public static final int IN_SWITCH = 15;
	public static final int IN_OTHER = 16;

	// Classfile sort order
	public static final int CLORD_BYCLASS = 0;
	public static final int CLORD_BYPACKAGE = 1;

	// Types of statements.
	// Code inserted via addXxx Trek call
	static final int ST_BEFORE = -8;
	static final int ST_AT = -7;
	static final int ST_AFTER = -6;
	static final int  ST_INSERTED = -6;

	// Real code that is special in some way
	static final int ST_ENTER = -5;			// Any start of constructor
	static final int ST_EXIT = -4;			// Any return stmt
	static final int ST_PIECE = -3;			// Piece of stmt split by addXxx
	static final int ST_PRIMORD = -2;		// All the code (before explode)
	public static final int ST_UNKNOWN = -1; // Becomes break, continue, or else

	// Those ending in semi-colon
	public static final int ST_ASSIGN = 0;
	public static final int ST_BREAK = 1;
	public static final int ST_CONTINUE = 2;
	public static final int ST_RETURN = 3;
	public static final int ST_EMPTY = 4;
	public static final int ST_THROW = 5;
	public static final int ST_CALL = 6;
	public static final int ST_LOCAL = 7;
	public static final int  ST_SEMI = 7;

	public static final int ST_TRY = 8;
	public static final int ST_CATCH = 9;
	public static final int ST_FINALLY = 10;
	public static final int ST_IF = 11;
	public static final int ST_THEN = 12;
	public static final int ST_ELSE = 13;
	public static final int ST_FOR = 14;
	public static final int ST_WHILE = 15;
	public static final int ST_DO = 16;
	public static final int ST_SWITCH = 17;
	public static final int ST_CASE = 18;
	public static final int ST_LABEL = 19;
	public static final int ST_SYNC = 20;

	// All the blk-end statements
	public static final int  ST_END = 21;
	public static final int ST_ENDTHEN = 21;
	public static final int ST_ENDELSE = 22;
	public static final int ST_ENDTRY = 23;
	public static final int ST_ENDCATCH = 24;
	public static final int ST_ENDFINALLY = 25;
	public static final int ST_ENDSYNC = 26;

	// Breakable blks: switches, labels, loops
	public static final int  ST_ENDBREAKABLE = 27;
	public static final int ST_ENDSWITCH = 27;
	public static final int ST_ENDLABEL = 28;

	// End of loops
	public static final int  ST_ENDLOOP = 29;
	public static final int ST_ENDDO = 29;
	public static final int ST_ENDFOR = 30;
	public static final int ST_ENDWHILE = 31;

	// The default (noop) action routines for a doTrek scan
	protected void atReference(ClassFile cf, Member mem) {}
	protected void at(Field field) {}
	protected void at(Local loc) {}
	protected void at(Instruction inst) {}

	protected void atStartOf() {}
	protected void atStartOf(ClassFile cf) {}
	protected void atStartOf(Method method) {}
	protected void atStartOf(Statement stmt) {}

	protected void atEndOf() {}
	protected void atEndOf(ClassFile cf) {}
	protected void atEndOf(Method method) {}
	protected void atEndOf(Statement stmt) {}

	// Add this class to program's list of classes, and then recurses if applic
	// classname = internal-fmt (/'s not .'s) fully qualified class name.
	// This method is slave of initTrek.
	// hash = table of seen class names
	// Returns the class it adds or throws exception
	// Notes:
	//	hashtable used to decide whether to read file rather than Trek.lookup
	//	because latter only includes successfully read files, and part of
	//	goal is to not re-open files that failed to read in the 1st time.
	//	read() enforces name of file matching case of class name. So if
	//	user specified wrong case in a base class, result is file not found.
	private ClassFile addClasses(String classname, Hashtable qhash)
	{
		writeVerbose("Reading " + classname.replace('/', '.'));

		// Create the object for this classfile & hash it so that it is
		// not processed mult times. If read it successfully, add it to trek.
		ClassFile cf = new ClassFile(this);
		qhash.put(classname, classname);
		cf.read(inDirs, classname);
		insertSorted(cfiles, cf);

		// Scan referenced classes to recursively read further classes.
		for (TypeRef ref=cf.firstTypeRef(); ref!=null; ref=ref.next()) {
			String refclass = ref.toInternalString();

			// Read a reffed classs only if it is not an array ref,
			// it is within requested scope of recursion, and it has 
			// not already been seen
			if (!refclass.startsWith("[")
				&& isInScope(classname, refclass)
				&& !qhash.containsKey(refclass))
			{
				try {
					addClasses(refclass, qhash);
				}

				// Log problem with this classfile, and continue with rest
				catch (Exception e) {
					String refext = refclass.replace('/', '.');
					logError("C="+refext+";D=initializing;", e);
				}
			}
		}									// End loop
		return cf;
	}								// End addClasses

	// Output the msg to System.out
	public void atVerbose(String msg)
	{
		Trek.println(System.out, msg);
	}

	// Return classFile at specified index
	public ClassFile classFileAt(int cfidx)
	{
		if (cfidx<0 || cfidx>=mainTrek.cfiles.size())
			return null;
		else return (ClassFile)mainTrek.cfiles.elementAt(cfidx);
	}

	// Remove all code-insertions from in-memory representation of program
	public void cleanTrek()
	{
		// Scan classfiles, removing inserted pool entries and code
		for (ClassFile cf=firstClassFile(false); cf!=null; cf=cf.next(false)) {
			cf.clean();
		}
	}

	// Trek thru code of specified method (slave of doTrek)
	private void doMethod(Method meth, int levels)
	{
		// Trek thru method's locals
		for (Local loc=meth.firstLocal(); loc!=null; loc=loc.next()) {
			at(loc);
			if (mainTrek.skipObject>=0 && mainTrek.skipObject<=2) return;
			mainTrek.skipObject = -1;
		}

		// Trek thru method's statements
		Statement stmt = meth.firstStatement();
		for (; stmt!=null; stmt=stmt.next()) {
			atStartOf(stmt);

			// Trek thru statement's instructions
			if (levels>=4 && mainTrek.skipObject<0)
			{
				Instruction inst = stmt.firstInstruction();
				for (; inst!=null; inst=inst.next()) {
					at(inst);
					if (mainTrek.skipObject>=0 && mainTrek.skipObject<=3) break;
					mainTrek.skipObject = -1;
				}
			}

			// Processing when scan past curr stmt
			if (mainTrek.skipObject<0)
				atEndOf(stmt);
			if (mainTrek.skipObject>=0 && mainTrek.skipObject<=2) return;
			mainTrek.skipObject = -1;
		}						// End statements
	}

	// For each class that member is referenced from, call atReference.
	public void doReferences(Member mem)
	{
		// Fastest case is private member. Just scan curr classfile
		ClassFile cf = mem.getClassFile();
		if (mem.hasModifier(ACC_PRIVATE))
		{
			if (cf.references(mem))
				atReference(cf, mem);
			return;
		}

		// Get each class and decide whether to process it:
		// For public members, all classes match.
		// For protected members, descendant classes and cf match.
		// For package-scope, package name needs to match
		String pkg = cf.toPackage();
		for (ClassFile ccf=firstClassFile(false); ccf!=null; ccf=ccf.next(false)) {
			if (mem.hasModifier(ACC_PUBLIC))
			{
				if (ccf.references(mem))
					atReference(ccf, mem);
			}
			else if (mem.hasModifier(ACC_PROTECTED))
			{
				if ((cf==ccf || cf.isAncestor(ccf)) && ccf.references(mem))
					atReference(ccf, mem);
			}
			else if (pkg.equals(ccf.toPackage()))
			{
				if (ccf.references(mem))
					atReference(ccf, mem);
			}
			if (mainTrek.skipObject==0) return;
		}
	}

	// Scan thru entire hierarchy of program identified by inClasses,
	// reading program's classfiles now if havent already done so.
	// If global error occurs, caller will get TrekException.
	// File-specific errors are gathered in currdir\trek.log
	public void doTrek()
	{
		doTrek(4);
	}

	// Scan the specified number of levels
	// levels = 1 (classfiles) thru 4 (instructions). <1 is 1. >4 is 4.
	public void doTrek(int levels)
	{
		boolean inited;

		// Read program if havent yet, and start processing at root,
		// after binding to user-supplied action routines.
		if (mainTrek==null)
		{
			initTrek();
			inited = true;		// Scan will itself close trek
		}
		else inited = false;

		// Throw exception if nothing to scan
		ClassFile cf1 = firstClassFile();
		if (cf1==null)
			throw new TrekException("Trek's class list is empty");

		// Trek thru the relevant classes
		mainTrek.skipObject = -1;
		atStartOf();
lev1:
		for (ClassFile cf=cf1; cf!=null; cf=cf.next()) {
			atStartOf(cf);

			// Go into fields/methods if this is in scope of scan
			if (levels>=2 && mainTrek.skipObject<0)
			{
				// Scan current classfile's fields
				for (Field fld=cf.firstField(); fld!=null; fld=fld.next()) {
					at(fld);
					if (mainTrek.skipObject==0) return;
					if (mainTrek.skipObject==1) continue lev1;
					mainTrek.skipObject = -1;
				}

				// Scan current classfile's methods
				for (Method meth=cf.firstMethod(); meth!=null; meth=meth.next()) {
					atStartOf(meth);

					// Do work for method
					if (levels>=3 && mainTrek.skipObject<0)
						doMethod(meth, levels);
					if (mainTrek.skipObject<0)
						atEndOf(meth);
					if (mainTrek.skipObject==0) return;
					if (mainTrek.skipObject==1) continue lev1;
					mainTrek.skipObject = -1;
				}							// End methods
			}

			// End of classfile
			if (mainTrek.skipObject<0)
				atEndOf(cf);
			if (mainTrek.skipObject==0) return;
			mainTrek.skipObject = -1;
		}								// End classfiles

		// Scan completed
		atEndOf();
		if (inited)
			endTrek();
	}									// End doTrek

	// End just the object at the specified level
	public void endTrek(int level)
	{
		if (level<1 || level>4)
			throw new TrekException("Invalid trek object level");
		mainTrek.skipObject = level;
	}

	// Cleanup the program's storage and write out changes
	// outdir = path of root dir to write modified classfiles to
	public void endTrek()
	{
		// If trek not currently inited, just succeed quick
		if (mainTrek!=null)
			mainTrek.endTrek_();
	}
	private void endTrek_()
	{
		// Save changes and free resources
		saveTrek();
		cfiles = cfiles2 = null;
		icache.setSize(0);
		icache = null;
		mainTrek = null;
		skipObject = 0;					// Set skip-level to whole trek

		// Close error log
		logChan.close();
		logChan = null;
		int errs = loggedError;
		loggedError = 0;

		// Report if any problems logged
		if (errs > 0)
		{
			String msg = errs==1 ? " error" : " errors";
			throw new TrekException(
				"Warning - " + errs + msg + " written to trek.log");
		}
	}									// End of endTrek

	// Returns ClassFile object identified by data type, or null.
	// dtype = a fully qualified external class name or internal class name.
	public ClassFile findClassFile(String dtype)
	{
		return mainTrek.findClassFile_(dtype);
	}
	private ClassFile findClassFile_(String dtype)
	{
		// Consis chks
		if (cfiles==null)
			throw new TrekException("Trek not initialized yet");

		// See if it is classfile that was part of trek
		dtype = dtype.replace('.', '/');	// Need internal form
		ClassFile cf = (ClassFile)Trek.lookup(cfiles, dtype);
		if (cf!=null)
			return cf;

		// It's not. Perhaps it was read/created here earlier
		// If simply internal-to-pkg lookup which failed, return null
		cf = (ClassFile)Trek.lookup(cfiles2, dtype);
		if (cf!=null)
			return cf;
		else return null;
	}

	// Return 1st classfile of this program
	public ClassFile firstClassFile()
	{
		if (mainTrek!=null)
			return mainTrek.firstClassFile_(true);
		else return null;
	}
	ClassFile firstClassFile(boolean ispub)
	{
		return mainTrek.firstClassFile_(ispub);
	}
	private ClassFile firstClassFile_(boolean ispub)
	{
		ClassFile cf;

		if (cfiles==null || cfiles.size()==0)
			return null;
		else cf = (ClassFile)cfiles.firstElement();
		if (ispub)
			writeVerbose("Scan at " + cf.toQualifiedName());
		return cf;
	}

	// Returns ClassFile object identified by data type or throws excep.
	// dtype = a fully qualified external class name or internal class name
	public ClassFile getClassFile(String dtype)
	{
		return mainTrek.getClassFile_(dtype);
	}
	private ClassFile getClassFile_(String dtype)
	{
		// See whether already in memory if trek initialized
		ClassFile cf;
		if (cfiles!=null)
		{
			cf = findClassFile(dtype);
			if (cf!=null)
				return cf;
		}

		// Try to read it now. (NoClassDef occurs in v1.0.2 if
		// class is in zip or jar file)
		cf = new ClassFile(this);
		dtype = dtype.replace('.', '/');	// Need internal form
		cf.read(inDirs, dtype);
		insertSorted(cfiles2, cf);
		return cf;
	}

	// Can be overridden as way to process user-defined arg.
	// argv = array of words passed to getCmdLine
	// i = index of arg not recognized by getCmdLine
	// Returns arg to resume processing at, or 0 for not processed
	protected int getCmdLineArg(String argv[], int i)
	{
		return 0;
	}

	// Materialize arguments of scan from (cmd line) array,
	// validating the trek-defined args
	public void getCmdLine(String argv[])
	{
		if (mainTrek!=null)
			throw new TrekException("Trek has already been initialized or joined");
		if (argv.length < 1)
			throw new TrekException("Program's class not specified");

		// Init result vars
		String inclasses = argv[0];
		String indirs = null;
		String outdir = null;
		String rawscope = null;

		// Process optional args (many are in pairs: switch & value)
		for (int i=1; i<argv.length; ) {

			// Let user process one of his own args if present
			int newi = getCmdLineArg(argv, i);
			if (newi > 0)
			{
				i = newi;				// User ate this arg
				continue;
			}

			// Verbose mode set?
			String arg = argv[i];
			if (arg.equals("-verbose"))
			{
				verbose = true;
				i += 1;
			}

			// classpath for inClasses
			else if (arg.equals("-ip"))
			{
				if (i>argv.length-2)
					throw new TrekException("No value specified for -ip");
				indirs = argv[i+1];
				i += 2;
			}

			// Where to output modified classfiles to
			else if (arg.equals("-op"))
			{
				if (i>argv.length-2)
					throw new TrekException("No value specified for -op");
				outdir = argv[i+1];
				if (!outdir.endsWith(File.separator))
					outdir += File.separator;
				i += 2;
			}

			// Chk for scope of scan
			else if (arg.equals("-sc"))
			{
				if (i>argv.length-2)
					throw new TrekException("No value specified for -sc");
				rawscope = argv[i+1];
				i += 2;
			}

			// Set debug mode for error reporting if desired
			else if (arg.startsWith("-debug"))
			{
				if (arg.length()==6)
					debug = Db_trace;
				else {
					arg = arg.substring(6);
					debug = Integer.parseInt(arg);
				}
				i += 1;
			}

			// If to here, no one recog this arg.
			// Msg contains i+1 cause [0] means arg1 to user.
			else throw new TrekException(
				"Argument" + (i+=1) + " (" + arg + ") not recognized");
		}								// End loop

		// Check/default args
		setClassList(inclasses, indirs, rawscope);
		setOutPath(outdir);
	}									// End getCmdLine

	// Return default overall classpath.
	private String getDefaultClassPath()
	{
		// Get classpath (in jdk1.2, this is just user classpath)
		String dirs = System.getProperty("java.class.path");
		if (dirs==null)
			dirs = ".";

		// Get boot and extension if 1.2 envir
		String pre = System.getProperty("java.ext.dirs");
		if (pre!=null)
			dirs = pre + ";" + dirs;
		pre = System.getProperty("sun.boot.class.path");
		if (pre!=null)
			dirs = pre + ";" + dirs;
		return dirs;
	}

	// Returns main method of the program, if any
	public Method getMainMethod()
	{
		// If un-inited trek or no main class set for trek, just ret null
		if (mainTrek==null) return null;
		ClassFile cf = mainTrek.progCf;
		if (cf==null) return null;
		return cf.lookupMethod("main", "([Ljava/lang/String;)");
	}

	// Set mask corresponding to access-flag tokens in flaglist
	// flaglist = 1 or more access-flag keywords sep by ;'s
	public static int getModifiers(String flaglist)
	{
		int flags = 0;

		// Get a keyword out of flaglist
		int cp[] = {0, -1};
		while (true) {
			String token = Trek.nextItem(flaglist, cp);
			if (token==null) break;

			// Chk each access keyword
			if (token.equals("public")) flags |= ACC_PUBLIC;
			else if (token.equals("private")) flags |= ACC_PRIVATE;
			else if (token.equals("protected")) flags |= ACC_PROTECTED;
			else if (token.equals("static")) flags |= ACC_STATIC;
			else if (token.equals("final")) flags |= ACC_FINAL;
			else if (token.equals("volatile")) flags |= ACC_VOLATILE;
			else if (token.equals("synchronized")) flags |= ACC_SYNCHRONIZED;
			//else if (token.equals("threadsafe")) flags |= ACC_THREADSAFE;
			else if (token.equals("transient")) flags |= ACC_TRANSIENT;
			else if (token.equals("native")) flags |= ACC_NATIVE;
			else if (token.equals("strictfp")) flags |= ACC_STRICT;
			else if (token.equals("interface")) flags |= ACC_INTERFACE;
			else if (token.equals("abstract")) flags |= ACC_ABSTRACT;
			else throw new TrekException(token + " is not a valid access keyword");
		}
		return flags;
	}

	// Return the object corresponding to the specified unique name, or null.
	// name = object name to lookup (obj must be in trek):
	//	Qual-class-name or  qualclass.field or qualclass.methodref
	//	or qualclass.methodref.local#addr or qualclass.methodref#addr
	//	methodref is method("method's descriptor"). Desc enclosed in
	//	quotes since it can contain )'s (which would be hard to parse).
	//	addr is stmtnumber (or maybe in future num+num for an inst)
	public Object getObject(String name)
	{
		return mainTrek.getObject_(name);
	}
	private Object getObject_(String name)
	{
		int memidx;

		// If no (, whole name may be class name. 
		int lpidx = name.indexOf("(\"");
		if (lpidx<0)
		{
			try {
				return getClassFile(name);	// Is classname if no throw
			}
			catch (TrekException trex) {
			}
			memidx = name.lastIndexOf('.');
		}
		else memidx = name.lastIndexOf('.', lpidx);

		// Whole name is not a class name, so must be at least 1 dot
		// Qcname must be valid classname, so fail if not
		if (memidx<0) return null;			// No member piece
		String qcname = name.substring(0, memidx);
		ClassFile cf;
		try {
			cf = getClassFile(qcname);
		}
		catch (TrekException trex) {
			return null;			// Bad class name
		}

		// Got the classfile now. If fld, all done on lookup
		if (lpidx<0)
			return Trek.lookup(cf.fields, name.substring(memidx+1));

		// For method, isolate name and descriptor. Then lookup.
		int rpidx = name.lastIndexOf("\")") + 1;
		if (rpidx < 0) return null;
		String methname = name.substring(memidx+1, lpidx);
		String desc = name.substring(lpidx+2, rpidx-1);
		Method meth = cf.lookupMethod(methname, desc);
		if (meth==null) return null;

		// Is method what we want? If not, see if local or stmt
		// Either way, name ends in stmt number
		int namelen = name.length();
		if (rpidx==namelen-1)
			return meth;
		int pndidx = name.lastIndexOf('#');
		if (pndidx<0 || pndidx>=namelen-1) return null;

		// Map from stmt number to statement. First stmt in a method is 1.
		// Note: 0 happens for local that is in arglist: meth().loc#0.
		int stnum = Integer.parseInt(name.substring(pndidx+1));
		Statement st = meth.firstStatement();
		for (int i=1; i<stnum; i++)
		{
			if (st==null) return null;		// Shouldnt happen
			st = st.next();
		}

		// Id is for stmt if # is right after meth(xxx)
		if (pndidx==rpidx+1)
			return st;
		if (name.charAt(rpidx+1)!='.')
			return null;			// Bad syntax (. wasnt after ))

		// Syntactically is local. Isolate name and get it.
		// Stmt# of 0 for local means a method arg or a catch arg.
		// Get can return null if the name is not defined at defpc
		String locname = name.substring(rpidx+2, pndidx);
		int defpc = stnum==0 ? 0 : st.start + st.insts.length;
		return meth.getLocal(locname, defpc);
	}

	// Get target obj of variable ref, wrt inst
	public Object getObject(String name, Instruction wrt)
	{
		return new Code(null, wrt.parent, wrt, Trek.ST_AT).getObject(name);
	}

	// Get target obj of variable ref, wrt stmt
	// Notes:
	//	Use after since in x=a, x's scope starts at end of stmt
	public Object getObject(String name, Statement wrt)
	{
		return new Code(null, wrt, null, Trek.ST_AFTER).getObject(name);
	}

	// Version of the kit
	public static int getVersion()
	{
		return Ver_num;
	}

	// Read all the classfiles of the program
	// inClasses = fully qualified internal names of classes on cmd line
	// inDirs = the dirs it classfiles can be in
	public void initTrek()
	{
		// Init/check trek state.
		if (mainTrek!=null)
			throw new TrekException("Trek already initialized");
		mainTrek = this;

		// Line separator (eg CRLF) varies between systems
		lineSep = System.getProperty("line.separator");
		if (lineSep==null)
			lineSep = "\n";

		// Set default verbose-msg outputter if need be
		if (verboseSink==null)
			verboseSink = this;

		// Create logfile for per-object errors in curr dir if none set yet.
		// If cannot, throw the failure, aborting the trek.
		loggedError = 0;
		if (logChan==null)
		{
			try {
				logChan = new PrintWriter(new FileWriter("trek.log"), true);
			}
			catch (Exception e) {
				mainTrek = null;
				e.printStackTrace();
				throw new TrekException("Could not open trek.log");
			}
		}
		println(logChan, toVersion());

		// Alloc space for classfiles & explode-method inst cache.
		// Note: cfiles non-null means body of trek started
		cfiles = new Vector(50, 25);
		cfiles2 = new Vector();
		icache = new Vector(1000, 500);

		// Chk if anything to scan
		packages = ";";
		if (inClasses==null)
			return;

		// Extract packages out in inClasses for scope processing.
		// Also indic if 1st item has wildcarding in it.
		boolean wild1 = false;
		int cp[] = {0, -1};
		for (int i=0; ; i++) {
			String classname = Trek.nextItem(inClasses, cp);
			if (i==0 && classname.indexOf('*')>=0)
				wild1 = true;
			if (classname==null) break;
			packages += toPackage(classname) + ";";
		}

		// Recursively read entire program. Avoid reading in same class more
		// than once by chking if name already in hashtable of read-in classes.
		// Scan thru semi-separated list of class names. If cant read any
		// of the explicitly specified classes, an exception is thrown.
		writeVerbose(toString_(false));
		Hashtable hash = new Hashtable(50);
		cp[0] = 0;
		while (true) {
			String classname = Trek.nextItem(inClasses, cp);
			if (classname==null) break;
			if (hash.containsKey(classname)) continue;

			// Define explicit 1st class (if any) as main class
			ClassFile cf = addClasses(classname, hash);
			if (progCf==null && !wild1)
				progCf = cf;
		}
		ClassFile.bindIndexes(this);

		// Output summary message if mult and user requested
		int read = cfiles.size();
		if (read>1)
			writeVerbose("Read " + read + " classfiles");
	}

	// Insert a gotten classfile into trek.
	// cf = classfile
	// Notes:
	//	Updates order of objects in firstCF/next loop
	public synchronized void insertClassFile(ClassFile cf)
	{
		// Chk if this classfile is associated with curr trek
		if (cf.parent!=mainTrek)
			throw new TrekException
				(cf.toQualifiedName() + " not associated with this trek");

		// Remove from cfiles2 if there, insert in cfiles, and re-sort
		if (!mainTrek.cfiles2.removeElement(cf))
			throw new TrekException("Classfile already in this trek");
		insertSorted(mainTrek.cfiles, cf);
		ClassFile.bindIndexes(mainTrek);
	}

	// Insert an object into a sorted vector, or look one up. If mult
	// objects with same name (eg. overloaded methods), lookup finds 1st.
	// vec = vector to process (can start empty but must be non-null)
	// obj = object to insert/lookup (must implement Sortable interface)
	// insert = true for insert, false for lookup
	// Notes:
	//	To sort an arbitrary type of object, (define)use a jacket class:
	//	class Jacket implements Sortable
	//	  Object obj;
	//	  String key;
	//    public int compareTo(Sortable sor) {return key.compareTo(sor.key)}
	//  and insert obj/key to be sorted into Jacket object & call insertSorted
	static Sortable insertLookup(Vector vec, Object obj, boolean insert)
	{
		// Setup to chk against middle item of vector
		int idx;
		int lb = -1;			// Key is known to be > than key at this idx
		int ub = vec.size();	// Key is known to be < than key at this idx

		// Find where this one fits by successively halving table
		while (true) {
			idx = (ub + lb)/2;

			// Bump into bnd we already know about?
			if (idx == ub)
				break;
			if (idx == lb)
			{
				idx += 1;
				break;
			}

			// See which direction to move for next narrowing
			Sortable cobj = (Sortable)vec.elementAt(idx);
			int result = cobj.compareTo(obj);

			// Done if exact match. If doing lookup, return the 1st
			// occurrence of the matched name.
			if (result==0)
			{
				if (insert) break;
				for (int i=idx-1; i>=0; i--) {
					Sortable pobj = (Sortable)vec.elementAt(i);
					if (pobj.compareTo(obj)!=0) break;
					cobj = pobj;
				}
				return cobj;
			}

			// Move to next narrower "half" of table
			if (result < 0)
				lb = idx;
			else ub = idx;
		}									// End of loop

		// Hit a bound, insert the new object at this position
		if (insert)	vec.insertElementAt(obj, idx);
		return null;
	}

	static void insertSorted(Vector vec, Sortable obj)
	{
		insertLookup(vec, obj, true);
		return;
	}										// End of insertSorted

	// Returns true if toclass is within scope of trek
	// fromclass = fully qualified class name of referencing class
	// toclass = fully qualified class name of reffed class
	// May be called for main trek only.
	private boolean isInScope(String fromclass, String toclass)
	{
		switch (scope) {
		case Sc_none:
			return false;
		case Sc_pkg:
			String topkg = ";" + toPackage(toclass) + ";" ;
			return packages.indexOf(topkg)>=0;
		case Sc_p1:
			String frompkg = ";" + toPackage(fromclass) + ";" ;
			return packages.indexOf(frompkg)>=0;
		case Sc_user:
			return !toclass.startsWith("java/");
		case Sc_u1:
			return !fromclass.startsWith("java/");
		case Sc_all:
			return true;
		default:
			throw new TrekException("Invalid scope");
		}
	}

	// Join a main trek. This allows multiple method vectors to share one
	// set of program data. That is, it allows a developer to aggregate
	// simple apps that use doTrek (and thus atXxx methods).
	// trek = trek to join, or null to unjoin
	public void joinTrek(Trek trek)
	{
		// Unjoin, join to uninited trek, and join to inited trek ok
		if (trek==null || trek.mainTrek==null || trek.mainTrek==trek)
			mainTrek = trek;
		else throw new TrekException("Trek being joined cannot be a sub-trek");
	}

	// Return last classfile of this program
	public ClassFile lastClassFile()
	{
		if (mainTrek!=null)
			return mainTrek.lastClassFile_();
		else return null;
	}
	private ClassFile lastClassFile_()
	{
		ClassFile cf;

		if (cfiles==null || cfiles.size()==0)
			return null;
		else cf = (ClassFile)cfiles.lastElement();
		writeVerbose("Scan at " + cf.toQualifiedName());
		return cf;
	}

	// Output msg line(s) to log file
	// context = info about what error occurred for
	// e = exception that just occurred
	void logError(String context, Exception e)
	{
		String msg = e.toString();
		TrekException tex = e instanceof TrekException ? (TrekException)e : null;

		// If live error, log original exception info). 
		// (Cant save. Later use of stack will clobber if dont output now).
		// Setup tex to store context info.
		if (tex==null || tex.context.length() <= 0)
		{
			if ((debug&Db_trace)==0)
				logChan.println(msg);
			else e.printStackTrace(logChan);
			tex = new TrekException("");	// Create context placeholder
		}

		// Merge new context info. Rethrow if not at top level. 
		tex.context += context;
		if (!context.startsWith("C=")) throw tex;

		// Else put out "while" msg, stopping immed if so requested
		loggedError += 1;
		String while_ = toPrettyError(tex);
		logChan.println(while_);
		if ((debug&Db_stop)!=0)
		{
			logChan.close();
			System.out.println("Aborted" + while_);
			System.exit(1);
		}
	}

	// Output error msg line(s) to log file
	// msg = msg to output
	void logError(String msg)
	{
		loggedError += 1;
		logChan.println(msg);
	}

	// Output non-error msg line(s) to log file
	// msg = msg to output
	void logMsg(String msg)
	{
		logChan.println(msg);
	}

	// Returns obj if it is in the vector
	static Sortable lookup(Vector vec, Object obj)
	{
		if (vec==null) return null;
		return insertLookup(vec, obj, false);
	}

	// Returns the next item in ch-separated list. Items in the
	// list may contain quoted strings and parentheses (ie. ch's 
	// in quoted string and in parentheses are ignored)
	// list = the semi-separated string
	// cp = 2-item array containing the position of the next item to get
	//		(init must be 0) and the char to search for. The latter has 2
	//		special cases. 0 means use ; and -1 means use ;'s and/or +'s
	public static String nextItem(String list, int cp[])
	{
		if (cp[0] < 0 || list==null || list.length()==0)
			return null;

		// Establish match-char(s) and parse out the next item
		String match;
		switch (cp[1]) {
		case 0:
			match = ";";
			break;
		case -1:
			match = ";+";
			break;
		default:
			match = String.valueOf(cp[1]);
		}
		int matidx = Datum.smartIndexOf(list, 0, match, cp[0]);

		// Return next piece or to end depending on whether ch found
		String item;
		if (matidx==-1)
		{
			item = list.substring(cp[0]);
			cp[0] = -1;
		}
		else {
			item = list.substring(cp[0], matidx);
			cp[0] = matidx + 1;
		}
		return item;
	}

	// Print a long string literal as "xxx"<RET>+"xxx" etc
	// chan = where to output to
	// lit = 
	private static int printLiteral(PrintWriter chan, String lit)
	{
		// Establish bounds of quoted string
		int qstart = lit.indexOf('"');
		int qend = 0;
		if (qstart>=0)
		{
			do {
				qend = lit.indexOf('"', qstart + qend + 1);
				if (qend<0) break;
			} while (lit.charAt(qend-1)=='\\');
		}

		// Not terminated, just print whole thing.
		int litlen = lit.length();
		if (qend<=0)
		{
			chan.println(lit);
			return litlen;
		}

		// Incl trailing char if any (probably space, rparen, or ;)
		// If need to divide and there is trailing char that is not ;
		// treat it as sub-expr because of +'s and enclose in parens
		boolean subex;
		char end;
		if (qend<litlen-1)
		{
			qend += 1;
			end = lit.charAt(qend);
			subex = end!=';';
		}
		else {
			end = '"';
			subex = false;
		}

		// Print quoted string, watching out for imbedded \xxx's.
		int pos1 = 0;
		int pos2;
		while (true) {
			if (pos1!=0)
				chan.print("+ \"");
			pos2 = pos1 + 60;
			if (pos2>qend-15)
			{
				chan.print(lit.substring(pos1, qend));

				// If divided it and it is sub-expr, fin enclosing in parens
				if (pos1!=0 && subex)
					chan.print(")");
				chan.println(end);
				return qend+1;
			}

			// Dont split in middle of \u0000 or \x.
			else {
				String bscheck = lit.substring(pos2, pos2+5);
				int bs = bscheck.indexOf('\\');
				if (bs < 0)
					pos2 = pos2+5;
				else if (bs>0)
					pos2 += bs;

				// Bscheck starts with \. If have mult consec \, must split 
				// after even #. So find 1st to identify valid split point
				else {
					int prepos;		// 1st non-slash position
					for (prepos=pos2-1; prepos>=0; prepos--) {
						if (lit.charAt(prepos)!='\\')
							break;
					}
					int nslash = pos2-1 - prepos;	// # \'s before pos2's \
					if (nslash%2==1)
						pos2 += 1;
				}
			}

			// Output this line of quoted string, and move to next.
			// If 1st line and sub-expr, insert opening paren
			if (pos1==0 && subex)
			{
				chan.print(lit.substring(pos1, qstart));
				chan.println("(" + lit.substring(qstart, pos2) + "\"");
			}
			else chan.println(lit.substring(pos1, pos2) + "\"");
			pos1 = pos2;
		}
	}

	// Print src, splitting long src into multiple lines of output
	private static final String split[] = {",(;", ":?|&+*-<>=", "[."};
	static void printPretty(PrintWriter chan, String indent, String src)
	{
		int indlen = Math.min(indent.length(), 24);
		int srclen = src.length();
		int maxlen = 85;			// Allow single line to be longer

		// Find places to split long line if applic
		while (indlen+srclen>maxlen) {
			int pos = 0;
			int bestbad = 0;
			int pass = 0;
			while (true) {

				// Chk for Space insures only match binary operators
				// and only 1st char of things like == and &&.
				pos = Datum.smartIndexOf(src, -1, split[pass], pos);
				if (pos>0 && pass==1 && src.charAt(pos-1)!=' ')
				{
					pos += 1;
					continue;
				}

				// Decide what to do based on search result
				// 1) find failure or too near end of string
				if (pos<0 || pos>srclen-6 || pos>120)
				{
					pass += 1;
					if (pass<3)
					{
						pos = 0;
						continue;
					}
					else {
						pos = bestbad>0 ? bestbad : srclen;
						break;
					}
				}

				// 2) In target area. If 1st pass (primary punct) and
				// not a left parend, brk after the char (eg. after comma)
				boolean postsplit = pass==0 && src.charAt(pos)!='(';
				if (pos>=60-indlen)
				{
					if (postsplit) pos += 1;
					break;
				}

				// 3) Match too early (but keep biggest as last resort)
				else {
					if (bestbad<pos) bestbad = postsplit ? pos+1 : pos;
					pos += 1;
				}
			}						// End find-a-split-pos loop

			// Output curr line of stmt (exploding long string lit if neces)
			chan.print(indent);
			String prefix = src.substring(0, pos);
			if (pos>120)
				pos = printLiteral(chan, prefix);
			else println(chan, prefix);

			// Indent subsequent lines more, unless too-big indent
			if (maxlen>80)
			{
				maxlen = 80;
				if (indlen<32)
				{
					indent += "    ";
					indlen += 4;
				}
			}

			// Process residue
			src = src.substring(pos);
			srclen -= pos;
		}							// End outer loop
		chan.print(indent);
		println(chan, src);
	}

	// For all but debugging output to System.out
	// Insure new lines dont start at column where prev ends (LF is only
	// LF -- not implicitly CRLF -- on some systems)
	static void println(PrintWriter chan) {chan.print(lineSep);}
	static void println(PrintWriter ch, String str) {ch.print(str + lineSep);}

	// For debugging output
	static void println(PrintStream chan) {System.out.print(lineSep);}
	static void println(PrintStream ch, String str)
		{System.out.print(str + lineSep);}

	// Write out any changes done in trek
	public void saveTrek()
	{
		mainTrek.saveTrek_();
	}
	private void saveTrek_()
	{
		ClassFile cf;
		int vecsize = cfiles.size();

		// Write/cleanup the program's class files
		for (int i=0; i<vecsize; i++) {
			cf = (ClassFile)cfiles.elementAt(i);
			try {
				cf.write(outDir);
				//cf.finalize();
			}
			catch (Exception e) {
				logError("C="+cf.toQualifiedName()+";", e);
			}
			catch (Throwable th) {
				th.printStackTrace(logChan);
			}
		}
		if (wrote>1)
			writeVerbose("Wrote " + wrote + " classfiles");
	}

	// Set and validate scope-of-trek properties from results of a toString
	private void setClassListHack(String tostr)
	{
		int within = tostr.indexOf(" within ");
		String inclasses = tostr.substring("Trek of ".length(), within);

		// Get indirs
		String indirs;
		int indiridx = within + " within ".length();
		int withsc = tostr.indexOf(" with scope of ");
		if (withsc < 0)
			indirs = tostr.substring(indiridx);
		else indirs = tostr.substring(indiridx, withsc);
		if (indirs.equals("<default>")) indirs = "";

		// Get scope
		String scope;
		if (withsc >= 0)
			scope = tostr.substring(withsc + " with scope of ".length());
		else scope = null;
		setClassList(inclasses, indirs, scope);
	}

	// Set and validate scope-of-trek properties from results of a toString
	public synchronized void setClassList(String trekid)
	{
		// Let old .drj files work
		if (trekid.startsWith("Trek of "))
		{
			setClassListHack(trekid);
			return;
		}

		// See if have valid trek id
		int within = trekid.indexOf('\t');
		int withsc = trekid.indexOf('\t', within+1);
		if (within<0 || withsc<0)
		{
			setClassList(null, null, null);
			return;
		}

		// Break it up and set class list
		String inclasses = trekid.substring(0, within);
		String indirs = trekid.substring(within+1, withsc);
		if (indirs.equals("<default>")) indirs = "";
		String scope = trekid.substring(withsc+1);
		setClassList(inclasses, indirs, scope);
	}

	// Set and validate scope-of-trek properties
	// inclasses=class names (proper case required) sep by ;'s, or null
	// indirs = null (for default) or dirs sep by ;'s
	// rawscope = null (for scope "none") or scope keyword
	public synchronized void setClassList(
		String inclasses, String indirs, String rawscope)
	{
		if (mainTrek!=null)
			throw new TrekException("Trek has already been initialized or joined");

		// Canonicalize class-list and scope
		if (inclasses!=null && inclasses.length()!=0)
			inClassesRaw = inclasses;
		else inclasses = null;
		if (rawscope==null)
			rawscope = "";
		scopeRaw = rawscope;

		// Get default classpath (and insure in canon form) if none specified
		inDirsRaw = indirs;
		if (indirs==null || indirs.length()==0)
		{
			inDirsRaw = "<default>";
			indirs = getDefaultClassPath();
		}

		// Append default classpath info if requested (-ip xxx;+)
		else if (indirs.endsWith("+"))
		{
			int plusidx = indirs.length()-1;
			indirs = indirs.substring(0, plusidx) + getDefaultClassPath();
		}

		// On unix, path sep is :, so user's value and/or pieces of
		// default may have :'s. Force all to be ;'s
		String pathsep = System.getProperty("path.separator");
		if (pathsep!=null && pathsep.equals(":"))
			indirs = indirs.replace(':', ';');

		// Internalize class list
		if (inclasses!=null)
			inclasses = inclasses.replace('.', '/');

		// Map scope string to known scope types
		if (rawscope.length()==0)
			scope = Sc_none;
		else if (rawscope.equals("pkg"))
			scope = Sc_pkg;
		else if (rawscope.equals("p1"))
			scope = Sc_p1;
		else if (rawscope.equals("user"))
			scope = Sc_user;
		else if (rawscope.equals("u1"))
			scope = Sc_u1;
		else if (rawscope.equals("all"))
			scope = Sc_all;
		else throw new TrekException("Invalid scope specified");

		// Store them
		inClasses = inclasses;
		inDirs = indirs;
	}

	// Set order to by-class or by-package.
	// Notes:
	//	class order built by initTrek() so must be done for un-inited trek
	public void setClassOrder(int sortby)
	{
		if (mainTrek!=null)
			throw new TrekException("Trek has already been initialized or joined");
		sortBy = sortby;
	}

	// Debug mode on/off (static so mainTrek doesnt matter)
	public void setDebugMode(boolean onoff)
	{
		debug = onoff ? Db_trace : 0;
	}

	// Undocumented more general version
	public void setDebugMode(int debopt)
	{
		debug = debopt;
	}

	// Verbose mode on/off, & defining verbose sink
	public void setVerboseMode(boolean onoff)
	{
		Trek trek = mainTrek!=null ? mainTrek : this;
		trek.verbose = onoff;
	}
	public void setVerboseSink(VerboseSink sink)
	{
		Trek trek = mainTrek!=null ? mainTrek : this;
		trek.verboseSink = sink!=null ? sink : trek;
	}

	// Set the log-file stream
	public void setLogStream(PrintStream chan)
	{
		Trek trek = mainTrek!=null ? mainTrek : this;
		trek.logChan = new PrintWriter(chan, true);
	}
	public void setLogStream(PrintWriter chan)
	{
		Trek trek = mainTrek!=null ? mainTrek : this;
		trek.logChan = chan;
	}


	// Set and validate output directory
	public void setOutPath(String outdir)
	{
		Trek trek = mainTrek!=null ? mainTrek : this;

		// Nothing to validate for default
		if (outdir==null)
			outdir = "";

		// Validate and save it
		else {
			if (!new File(outdir).isDirectory())
				throw new TrekException(outdir + " is not a directory");
			if (!outdir.endsWith(File.separator))
					outdir += File.separator;
		}
		trek.outDir = outdir;
	}

	// Return error msg (and display trace if virgin error in debug mode)
	// e = error thrown on call to Trek package
	public String toErrorMessage(Throwable e)
	{
		Trek trek = mainTrek!=null ? mainTrek : this;
		return trek.toErrorMessage_(e);
	}
	private String toErrorMessage_(Throwable e)
	{
		// For TrekException's, desired msg is the getMessage data
		// Otherwise use toString
		boolean trekex = e instanceof TrekException;
		String msg = trekex ? e.getMessage() : e.toString();

		// If debug mode, error string (as opposed to error context), and in
		// body of trek or serious error (as opposed to user error in cmd
		// line), print stack trace msg to error stream as well as returning
		// approp msg.
		int msglen = msg.length();
		if ((debug&Db_trace)!=0	&& msglen>0 && (cfiles!=null || !trekex))
			e.printStackTrace();

		// Get real error string or TrekException context
		if (msglen > 0)
			return msg;
		else if (trekex)
			return "Error logged" + toPrettyError((TrekException)e);
		else return "No error string or error context available";
	}

	// Return string that uniquely identifies trek (see setClassList(trekid))
	public String toObjectId()
	{
		Trek trek = mainTrek!=null ? mainTrek : this;
		return trek.toObjectId_();
	}
	private String toObjectId_()
	{
		// Output the pieces, sep by tabs
		return inClassesRaw + '\t' + inDirsRaw + '\t' + scopeRaw;
	}

	// Returns the package name of the specified class name
	// clname = fully qualified internal-fmt class name
	static String toPackage(String clname)
	{
		int dotpos = clname.lastIndexOf('/');
		if (dotpos>=0)
			return clname.substring(0, dotpos);
		else return "";
	}

	// Convert context to readable form
	private String toPrettyError(TrekException tex)
	{
		String doing = tex.piece(" ", "D=");
		String clname = tex.piece(" ", "C=");
		String memname = tex.piece(".", "M=");
		String element = tex.piece(".", "E=");
		return " while" + doing + clname + memname + element;
	}

	// Return string describing the trek that is being done
	// defpathlabel = show default path as <default> if true
	public String toString()
	{
		Trek trek = mainTrek!=null ? mainTrek : this;
		return trek.toString_(true);
	}
	private String toString_(boolean defpathlabel)
	{
		// The normal output string
		String scostr = (scope==Sc_none) ? "" : " with scope of " + scopeRaw;
		String dirs = defpathlabel ? inDirsRaw : inDirs;
		return "Trek of " + inClassesRaw + " within " + dirs + scostr;
	}

	public static String toVersion()
	{
		return Ver_string;
	}

	void writeVerbose(String msg)
	{
		if (verbose)
			verboseSink.atVerbose(msg);
	}

}
