//
// THIS SOFTWARE IS FURNISHED UNDER A LICENSE AND  MAY
// BE USED OR COPIED ONLY IN ACCORDANCE WITH THE TERMS
// OF SUCH LICENSE.
//
//
// COPYRIGHT (c) 1997,1998 BY DIGITAL EQUIPMENT CORPORATION.
// ALL RIGHTS RESERVED.
//

// Import the Trek class library and any other libraries used
import dec.trek.*;
import java.io.*;

//
// profile -- a very simple profiling application.
// It inserts code at start and end of each non-trivial method in a program.
// See profileLog to see what the inserted code does.
//
// To generate the profiling report, run your program. When it
// completes, the time used by each method (and its inferiors) per
// call and the # of times it was called will be written to profile.log.
// Notes:
//	Sub-class Trek because need to override atStartOf in order
//	to use automatic scanning provided by doTrek
//
public final class profile extends Trek
{
	// Each method needs a slot where its profiling info can be stored
	// 0 is always the program's main method
	int methIdx = 0;

	// When the scan reaches a method, insert calls at the start and end 
	// of the method
	// meth = current object of trek
	protected void atStartOf(Method meth)
	{
		// Dont insert stuff in <clinit>'s because exec before main method
		// Dont insert reg stuff in program's main method (see atEndOf()).
		if (meth.toName().equals("<clinit>")) return;
		boolean ismain = meth==getMainMethod();

		// See if anything worth tracking: loops or calls
		// And instrument System.exit calls
		boolean work = false;
		Statement st;
		for (st=meth.firstStatement(); st!=null; st=st.next()) {
			if (st.getType()==Trek.ST_DO
				|| st.getType()==Trek.ST_FOR
				|| st.getType()==Trek.ST_WHILE)
				work = true;
			else if (st.getType()==Trek.ST_CALL)
			{
				work = true;
				checkInstrumentExit(st);
			}

			// Chk for function calls within other stmts if none fnd yet
			if (work || ismain) continue;
			for (Instruction in=st.firstInstruction(); in!=null; in=in.next()) {
				if (in.getType()==Trek.IN_CALL)
				{
					work = true;
					break;
				}
			}
		}

		// Put regular cnting code in if approp
		if (ismain || !work)
			return;						// Main method or fast method

		// ProfileLog.startMethod stores time and bumps "was-called counter".
		Call call = Call.addBefore("ProfileLog.startMethod", meth);
		if (call==null) return;			// Its possible method has no code
		call.passInt(++methIdx);		// Tell startMethod meth's slot

		// Tell it meth's name (which is just class name for constructor)
		String clname = meth.getClassFile().toName();
		String mname = meth.toName();
		if (!clname.equals(mname))
			mname = clname + "." + mname;
		call.passString(mname);
		call.done();

		// ProfileLog.endMethod tabulates time spent in a usage of meth.
		call = Call.addAfter("ProfileLog.endMethod", meth);
		if (call==null) return;
		call.passInt(methIdx);			// Again, tell the slot to use
		call.done();
	}

	// This inserts code at the start and end of your program's main method.
	// Without these calls, no profiling can be done.
	// Notes:
	//	atXxx(Method) nominally inserts code at these locations. Since the
	//	rule is last-inserted goes at end of insert point, could have
	//	simply inserted both endXxx calls, but atStartOf(Method) is executed
	//	before atEndOf(), so startMethod would be before startTrek if
	//	this approach were being used. Instead, just let startTrek/endTrek
	//	handle main method fully.
	protected void atEndOf()
	{
		// ProfileLog.startTrek is what allocates storage to keep profile data.
		Call call = Call.addBefore("ProfileLog.startTrek", this);
		if (call==null)
			throw new TrekException("Need a main program to setup profiling");
		call.passInt(methIdx+1);	// # of slots needed to store profiling data
		call.passString(getMainMethod().getClassFile().toName() + ".main");
										// Tell it meth's name
		call.done();

		// ProfileLog.endTrek is what builds profile.log.
		call = Call.addAfter("ProfileLog.endTrek", this);
		if (call==null)
			throw new TrekException("Could not setup profiling");
		call.done();
	}

	// Check if java.lang.System.exit is called in specified Call stmt
	// Notes:
	//	Conceivably the arg to exit could involve use of Call inst.
	//	So dont return because 1st call is not to exit().
	private void checkInstrumentExit(Statement st) {
		for (Instruction in=st.firstInstruction(); in!=null; in=in.next()) {
			if (in.getType()!=Trek.IN_CALL)
				continue;
			MethodRef ref = (MethodRef)in.getReferenceArg();
			if (!ref.toName().equals("exit"))
				continue;
			if (!ref.getTypeRef().toQualifiedName().equals("java.lang.System"))
				continue;

			// System.exit is called. Insert end-trek code here
			Call call = Call.addAt("ProfileLog.endTrek", in);
			if (call==null)
				throw new TrekException("Could not setup profiling");
			call.done();
			return;
		}
	}

	// Do the scan to insert profiling points in your program.
	// There are no profile-specific cmd line switches. On the other
	// hand, the common argument -sc controls the granularity of profiling
	// done. For example, -sc pkg will lead to there being only entries
	// for methods in your program's package in profile.log.
	// argv = args of the cmd line
	public static void main(String argv[])
	{
		// Create a trek object, get profile cmd line, and do 2-level
		// (classfiles and methods/fields) scan of what was read in 
		// when doTrek called initTrek.
		profile ptrek = new profile();
		try {
			ptrek.getCmdLine(argv);
			ptrek.doTrek(2);
		}

		// Catch any error thrown by Trek package, and output it to console.
		catch (Throwable e) {
			System.out.println(ptrek.toErrorMessage(e));
			System.exit(1);
		}
	}

}

