//
// THIS SOFTWARE IS FURNISHED UNDER A LICENSE AND  MAY
// BE USED OR COPIED ONLY IN ACCORDANCE WITH THE TERMS
// OF SUCH LICENSE.
//
//
// COPYRIGHT (c) 1997 BY DIGITAL EQUIPMENT CORPORATION.
// ALL RIGHTS RESERVED.
//

import dec.trek.*;
import java.io.*;

//
//
// stats
//
//
class stats extends Trek
{
	private PrintWriter outChan;	// Identies report file
	private boolean onlySummary;	// Do only trek totals
	private boolean doClasses;		// Do class-summary lines
	private boolean doMethods;		// Do detailed method report

	// Data to display at trek-level for -classes and -summary
	private int t_cfiles;		// Number of classes seen in trek
	private long t_size;		// Number of bytes in all class files

	private int t_instvars[] = new int[2];	// Number of inst vars in trek (dir, inher)
	private int t_ivbytes[] = new int[2];	// Their size in bytes in trek (dir, inher)
	private int t_statvars[] = new int[2];	// Number of static vars in trek
	private int t_svbytes[] = new int[2];	// Their size in bytes in trek

	// Field data to display at class-level for -classes
	private int c_instvars[] = new int[2];	// Number of inst vars in class
	private int c_ivbytes[] = new int[2];	// Their size in bytes directly in class
	private int c_statvars[] = new int[2];	// Number of static vars in class
	private int c_svbytes[] = new int[2];	// Their size in bytes in class

	// Data to display at method-level

	// Data for methods at method, class, and overall trek
	private int methods[] = new int[3];	// Number of methods (so [0] not used)
	private int stmts[] = new int[3];	// Number of stmts in method
	private int mbytes[] = new int[3];	// Number of inst bytes in method

	private int assigns[] = new int[3];	// Number of assignment stmts
	private int ifsws[] = new int[3];	// Number of if-stmts & switch-stmts
	private int loops[] = new int[3];	// Number of loops in method
	private int thros[] = new int[3];	// Number of throw stmts
	private int trys[] = new int[3];	// Number of try blocks

	// Cell widths for formatting
	private static final int Wid_meth = 28;
	private static final int Wid_nmeth = 7;
	private static final int Wid_var = 22;
	private static final int Wid_st = 8;
	private static final int Wid_stype = 6;
	private static final int Wid_b = 8;

	protected void atStartOf()
	{
		doClasses = !onlySummary && !doMethods;

		// Output report hdr for class report
		if (!doMethods)
		{
			outChan.println(fit("Instance Variables", Wid_var) +
				fit("Static Variables", Wid_var+2) +
				"    Classfile's Code");
			outChan.println(fit("#/Size (Inherited)", Wid_var) +
				fit("#/Size (Inherited)", Wid_var+2) +
				fit("Methods", Wid_nmeth) +
				fit("  Stmts", Wid_st) +
				" Bytecodes");
			outChan.println(
				"-------------------------------------------------" +
				"--------------------");
		}

		// Output report hdr for methods report
		else {
			outChan.println(fit("Methods", Wid_meth+1) +
				fit("  Stmts", Wid_st) +
				fit(" Asgn", Wid_stype) +
				fit("If/Sw", Wid_stype) +
				fit(" Loop", Wid_stype) +
				fit("Throw", Wid_stype) +
				fit("  Try", Wid_stype) +
				" Bytecodes");
			outChan.println(
				"-------------------------------------------------" +
				"----------------------------");
		}
	}
	protected void atEndOf()
	{
		// Output report totals
		if (doClasses || onlySummary)
			printTotals(2, null, t_instvars, t_statvars, t_ivbytes, t_svbytes);
		else {
			String suffix = methods[2]==1 ? "" : "s";
			printMethodInfo(2, methods[2] + " Method" + suffix);
		}

		// Cleanup this level
		finishLevel(2, t_instvars, t_statvars, t_ivbytes, t_svbytes);
		t_cfiles = 0;
		t_size = 0;
	}

	protected void atStartOf(ClassFile cf)
	{
		t_cfiles += 1;
	}
	protected void atEndOf(ClassFile cf)
	{
		// Update summary counters
		countInheritedFields(cf);
		t_size += cf.getSize();
		t_instvars[0] += c_instvars[0];
		t_instvars[1] += c_instvars[1];
		t_ivbytes[0] += c_ivbytes[0];
		t_ivbytes[1] += c_ivbytes[1];

		t_statvars[0] += c_statvars[0];
		t_statvars[1] += c_statvars[1];
		t_svbytes[0] += c_svbytes[0];
		t_svbytes[1] += c_svbytes[1];

		// Print totals for class in approp fmt
		if (doClasses || doMethods)
		{
			if (doClasses)
				printTotals(1, cf, c_instvars, c_statvars, c_ivbytes, c_svbytes);

			// Print methods & method-totals for class
			else {
				printMethodInfo(1, cf.toName());
				for (Method meth=cf.firstMethod(); meth!=null; meth=meth.next()) {
					outChan.println(meth.userDefined);
					meth.userDefined = null;
				}
			}
			outChan.println();
		}

		// Cleanup this level
		finishLevel(1, c_instvars, c_statvars, c_ivbytes, c_svbytes);
	}

	// Accum info for variables (ignore constants)
	protected void at(Field field)
	{
		tabulateField(field, 0);
	}

	protected void atStartOf(Method meth)
	{
		methods[0] += 1;
	}
	protected void atEndOf(Method meth)
	{
		// Build line for the method. Include arg list because of overloading
		if (doMethods)
		{
			String sig = " " + meth.toName() + meth.toArgList();
			meth.userDefined = printMethodInfo(0, sig);
		}
		finishLevel(0, null, null, null, null);
	}

	protected void atStartOf(Statement stmt)
	{
		stmts[0] += 1;
		switch (stmt.getType()) {
		case ST_DO:
		case ST_FOR:
		case ST_WHILE:
			loops[0] += 1;
			break;
		case ST_ASSIGN:
			assigns[0] += 1;
			break;
		case ST_IF:
		case ST_SWITCH:
			ifsws[0] += 1;
			break;
		case ST_THROW:
			thros[0] += 1;
			break;
		case ST_TRY:
			trys[0] += 1;
			break;
		}
	}
	protected void at(Instruction inst)
	{
		mbytes[0] += inst.getSize();
	}

	// Recursively scan super classes
	private void countInheritedFields(ClassFile cf)
	{
		// Recurse if more levels to go
		String supname = cf.toSuperClass();
		if (supname==null)
			return;

		// Access the super class's classfile
		ClassFile sup;
		try {
			sup = getClassFile(supname);
		}
		catch (Throwable th) {
			System.out.println("Could not get super class of " + cf.toName()
				+ " because:");
			System.out.println(toErrorMessage(th));
			return;
		}
		
		// Scan current classfile's instvars
		for (Field field=sup.firstField(); field!=null; field=field.next()) {
			tabulateField(field, 1);
		}
		countInheritedFields(sup);
	}

	// Tabulate next-level method info and clear curr level's method & field info
	private void finishLevel(int level,
		int instvars[], int statvars[], int ivbytes[], int svbytes[])
	{
		// Tabulate method info
		if (level < 2)
		{
			methods[level+1] += methods[level];
			stmts[level+1] += stmts[level];
			mbytes[level+1] += mbytes[level];
			assigns[level+1] += assigns[level];
			ifsws[level+1] += ifsws[level];
			loops[level+1] += loops[level];
			thros[level+1] += thros[level];
			trys[level+1] += trys[level];
		}

		// Clear accum data
		if (level > 0)
		{
			instvars[0] = instvars[1] = statvars[0] = statvars[1] = 0;
			ivbytes[0] = ivbytes[1] = svbytes[0] = svbytes[1] = 0;
		}
		methods[level] = mbytes[level] = stmts[level] = 0;
		assigns[level] = ifsws[level] = loops[level] = 0;
		thros[level] = trys[level] = 0;
	}

	// Pad a left-aligned value to wid
	private static final String cellPad = "                                        ";
	private String fit(String val, int wid)
	{
		return (val+cellPad).substring(0, wid);
	}
	// Pad a right-aligned value to wid
	private String fit(int val, int wid)
	{
		String pval = cellPad+val;
		return pval.substring(pval.length() - wid);
	}

	// Process members-utility command-line stuff
	protected int getCmdLineArg(String argv[], int i)
	{
		// Process one of the common args if present
		int newi = Jtu.getCmdLineArg(argv, i);
		if (newi > 0) return newi;		// Ate this arg

		// Type of report. Last one wins if mult -inc switches
		if (argv[i].equals("-inc"))
		{
			if (i>argv.length-2)
				throw new TrekException("No value specified for -inc");
			i += 1;
			if (argv[i].equals("s"))
			{
				doClasses = doMethods = false;
				onlySummary = true;
				return i + 1;
			}
			if (argv[i].equals("c"))
			{
				onlySummary = doMethods = false;
				doClasses = true;
				return i + 1;
			}
			if (argv[i].equals("m"))
			{
				doClasses = onlySummary = false;
				doMethods = true;
				return i + 1;
			}
		}
		return 0;					// Not a custom arg
	}

	// Do the scan to insert the watch-calls
	public static void main(String argv[])
	{
		stats strek = new stats();
		try {
			strek.getCmdLine(argv);
			strek.outChan = Jtu.openReportFile();
			strek.doTrek();
			strek.outChan.close();
		}
		catch (Throwable e) {
			System.out.println(strek.toErrorMessage(e));
			System.exit(1);
		}
	}

	// Outline line describe curr grouping of method info
	// level = trek, class, or method level
	// methstr = string to identify curr method (grouping)
	private String printMethodInfo(int level, String methstr)
	{
		// Build the line describing the curr group of methods
		String linestr = fit(methstr, Wid_meth) +
			fit(stmts[level], Wid_st) +
			fit(assigns[level], Wid_stype) +
			fit(ifsws[level], Wid_stype) +
			fit(loops[level], Wid_stype) +
			fit(thros[level], Wid_stype) +
			fit(trys[level], Wid_stype) +
			fit(mbytes[level], Wid_b);

		// Print the line or store it
		if (level!=0)
			outChan.println(linestr);
		return linestr;
	}

	// Print totals for methods and instvars
	private void printTotals(int level, ClassFile cf,
		int instvars[], int statvars[], int ivbytes[], int svbytes[])
	{
		// Output classfile line for class or summary
		if (level==2)
		{
			String suffix = t_cfiles==1 ? "" : "s";
			String numcl = t_cfiles + " Classfile" + suffix;
			outChan.println(numcl + " (Total Size=" + t_size + " bytes)");
		}
		else {
			outChan.println(cf.toQualifiedName() + " (File Size=" + cf.getSize() + " bytes)");
		}

		// Compute fld info
		String inststr = instvars[0] + "/" + ivbytes[0]
			+ " (" + instvars[1] + "/" + ivbytes[1] + ")";
		String statstr = statvars[0] + "/" + svbytes[0]
			+ " (" + statvars[1] + "/" + svbytes[1] + ")";

		// Output totals line for class
		outChan.println(fit(inststr, Wid_var) +
			fit(statstr, Wid_var) +
			fit(methods[level], Wid_nmeth) +
			fit(stmts[level], Wid_st) +
			fit(mbytes[level], Wid_b));
	}

	// Tabulate stats for curr field
	// field = fld to process
	// dirinh = direct or inherited index
	private void tabulateField(Field field, int dirinh)
	{
		if (field.hasModifier(ACC_FINAL)) return;

		// Accum static vs. instance vars
		if (field.hasModifier(ACC_STATIC))
		{
			c_statvars[dirinh] += 1;
			c_svbytes[dirinh] += field.getSize();
		}
		else {
			c_instvars[dirinh] += 1;
			c_ivbytes[dirinh] += field.getSize();
		}
	}
}

